/*
  extsmdromot.c

  Dispatcher of external motion functions to parallel port stepper motor
  implementation, based on original extsmmot.c, but using the DRO digital
  readout board for actual encoder feedback.

  Modification history:

  13-Mar-2000 WPS added unused attribute to ident to avoid 'defined but not used' compiler warning.
  24-Feb-2000  FMP created from extsmmot.c
  */

#include "extintf.h"            /* these decls */
#include "parport.h"            /* parallel port funcs */
#include "dro.h"                /* DRO functions */

/* ident tag */
#ifndef __GNUC__
#ifndef __attribute__
#define __attribute__(x)
#endif
#endif

static char __attribute__((unused)) ident[] = "$Id: extsmdromot.c,v 1.1.1.1 2005/05/31 11:12:19 paul Exp $";

#define STEPPER_MAX 6           /* can only handle 6 motors */

char *extModel(void)
{
    return "Parallel port interface with encoder feedback";
}

int extMotInit(const char * stuff)
{
  if (0 != pptDioInit(0)) {
    return -1;
  }

  if (0 != LS7266Init()) {
    return -1;
  }

  return 0;
}

int extMotQuit(void)
{
  return pptDioQuit();
}

int extMotCycle(const char * stuff)
{
  return 0;
}

int extDacNum(void)
{
  return 0;
}

int extDacWrite(int dac, double volts)
{
  return 0;
}

int extDacWriteAll(int max, double * volts)
{
  return 0;
}

int extEncoderSetIndexModel(unsigned int model)
{
  return 0;
}

int extEncoderNum(void)
{
  return 0;
}

int extEncoderRead(int encoder, double * counts)
{
  if (encoder < 0 ||
      encoder >= DRO_MAX) {
    *counts = 0.0;
  }
  else {
    *counts = (double) LS7266Read(encoder);
  }

  return 0;
}

int extEncoderReadAll(int max, double * counts)
{
  int t;

  for (t = 0; t < max; t++) {
    if (t >= DRO_MAX) {
      counts[t] = 0.0;
    }
    else {
      counts[t] = (double) LS7266Read(t);
    }
  }

  return 0;
}

int extEncoderResetIndex(int encoder)
{
  return 0;
}

int extEncoderReadLatch(int encoder, int * flag)
{
  *flag = 1;                    /* always on "index pulse" */

  return 0;
}

int extEncoderReadLevel(int encoder, int * flag)
{
  return 0;
}

int extAmpEnable(int axis, int enable)
{
  return 0;
}

int extAmpFault(int axis, int * fault)
{
  return 0;
}

/*
  We gang the max limits onto 1, mins onto another, and homes onto a third:

  (DioRead)
  0  1-6 joints lim + (S3)
  1  1-6 joints lim - (S4)
  2  1-6 joints home  (S5)
  3  unused
  4  unused
*/

int extMaxLimitSwitchRead(int axis, int * flag)
{
  if (axis < 0 || axis >= STEPPER_MAX) {
    *flag = 0;
    return 0;
  }

  pptDioRead(0, flag);

  return 0;
}

int extMinLimitSwitchRead(int axis, int * flag)
{
  if (axis < 0 || axis >= STEPPER_MAX) {
    *flag = 0;
    return 0;
  }

  pptDioRead(1, flag);

  return 0;
}

int extHomeSwitchRead(int axis, int * flag)
{
  if (axis < 0 || axis > STEPPER_MAX) {
    *flag = 0;
    return 0;
  }

  pptDioRead(2, flag);

  return 0;
}
