/*
  parport.c

  External interface to parallel port I/O

  Processes need to be root, or run setuid-root, in order to run this,
  even with the call to ioperm().

  Modification history:

  23-Oct-2003  FMP changed pptXXXCheck so that they set the values
  to zero if they're out of range
  15-May-2002 P.C. Changed PARPORT_IO_ADDRESS to IO_BASE_ADDRESS
  28-Jun-2000 WPS changed %i to %l in sscanf to avoid warning.
  27-Jun-2000 WPS changed asm/io.h to sys/io.h to avoid a bunch of warnings under linux_2_2_14.
  20-Mar-2000 WPS added unused attribute to parport_h to avoid 'defined but not used' compiler warning
  29-Sep-1999  FMP moved #include "rcs_prnt.h" into don't DO_IT section
  21-Sep-1999  WPS eliminate sscanf and printf calls not supported under CE.
  7-Aug-1998  FMP changed extInit/Quit() to extDioInit/Quit()
  5-Aug-1998  FMP added #ifndef RTLINUX around ioperm() call in pptInit(),
  since it's not needed for RT-Linux
  10-Jul-1998  FMP added -addr flag to set io address
  8-Jul-1998  FMP changed PPT_ to PARPORT_ for consistency with INI file
  6-Nov-1997  FMP added pptDioCheck(); removed inversion correction
  29-Oct-1997 FMP fixed bugs, inverted the reverse polarity ones
  24-Oct-1997 FMP created
  */

/* ident tag */
#ifndef __GNUC__
#ifndef __attribute__
#define __attribute__(x)
#endif
#endif

static char __attribute__ ((unused)) ident[] =
    "$Id: parport.c,v 1.2 2005/08/08 22:56:38 paul Exp $";

#ifdef MAIN

#include "config.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#endif

#include <asm/bitops.h>
#include "parport.h"		/* these decls */

/* the IO address of the board; override by setting global before call
   to pptDioInit() */
int IO_BASE_ADDRESS = DEFAULT_IO_BASE_ADDRESS;

#define PARPORT_IO_WIDTH 3	/* number of bytes taken up */


/* real parallel port IO */

#ifdef nonrealtime
#include <unistd.h>		/* ioperm() */
#include <sys/io.h>		/* outb, inb */
#endif

#ifdef realtime
#ifdef __KERNEL__
#include <linux/kernel.h>
#endif
#include <linux/fs.h>
#include <linux/parport.h>
#include <asm/io.h>

#endif

static int port_void = 1;

#ifdef realtime
static struct pardevice *ppt_dev = NULL;

static void ppt_attach(struct parport *port)
{
    /* At some point we may want to respond to an "event" such as a
       pcmcia parport being plugged in... */
    return;
}

static void ppt_detach(struct parport *port)
{
    port_void = 1;
    return;
}

static struct parport_driver ppt_driver = {
    .name = "emcmot",
    .attach = ppt_attach,
    .detach = ppt_detach,
};
#endif

/* set up bits for R/W. In this case, they're not programmable by
   register so nothing need be done. */
int pptDioInit(const char *stuff)
{
    int retval = 0;

#ifdef realtime
    struct parport *ppt;

    if (ppt_dev != NULL) {
	printk(KERN_ERR "emcmot: pptDioInit called multiple times.\n");
	return 0;
    }
    /* Register with the parport subsystem so that in the unlikely event of
       a pcmcia card disappearing, we don't read or write in to empty space. */
    if (parport_register_driver(&ppt_driver)) {
	printk(KERN_ERR "emcmot: unable to register with parport\n");
	return -EIO;
    }

    /*! \note parport_find_number(num) could be used here. That would
       simplify configuration and would not require the user to track down
       base addresses (problematic for some PCI cards). */
    ppt = parport_find_base(IO_BASE_ADDRESS);

    if (ppt == NULL) {
	printk(KERN_ERR "emcmot: No parport found at 0x%x\n",
	       IO_BASE_ADDRESS);
	parport_unregister_driver(&ppt_driver);
	return -ENODEV;
    }
    ppt_dev = parport_register_device(ppt, "emcmot",
				      NULL, NULL, NULL, 0, NULL);
    parport_put_port(ppt);

    if (!ppt_dev) {
	parport_unregister_driver(&ppt_driver);
	ppt_dev = NULL;
	return -ENODEV;
    }

    if (parport_claim(ppt_dev)) {
	printk(KERN_ERR
	       "emcmot: %s busy. Can not claim for realtime control.\n",
	       ppt_dev->port->name);
	parport_unregister_device(ppt_dev);
	parport_unregister_driver(&ppt_driver);
	ppt_dev = NULL;
	return -EBUSY;
    }
#else
    /* turn on port access */
    retval = ioperm(IO_BASE_ADDRESS, PARPORT_IO_WIDTH, 1);
    if (retval)
	retval = iopl(3);
#endif
    port_void = retval;
    return retval;
}

int pptDioQuit(void)
{
    int retval = 0;

#ifdef realtime
    if (ppt_dev == NULL) {
	printk(KERN_ERR
	       "emcmot: pptDioQuit called without parport being registered.\n");
	return 0;
    }
    parport_release(ppt_dev);
    parport_unregister_device(ppt_dev);
    parport_unregister_driver(&ppt_driver);
    ppt_dev = NULL;
#else
    /* turn off port access */
    retval = ioperm(IO_BASE_ADDRESS, PARPORT_IO_WIDTH, 0);
#endif
    port_void = 1;
    return retval;
}

/*
  Digital I/O

  Bitmap:
  Data Out        W   Base+0      D0-D7   8 LS TTL outputs
  Status In       R   Base+1      S3-S7   5 LS TTL inputs
  Control Out     W   Base+2      C0-C3   4 TTL Open Collector outputs

 */

int pptDioMaxInputs(void)
{
    return 5;
}

int pptDioMaxOutputs(void)
{
    return 12;
}

/*
  pptDioRead() maps 5 index values 0-4 onto S3-S7, respectively,
  providing 5 digital inputs

  Note that S7=4 is reverse polarity.
  */

int pptDioRead(int index, int *value)
{
    unsigned char inByte;
    unsigned char mask = 8;	/* start at 00001000 for S3 */

    if ((index < 0 || index > 4) && port_void) {
	*value = 0;
	return -1;
    }

    /* latch input */
    inByte = inb(IO_BASE_ADDRESS + 1);
    mask <<= index;
    *value = (inByte & mask) ? 1 : 0;

    return 0;
}

/*
  pptDioCheck() maps 8 index values 0-7 onto D0-D7, respectively, and
  4 index values 8-11 onto C0-C3, respectively, providing read checks
  on the 12 digital outputs.

  Note that C0, C1, and C3 are inverted outputs; 'value' will return
  what was written to these, so checking a 1 means a 1 was last written
  and the output is 0V.
  */

int pptDioCheck(int index, int *value)
{
    unsigned char inByte;
    unsigned char mask = 1;	/* start at 00000001 for D0, C0 */

    if ((index < 0 || index > 11) && port_void) {
	*value = 0;
	return -1;
    }

    if (index < 8) {
	inByte = inb(IO_BASE_ADDRESS);
	mask <<= index;
	*value = (inByte & mask) ? 1 : 0;
	return 0;
    }

    /* else index is 8-11 */
    inByte = inb(IO_BASE_ADDRESS + 2);
    mask <<= index - 8;
    *value = (inByte & mask) ? 1 : 0;
    return 0;
}

/*
  pptDioWrite() maps 8 index values 0-7 onto D0-D7, respectively, and
  4 index values 8-11 onto C0-C3, respectively, providing 12 digital outputs.

  Note that C0, C1, and C3 are inverted outputs.
  */
int pptDioWrite(int index, int value)
{
    unsigned char outByte;
    unsigned char mask = 1;	/* start at 00000001 for D0, C0 */

    if ((index < 0 || index > 11) && port_void) {
	return -1;
    }

    if (index < 8) {
	/* read current outputs */
	outByte = inb(IO_BASE_ADDRESS);
	mask <<= index;
	if (value) {
	    outByte |= mask;
	} else {
	    outByte &= ~mask;
	}
	outb(outByte, IO_BASE_ADDRESS);
	return 0;
    }

    /* read current outputs */
    outByte = inb(IO_BASE_ADDRESS + 2);
    mask <<= index - 8;
    if (value) {
	outByte |= mask;
    } else {
	outByte &= ~mask;
    }
    outb(outByte, IO_BASE_ADDRESS + 2);
    return 0;
}

int pptDioByteRead(int index, unsigned char *byte)
{
    if (index != 0 && port_void) {
	*byte = 0;
	return -1;
    }

    *byte = inb(IO_BASE_ADDRESS + 1);
    *byte >>= 3;		/* shift down since S3 is bit 0, etc. */
    return 0;
}

int pptDioByteWrite(int index, unsigned char byte)
{
    if (index == 0 && !port_void) {
	outb(byte, IO_BASE_ADDRESS);
	return 0;
    }

    if (index == 1 && !port_void) {
	outb(byte, IO_BASE_ADDRESS + 2);
	return 0;
    }

    return -1;
}

int pptDioByteCheck(int index, unsigned char *byte)
{
    if (index == 0 && !port_void) {
	*byte = inb(IO_BASE_ADDRESS);
	return 0;
    }

    if (index == 1 && !port_void) {
	*byte = inb(IO_BASE_ADDRESS + 2);
	return 0;
    }

    *byte = 0;
    return 0;
}

#ifdef MAIN

static void fillIn(int *ins)
{
    int t;

    for (t = 0; t < 5; t++) {
	pptDioRead(t, &ins[t]);
    }
}

static void printIn(int *ins)
{
    int t;

    for (t = 0; t < 5; t++) {
	printf("%d ", ins[t]);
    }
    printf("\n");
}

static void fillOut(int *outs)
{
    int t;

    for (t = 0; t < 12; t++) {
	pptDioCheck(t, &outs[t]);
    }
}

static void printOut(int *outs)
{
    int t;

    for (t = 0; t < 12; t++) {
	printf("%d ", outs[t]);
    }
    printf("\n");
}

int main(int argc, char *argv[])
{
    int t;
    char input[LINELEN];
    char cmd[LINELEN];
    int ins[5];
    int outs[12];
    int index;

    /* check -addr 0x378 */
    for (t = 1; t < argc; t++) {
	if (!strcmp(argv[t], "-addr")) {
	    if (t == argc - 1) {
		fprintf(stderr, "no address specified\n");
		exit(1);
	    } else if (1 != sscanf(argv[t + 1], "%i", &IO_BASE_ADDRESS)) {
		fprintf(stderr, "bad address specified: %s\n",
			argv[t + 1]);
		exit(1);
	    } else {
		t++;		/* step over address */
		continue;
	    }
	} else {
	    fprintf(stderr, "syntax: %s -addr <io address>\n", argv[0]);
	    exit(1);
	}
    }

    if (-1 == pptDioInit(0)) {
	fprintf(stderr, "can't init\n");
	exit(1);
    }

    while (!feof(stdin)) {
	printf("ppt> ");
	fflush(stdout);

	if (NULL == fgets(input, LINELEN, stdin)) {
	    /* EOF */
	    break;
	}

	if (1 != sscanf(input, "%s", cmd)) {
	    /* blank line-- print input */
	    fillIn(ins);
	    printIn(ins);
	    continue;
	}

	if (!strcmp(cmd, "help") || !strcmp(cmd, "?")) {
	    printf("help/?           print this help\n");
	    printf("s <#>            set bit #\n");
	    printf("c <#>            clear bit #\n");
	    printf("k                check outputs\n");
	    printf("ENTER            print inputs\n");
	    printf("q                quit\n");
	} else if (!strcmp(cmd, "s")) {
	    /* set */
	    if (1 != sscanf(input, "%*s %d", &index)) {
		printf("bad input: %s\n", input);
		continue;
	    }

	    if (0 != pptDioWrite(index, 1)) {
		printf("index out of range\n");
	    }
	} else if (!strcmp(cmd, "c")) {
	    /* set */
	    if (1 != sscanf(input, "%*s %d", &index)) {
		printf("bad input: %s\n", input);
		continue;
	    }

	    if (0 != pptDioWrite(index, 0)) {
		printf("index out of range\n");
	    }
	} else if (!strcmp(cmd, "k")) {
	    /* check outputs */
	    fillOut(outs);
	    printOut(outs);
	} else if (!strcmp(cmd, "q")) {
	    break;
	} else {
	    printf("?\n");
	}
    }

    pptDioQuit();

    exit(0);
}

#endif
