/*
  ppmc_dio.c

  dio functions for PPMC and Universal Stepper boards.
  
  Modification history:

   25-Apr-2005 JE detect DIO card with 11/20/2004 mod (module ID register
                      appears at address more consistent with other modules)
   4-Apr-2005 JE  disable MAX_LIM_3 bit for Univ stepper and univ PWM
                     as that bit is taken over by OK-to-come-out-of-ESTOP
   7-Aug-2004 JE turn on SSR8 when out of estop
  14-Jan-2004 JE added support for Universal PWM controller
  16-Nov-2002 P.C. Commented out shared memory writes in ppmcDioInit()
  14-Jul-2002 JE mod to make access to EMCMOT_STRUCT compatible with BDI
                      versions.
  11-Aug-2001 JE  mod to accept digital Output values from EMCMOT_IO structure
                      in EMCMOT_STRUCT meta-structure.
  10 oct 2000 eek.  split out ppmc_dio from other ppmc functions
  this is likely to become quite large as all the desirable functions
  are incorporated

  */
#include "ppmc.h"		/* decls for ppmc implementation */


/* ident tag */
#ifndef __GNUC__
#ifndef __attribute__
#define __attribute__(x)
#endif
#endif

static char __attribute__((unused)) ident[] = "$Id: ppmc_dio.c,v 1.3 2005/07/28 11:32:17 paul Exp $";

#include <asm/io.h>
#include "emcmot.h"
#include "diagnostics.h"	/* diagnostics printing */

/* need to point to emcmotIo */
extern EMCMOT_IO *emcmotIo;

#define CONTROLPORT ppmc_base_addr+2
#define DATAPORT    ppmc_base_addr+4

int ppmcDioInit(const char * stuff)
{
  int i,j;
  diagnostics("ppmc_dio entering ppmcDioInit \n");
  //  diagnostics("ppmc_dio emcmotIo at location %x\n",emcmotIo);
  //  diagnostics("ppmc_dio emcmotStruct at location %x\n",emcmotStruct);
  //  diagnostics("ppmc_dio emcmotIo->NumRead = %d\n",emcmotIo->NumRead);
  //  diagnostics("ppmc_dio emcmotIo->NumWrite = %d\n",emcmotIo->NumWrite);
  //  diagnostics("ppmc_dio emcmotIo->InBytes[0..2] = %d %d %d\n",emcmotIo->InBytes[0],
  //	      emcmotIo->InBytes[1],emcmotIo->InBytes[2]);
#ifdef UNIV_STEP
  j=ENCCNT0+15;
#else
  j=DINA_0+3;
#endif
  SelAddr(j);
  outb(0x24,CONTROLPORT);    /* set EPP port to input mode */
  i = inb(DATAPORT);
  if ((i & 0xf0) != 0x30 && (i & 0Xf0) != 0x40 && (i & 0xf0) != 0x50) {
    diagnostics("ppmc_dio could not find a DIO module at IEEE-1284 address 0x%2x\n",j); 
  }
  /* if not found at addr 0x63, try at addr 0x67 */
  SelAddr(j+4);
  outb(0x24,CONTROLPORT);    /* set EPP port to input mode */
  i = inb(DATAPORT);
  if ((i & 0xf0) != 0x30 && (i & 0Xf0) != 0x40 && (i & 0xf0) != 0x50) {
    diagnostics("ppmc_dio could not find a DIO module at IEEE-1284 address 0x%2x\n",j+4); 
    diagnostics("ppmc_dio IEEE-1284 address 0x%2x reads 0x%2x\n",j+4,i); 
    
    return -1;              /* blow it up ! */
  } else {
    if ((i & 0xf0) == 0x30) {
      diagnostics("ppmc_dio found a DIO ver %d module at IEEE-1284 address 0x%2x\n",i&0x0f,j);}
    if ((i & 0xf0) == 0x40) {
      diagnostics("ppmc_dio found a Universal Stepper ver %d module at IEEE-1284 address 0x%2x\n",i&0x0f,j); }
    if ((i & 0xf0) == 0x50) {
      diagnostics("ppmc_dio found a Universal PWM ver %d module at IEEE-1284 address 0x%2x\n",i&0x0f,j); }
  }
  /* Really don't want to be doing shared memory accesses here - The buffers
     have not been set up yet, so a segmentation fault WILL occur. */
  //  emcmotIo->InBytes[0] = 123;       /* make existance known */
  //  emcmotIo->OutBytes[0] = 9;  /* testing for DIO */
  //  emcmotIo->NumWrite = 1;
  return 0;
}

int DIOQuit(void)
{
	return 0;
}
/* PPMC/Universal Stepper digital input bit masks, 32-bit word */
/* maps to ports DINA_0 and DINB_0 */
#define HOME_0    0x00000001
#define MIN_LIM_0 0x00000002
#define MAX_LIM_0 0x00000004
#define FAULT_0   0x00000008
#define HOME_1    0x00000010
#define MIN_LIM_1 0x00000020
#define MAX_LIM_1 0x00000040
#define FAULT_1   0x00000080
#define HOME_2    0x00000100
#define MIN_LIM_2 0x00000200
#define MAX_LIM_2 0x00000400
#define FAULT_2   0x00000800
#define HOME_3    0x00001000
#define MIN_LIM_3 0x00002000
#ifdef UNIV_STEP  // defined for both USC and UPC
#define MAX_LIM_3 0x00002000
#else
#define MAX_LIM_3 0x00004000
#endif
#define FAULT_3   0x00008000    // really ESTOP

/* maps to ports DINA_1 and DINB_1 if present */
#define HOME_4    0x01000000
#define MIN_LIM_4 0x02000000
#define MAX_LIM_4 0x04000000
#define FAULT_4   0x08000000
#define HOME_5    0x10000000
#define MIN_LIM_5 0x20000000
#define MAX_LIM_5 0x40000000
#define FAULT_5   0x80000000


int ppmcMaxLimitSwitchRead(int axis, int * flag)
{
  long bits;
  int retval = 0;

  if (emcmotIo->NumRead == 0) return 0;
  if (emcmotIo->NumRead >= EMCMOT_IO_SIZE) return -1;
  bits = (emcmotIo->InBytes[0]) | ((emcmotIo->InBytes[1])<<8) | ((emcmotIo->InBytes[2]) << 16);
  *flag = 0;

  switch (axis)
    {
    case 0:
      if (bits & MAX_LIM_0)
	{
	  *flag = 1;
	}
      break;

    case 1:
      if (bits & MAX_LIM_1)
	{
	  *flag = 1;
	}
      break;

    case 2:
      if (bits & MAX_LIM_2)
	{
	  *flag = 1;
	}
      break;

    case 3:
      if (bits & MAX_LIM_3)
	{
	  *flag = 1;
	}
      break;

#ifdef PPMC_8_AXIS
    case 4:
      if (bits & MAX_LIM_4)
	{
	  *flag = 1;
	}
      break;

    case 5:
      if (bits & MAX_LIM_5)
	{
	  *flag = 1;
	}
      break;

#endif

    default:
      retval = -1;
      break;
    }

  return retval;
}


int ppmcMinLimitSwitchRead(int axis, int * flag)
{
  long bits;
  int retval = 0;

  if (emcmotIo->NumRead == 0) return 0;
  if (emcmotIo->NumRead >= EMCMOT_IO_SIZE) return -1;
  bits = (emcmotIo->InBytes[0]) | ((emcmotIo->InBytes[1])<<8) | ((emcmotIo->InBytes[2]) << 16);
  *flag = 0;

  switch (axis)
    {
    case 0:
      if (bits & MIN_LIM_0)
	{
	  *flag = 1;
	}
      break;

    case 1:
      if (bits & MIN_LIM_1)
	{
	  *flag = 1;
	}
      break;

    case 2:
      if (bits & MIN_LIM_2)
	{
	  *flag = 1;
	}
      break;

    case 3:
      if (bits & MIN_LIM_3)
	{
	  *flag = 1;
	}
      break;

#ifdef PPMC_8_AXIS
    case 4:
      if (bits & MIN_LIM_4)
	{
	  *flag = 1;
	}
      break;

    case 5:
      if (bits & MIN_LIM_5)
	{
	  *flag = 1;
	}
      break;

#endif

    default:
      retval = -1;
      break;
    }

  return retval;
}


int ppmcHomeSwitchRead(int axis, int * flag)
{
  long bits;
  int retval = 0;

  if (emcmotIo->NumRead == 0) return 0;
  if (emcmotIo->NumRead >= EMCMOT_IO_SIZE) return -1;
  bits = (emcmotIo->InBytes[0]) | ((emcmotIo->InBytes[1])<<8) | ((emcmotIo->InBytes[2]) << 16);
  *flag = 0;

  switch (axis)
    {
    case 0:
      if (bits & HOME_0)
	{
	  *flag = 1;
	}
      break;

    case 1:
      if (bits & HOME_1)
	{
	  *flag = 1;
	}
      break;

    case 2:
      if (bits & HOME_2)
	{
	  *flag = 1;
	}
      break;

    case 3:
      if (bits & HOME_3)
	{
	  *flag = 1;
	}
      break;

#ifdef PPMC_8_AXIS
    case 4:
      if (bits & HOME_4)
	{
	  *flag = 1;
	}
      break;

    case 5:
      if (bits & HOME_5)
	{
	  *flag = 1;
	}
      break;

#endif

    default:
      retval = -1;
      break;
    }

  return retval;
}


int ppmcAmpEnable(int axis, int enable)
{
  return 0;
}

int ppmcAmpFault(int axis, int * fault)
{
  return 0;
}


int ppmcDioQuit(void)
{
  return 0;
}

int ppmcDioMaxInputs(void)
{
  return 0;
}

int ppmcDioMaxOutputs(void)
{
  return 0;
}

int ppmcDioRead(int index, int *value)
     /* If index == -1 then read all input ports into memory */
{
  int i;
  if (emcmotIo->NumRead == 0) return 0;
  if (index == -1)
    {
      SelAddr(DINA_0);
      outb(0x24,CONTROLPORT);    /* set EPP port to input mode */
      for (i=0; i < emcmotIo->NumRead ; i++)
	{
	  emcmotIo->InBytes[i] = inb(DATAPORT);
	}
      return 0;
    }
  if (emcmotIo->NumRead >= EMCMOT_IO_SIZE) return -1;
  /* if index positive, then return content recorded in memory */
  *value = emcmotIo->InBytes[index];
  return 0;
}

/* this is the routine that writes the digital Output values out to
   the device */
int ppmcDioWrite(int index, int value)
     /* the paramemters are dummies */
{
  int i;
#ifdef DIO_DEBUG
  static int counter = 0;
#endif

  if (emcmotIo->NumWrite <= 0) return 0;
  if (emcmotIo->NumWrite >= EMCMOT_IO_SIZE) return -1;
#ifdef UNIV_STEP
#ifdef DIO_DEBUG
  counter += 1;
  if (counter > 10000)
    {
      counter = 0;
      //      diagnostics("ppmcDioWrite, io is at 8%x\n",(long) emcmotIo);
      //diagnostics("ppmcDioWrite, NumWrite %d, OutBytes[0] %d\n",emcmotIo->NumWrite,emcmotIo->OutBytes[0]);
    }
#endif
  SelAddr(DOUTA_0);
  i = 0;
  //    bit 7 (value = 128) is the estop bit, write 1 there to get OUT of estop
  outb(emcmotIo->OutBytes[i],DATAPORT);
  return 0;
#else
  SelAddr(ESTOP_0);    /* DIO card kludge right now to take out of ESTOP */
  outb(emcmotIo->OutBytes[2],DATAPORT);
  SelAddr(DOUTA_0);
  i = 0;  // force bit 7 on for PPMC DIO card 8/7/04 JE
     outb((emcmotIo->OutBytes[i] | 128),DATAPORT);
  return 0;
#endif
}

int ppmcDioCheck(int index, int *value)
{
  return 0;
}

int ppmcDioByteRead(int index, unsigned char *byte)
{
  return 0;
}

int ppmcDioShortRead(int index, unsigned short *sh)
{
  return 0;
}

int ppmcDioWordRead(int index, unsigned int *word)
{
  return 0;
}

int ppmcDioByteWrite(int index, unsigned char byte)
{
  return 0;
}

int ppmcDioShortWrite(int index, unsigned short sh)
{
  return 0;
}

int ppmcDioWordWrite(int index, unsigned int word)
{
  return 0;
}

int ppmcDioByteCheck(int index, unsigned char *byte)
{
  return 0;
}

int ppmcDioShortCheck(int index, unsigned short *sh)
{
  return 0;
}

int ppmcDioWordCheck(int index, unsigned int *word)
{
  return 0;
}
