/*
  ppmcparport.c

  External interface to parallel I/O through the PPAC card set

  Processes need to be root, or run setuid-root, in order to run this,
  even with the call to ioperm().

  Modification history:
  16-Mar-2004  FMP added if defined (HAVE_RTL) for mbuff.h include
  4-Oct-2002 P.C. Included rtai support.
  11-Aug-2001  JE  begin conversion
  29-Sep-1999  FMP moved #include "rcs_prnt.h" into don't DO_IT section
  21-Sep-1999  WPS eliminate sscanf and printf calls not supported under CE.
  7-Aug-1998  FMP changed extInit/Quit() to extDioInit/Quit()
  5-Aug-1998  FMP added #ifndef RTLINUX around ioperm() call in pptInit(),
                   since it's not needed for RT-Linux
  10-Jul-1998  FMP added -addr flag to set io address
  8-Jul-1998  FMP changed PPT_ to PARPORT_ for consistency with INI file
  6-Nov-1997  FMP added pptDioCheck(); removed inversion correction
  29-Oct-1997 FMP fixed bugs, inverted the reverse polarity ones
  24-Oct-1997 FMP created
  */

/* ident tag */
#ifndef __GNUC__
#ifndef __attribute__
#define __attribute__(x) 
#endif
#endif

static char __attribute__((unused)) ident[] = "$Id: ppmcparport.c,v 1.3 2005/08/08 22:56:38 paul Exp $";

#include <stdio.h>
#include <string.h>		/* memcpy() */
#include <stdlib.h>		/* sizeof() */
#include <unistd.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <math.h>

#include "rtai_shm.h"

#include "_timer.h"		/* rcslib esleep() */
#include "_shm.h"		/* rcslib shm_t, rcs_shm_open(), ... */
#include "rcs_print.hh"          /* set rcs_print_destination(), ... */

#include "emcmot.h"
#include "emcmot.h"            /* define shared-memory structures to talk to RT side */
#include "emcmotglb.h"
#include "inifile.hh"

#include "ppmcparport.h"		/* these decls */


#if !defined(USE_RCS_SHM_GET_ADDR) && !defined(rcs_shm_get_addr)
#define rcs_shm_get_addr(x) ((x)->addr)
#endif

/* the IO address of the board; override by setting global before call
   to pptDioInit() */
unsigned long int IO_BASE_ADDRESS = DEFAULT_IO_BASE_ADDRESS;
static int inited = 0;		/* flag if inited */
static int usingShmem = 0;
static shm_t *shmem = NULL;

static EMCMOT_STRUCT *emcmotStruct = 0;
static  EMCMOT_IO *emcmotIo = 0;            /* pointer to I/O intf structure in shared memory */

/* usrmotIniLoad() loads params (SHMEM_KEY, SHMEM_BASE_ADDRESS,
   COMM_TIMEOUT, COMM_WAIT) from named ini file */
int ppmcIniLoad(const char *filename)
{
  FILE *fp;
  const char *inistring;
  int saveInt;

  /* open it */
  if (NULL == (fp = fopen(filename, "r")))
    {
      rcs_print("can't find emcmot ini file %s\n", filename);
      return -1;
    }

  saveInt = SHMEM_KEY;
  if (NULL != (inistring = iniFind(fp, "SHMEM_KEY", "EMCMOT")))
    {
      if (1 == sscanf(inistring, "%i", &SHMEM_KEY))
	{
	  /* found it */
	}
      else
	{
	  /* found, but invalid */
	  SHMEM_KEY = saveInt;
	  rcs_print("invalid [EMCMOT] SHMEM_KEY in %s (%s); using default %d\n",
		    filename, inistring, SHMEM_KEY);
	}
    }
  else
    {
      /* not found, using default */
      rcs_print("[EMCMOT] SHMEM_KEY not found in %s; using default %d\n",
		filename, SHMEM_KEY);
    }

  fclose(fp);

  return 0;
}



/* instead of doing real I/O to the hardware, we process all changes to the
   values to be output, and keep them in a memory location available to
   the real-time section through the EMCMOT_IO structure in the
   EMCMOT_STRUCT meta-structure.

   The RT section picks up the values to be output and sends them out every
   servo update cycle when it is accessing the servo hardware.
   */

int pptDioInit(const char * stuff)
{
//  int i; // Used in a debug routine that is also commented out.
  RCS_PRINT_DESTINATION_TYPE dest;

  /* read comm parameters */
  if (-1 == ppmcIniLoad(EMCMOT_INIFILE)) {
    fprintf(stderr, "can't load ini file %s\n", EMCMOT_INIFILE);
    exit(1);
  }

  usingShmem = 0;

  /* try OS shared memory first, inhibiting error messages */
  dest = get_rcs_print_destination();
  set_rcs_print_destination(RCS_PRINT_TO_NULL);
  shmem = rcs_shm_open(SHMEM_KEY, sizeof(EMCMOT_STRUCT), 0);
  if (NULL != shmem &&
      NULL != rcs_shm_get_addr(shmem))
  {
    /* map shmem area into local address space */
    emcmotStruct = (EMCMOT_STRUCT *) rcs_shm_get_addr(shmem);
    usingShmem = 1;
  }
  set_rcs_print_destination(dest);

  if (! usingShmem)
  {
  emcmotStruct = rtai_malloc(SHMEM_KEY,sizeof(EMCMOT_STRUCT));
  if (emcmotStruct == NULL)
    {
      printf("rtai_alloc(%d (0x%X), %d (0x%X) ) failed\n",
	      SHMEM_KEY, SHMEM_KEY, sizeof(EMCMOT_STRUCT), sizeof(EMCMOT_STRUCT));
      return -1;
    }
  }

  /* got it */
  emcmotIo = &(emcmotStruct->io);
  printf("pptDioInit in ppmcparport.c, emcmotIo struct at 0x%8X \n",(int) emcmotIo);

  emcmotIo->OutBytes[0] = 0;     /* put a zero out there first to avoid turning
				  on any hardware before we are ready for it */
  emcmotIo->OutBytes[1] = 0;     /* additional bytes */
  emcmotIo->OutBytes[2] = 0;     /* additional bytes */
  emcmotIo->NumWrite = 3;        /* tell servo write logic that there is only one
				  byte of digital output to be sent */
  emcmotIo->NumRead = 3;        /* tell encoder read logic to read 3 bytes from
				   digital input */
  emcmotIo->InBytes[1] = 55;
  /*
    printf("NumWrite = %d  Out Bytes = ",emcmotIo->NumWrite);
    for (i=0;i<3; i++)
      {
	printf(" %d = %d ",i,emcmotIo->OutBytes[i]);
      }
    printf("\n");
  */
  inited = 1;
  return 0;
}

int pptDioQuit()
{
  printf("pptDioQuit in ppmcparport.c\n");

  /* turn off port access */
  emcmotIo->OutBytes[0] = 0;    /* put a zero out there first to avoid turning
				      on any hardware before we are ready for it */
  emcmotIo->NumWrite    = 0;    /* tell servo write logic to not write to ANY 
				    digital output ports */
  if (usingShmem)
  {
    if (NULL != shmem)
    {
      rcs_shm_close(shmem);
      shmem = NULL;
    }
  }
  else
  {
    rtai_free(SHMEM_KEY,emcmotStruct);
  }

  emcmotStruct = 0;
  emcmotIo = 0;
  inited = 0;
  return 0;
}

int pptDioMaxInputs()
{
  return 18;
}

int pptDioMaxOutputs()
{
  return 18;
}

/*
  pptDioRead() maps 5 index values 0-4 onto S3-S7, respectively,
  providing 5 digital inputs
 
  Note that S7=4 is reverse polarity.
  */

int pptDioRead(int index, int *value)
{
  unsigned char mask = 1;	/* start at 00000001 for S3 */

  if (index < 0 ||
      index > 18)
    {
      *value = 0;
      return -1;
    }

  if (index <8)
    {
      mask <<= index;
      *value = (emcmotIo->InBytes[0] & mask) ? 1 : 0;
    }
  else {
    if (index < 16)
      {
	mask <<= index -8;
	*value = (emcmotIo->InBytes[1] & mask) ? 1 : 0;
      }
    else
      {
	mask <<= index -16;
	*value = (emcmotIo->InBytes[2] & mask) ? 1 : 0;
      }
  }
      
  return 0;
}

/*
  pptDioCheck() maps 8 index values 0-7 onto D0-D7, respectively, and
  4 index values 8-11 onto C0-C3, respectively, providing read checks
  on the 12 digital outputs.

  Note that C0, C1, and C3 are inverted outputs; 'value' will return
  what was written to these, so checking a 1 means a 1 was last written
  and the output is 0V.
  */

int pptDioCheck(int index, int *value)
{
  unsigned char mask = 1;	/* start at 00000001 for D0, C0 */

  if (index < 0 ||
      index > 16)
    {
      return -1;
    }

  if (index < 8)
    {
      mask <<= index;
      *value = (emcmotIo->OutBytes[0] & mask) ? 1 : 0;
    }
  else {
    if (index < 16)       /* else index is 8-15 */
      {
	mask <<= index - 8;
	*value = (emcmotIo->OutBytes[1] & mask) ? 1 : 0;
      }
    else
      {
	mask <<= index - 16;
	*value = (emcmotIo->OutBytes[2] & mask) ? 1 : 0;
      }
  }
  return 0;
}

/*
  pptDioWrite() maps 8 index values 0-7 onto D0-D7, respectively, and
  4 index values 8-11 onto C0-C3, respectively, providing 12 digital outputs.

  Note that C0, C1, and C3 are inverted outputs.
  */
int pptDioWrite(int index, int value)
{
  unsigned char mask = 1;	/* start at 00000001 for D0, C0 */
#ifdef PPMC_DEBUG
  static int OldOutBytes[3] = {123, 123, 123};
  int i,j;
#endif

  if (index < 0 ||
      index > 16)
    {
      return -1;
    }

  if (index < 8)
    {
      mask <<= index;
      if (value)
	{
	  emcmotIo->OutBytes[0] |= mask;
	}
      else
	{
	  emcmotIo->OutBytes[0] &= ~mask;
	}
    }
  else {
    if (index < 16)
      {
	mask <<= index - 8;
	if (value)
	  {
	    emcmotIo->OutBytes[1] |= mask;
	  }
	else
	  {
	    emcmotIo->OutBytes[1] &= ~mask;
	  }
      }
  else
      {
	mask <<= index - 16;
	if (value)
	  {
	    emcmotIo->OutBytes[2] |= mask;
	  }
	else
	  {
	    emcmotIo->OutBytes[2] &= ~mask;
	  }
      }
  }
#ifdef PPMC_DEBUG
  j = 0;
  for (i=0; i<3;i++)
      if (emcmotIo->OutBytes[i] != OldOutBytes[i]) j=1;
  if (j != 0)
    {
    printf("ppmcparport output change :");
    printf("NumWrite = %d  Out Bytes = ",emcmotIo->NumWrite);
    for (i=0;i<3; i++)
      {
	printf(" %d %d ",i,emcmotIo->OutBytes[i]);
	OldOutBytes[i] = emcmotIo->OutBytes[i];
      }
    printf("\n");
    printf("NumRead = %d  In Bytes = ",emcmotIo->NumRead);
    for (i=0;i<3; i++)
      {
	printf(" %d %d ",i,emcmotIo->InBytes[i]);
      }
    printf("\n");
    }
#endif
  return 0;
}

int pptDioByteRead(int index, unsigned char *byte)
{
  if (index != 0)
    {
      *byte = 0;
      return -1;
    }

  //  *byte = inb(PARPORT_IO_ADDRESS + 1);
  *byte >>= 3;			/* shift down since S3 is bit 0, etc. */
  return 0;
}

int pptDioShortRead(int index, unsigned short *sh)
{
  if (index != 0)
    {
      *sh = 0;
      return -1;
    }

  //  *sh = (unsigned short) inb(PARPORT_IO_ADDRESS + 1);
  *sh >>= 3;			/* shift down since S3 is bit 0, etc. */
  return 0;
}

int pptDioWordRead(int index, unsigned int *word)
{
  if (index != 0)
    {
      *word = 0;
      return -1;
    }

  //  *word = (unsigned int) inb(PARPORT_IO_ADDRESS + 1);
  *word >>= 3;			/* shift down since S3 is bit 0, etc. */
  return 0;
}

int pptDioByteWrite(int index, unsigned char byte)
{
  if (index == 0)
    {
      //      outb(byte, PARPORT_IO_ADDRESS);
      return 0;
    }

  if (index == 1)
    {
      //      outb(byte, PARPORT_IO_ADDRESS + 2);
      return 0;
    }

  return -1;
}

int pptDioShortWrite(int index, unsigned short sh)
{
  return pptDioByteWrite(index, (unsigned char) sh);
}

int pptDioWordWrite(int index, unsigned int word)
{
  return pptDioByteWrite(index, (unsigned char) word);
}

int pptDioByteCheck(int index, unsigned char *byte)
{
  if (index == 0)
    {
      //      *byte = inb(PARPORT_IO_ADDRESS);
      return 0;
    }

  if (index == 1)
    {
      //      *byte = inb(PARPORT_IO_ADDRESS + 2);
      return 0;
    }

  *byte = 0;
  return -1;
}

int pptDioShortCheck(int index, unsigned short *sh)
{
  unsigned char b0 = 0;
  unsigned char b1 = 0;

  if (index != 0)
    {
      *sh = 0;
      return -1;
    }

  //  b0 = inb(PARPORT_IO_ADDRESS);
  //  b1 = inb(PARPORT_IO_ADDRESS + 2);

  *sh = b1;
  *sh <<= 8;
  *sh += b0;

  return 0;
}

int pptDioWordCheck(int index, unsigned int *word)
{
  unsigned char b0 = 0;
  unsigned char b1 = 0;

  if (index != 0)
    {
      *word = 0;
      return -1;
    }

  //  b0 = inb(PARPORT_IO_ADDRESS);
  //  b1 = inb(PARPORT_IO_ADDRESS + 2);

  *word = b1;
  *word <<= 8;
  *word += b0;

  return 0;
}

#ifdef MAIN

static void fillIn(int *ins)
{
  int t;

  for (t = 0; t < 5; t++)
    {
      pptDioRead(t, &ins[t]);
    }
}

static void printIn(int *ins)
{
  int t;

  for (t = 0; t < 5; t++)
    {
      rcs_print("%d ", ins[t]);
    }
  rcs_print("\n");
}

static void fillOut(int *outs)
{
  int t;

  for (t = 0; t < 12; t++)
    {
      pptDioCheck(t, &outs[t]);
    }
}

static void printOut(int *outs)
{
  int t;

  for (t = 0; t < 12; t++)
    {
      rcs_print("%d ", outs[t]);
    }
  rcs_print("\n");
}

int main(int argc, char *argv[])
{
  int t;
  char input[LINELEN];
  char cmd[LINELEN];
  int ins[5];
  int outs[12];
  int index;



  if (-1 == pptDioInit(0))
    {
      rcs_print_error( "can't init\n");
      exit(1);
    }

  while (! feof(stdin))
    {
      rcs_print("ppt> ");
      fflush(stdout);

      if (NULL == fgets(input, LINELEN, stdin))
	{
	  /* EOF */
	  break;
	}

      if (1 != sscanf(input, "%s", cmd))
	{
	  /* blank line-- print input */
	  fillIn(ins);
	  printIn(ins);
	  continue;
	}

      if (! strcmp(cmd, "help") ||
	  ! strcmp(cmd, "?"))
	{
	  rcs_print("help/?           print this help\n");
	  rcs_print("s <#>            set bit #\n");
	  rcs_print("c <#>            clear bit #\n");
	  rcs_print("k                check outputs\n");
	  rcs_print("ENTER            print inputs\n");
	  rcs_print("q                quit\n");
	}
      else if (! strcmp(cmd, "s"))
	{
	  /* set */
	  if (1 != sscanf(input, "%*s %d", &index))
	    {
	      rcs_print("bad input: %s\n", input);
	      continue;
	    }

	  if (0 != pptDioWrite(index, 1))
	    {
	      rcs_print("index out of range\n");
	    }
	}
      else if (! strcmp(cmd, "c"))
	{
	  /* set */
	  if (1 != sscanf(input, "%*s %d", &index))
	    {
	      rcs_print("bad input: %s\n", input);
	      continue;
	    }

	  if (0 != pptDioWrite(index, 0))
	    {
	      rcs_print("index out of range\n");
	    }
	}
      else if (! strcmp(cmd, "k"))
	{
	  /* check outputs */
	  fillOut(outs);
	  printOut(outs);
	}
      else if (! strcmp(cmd, "q"))
	{
	  break;
	}
      else
	{
	  rcs_print("?\n");
	}
    }

  pptDioQuit();

  exit(0);
}

#endif
