/********************************************************************
*
* Description: sensoray526.c
*   Driver for the Sensoray 526 PC104 card.
*
* Author: Paul Corner
* License: GPL Version 2
* System: Linux
*    
* Copyright (c) 2005 All rights reserved.
*
* Last change: 
* $Revision: 1.2 $
* $Author: paul $
* $Date: 2005/08/08 22:56:38 $
********************************************************************/
#include "config.h"

#ifdef realtime
#ifdef __KERNEL__
#include <linux/kernel.h>
#endif

#include <asm/io.h>

#else
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <stdlib.h>
#include <linux/types.h>
#include <ctype.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/io.h>

#endif

/* Use the standard error numbers to return a little more info
   to the calling process. Three possible errors to report.

   -EFAULT  Used for index or axis out of range
   -ENOSYS  Function not implemented
   -EIO     General IO error
*/
#include <linux/errno.h>

#include "emcmotcfg.h"
#include "diagnostics.h"
#include "sensoray526.h"

/* ident tag */
#ifndef __GNUC__
#ifndef __attribute__
#define __attribute__(x)
#endif
#endif

#ifdef realtime
MODULE_VERSION("$Id: sensoray526.c,v 1.2 2005/08/08 22:56:38 paul Exp $");
#endif

static char __attribute__((unused)) ident[] = "$Id: sensoray526.c,v 1.2 2005/08/08 22:56:38 paul Exp $";

int IO_BASE_ADDRESS = DEFAULT_IO_BASE_ADDRESS;
char buf[LINELEN];

char *s526Model(void)
{
    memset(buf, 0,LINELEN);	// Clear the buffer for the return string.
    snprintf(buf, 63, "Sensoray 526 PC104 card (rev. %x) found at 0x%x", 
		inw(ECC) & 0x000f, IO_BASE_ADDRESS);
    return buf;
}

int s526MotInit(const char * stuff)
{
    int tmp = 0;
    tmp = inw(ECC);
    if ((tmp & 0xfff0) != 0x5260) {
	diagnostics("Sensoray 526 card not found at 0x%x\n", IO_BASE_ADDRESS);
	return -ENXIO;
    } else {
	diagnostics("%s\n", s526Model());
    }
    outw(0x0480, C0M);  // Set the counter to x4 quadrature counting.
    outw(0x0480, C1M);
    outw(0x0480, C2M);
    outw(0x0480, C3M);
    outw(0x8007, C0C);
    outw(0x8007, C1C);
    outw(0x8007, C2C);
    outw(0x8007, C3C);
    outw(0x0001, MSC);
    diagnostics("Completed MotInit\n");
    return 0;
}
int s526MotQuit(void)
{
    outw(0x0000, C0M);
    outw(0x0000, C1M);
    outw(0x0000, C2M);
    outw(0x0000, C3M);
    outw(0x0000, MSC);
    diagnostics("Completed MotQuit\n");
    return 0;
}

int s526DioInit(const char * stuff)
{
    outw(DIO_CONFIG, DIO);
    return 0;
}

int s526DioQuit(void)
{
    outw(0x0000, DIO);
    return 0;
}

int s526AioInit(const char * stuff)
{
    outw(0x0008, DAC); // Reset the DACs.
    outw(0x7fff, ADD); // Write a zero so that the ISR has a reason
    outw(0x0001, DAC); // set the DAC done flag.
     /* Enable the watchdog timer and activate the relay.
        Set the little puppy for a 125mSec timeout */
    outw(0x000f, WDC);
    outw(0x8001, ADC); // Trigger an ADC conversion.
    return 0;
}

int s526AioQuit(void)
{
    outw(0x0008, DAC);	// Reset the DACs.
    return 0;
}

int s526DacNum(void)
{
    return S526_MAX_AXIS;
}

int s526DacWrite(int axis, double volts)
{

    int count = 0;
    /* DAC output is in the range:
       0x0000 = -10V
       0x7fff = 0V
       0xffff = +10v
    */
    // Clamp the voltage if out of range.
    if (volts > 10){volts = 10;}
    if (volts < -10){volts = -10;}
    // Now scale it to suit the DAC
    count = (int) (volts * 3276.75 + 32767.5);

    // Wait for the last DAC write to complete..
    do {} while((inw(ISR) & 0x0002) == 0);

    switch(axis){
	case 0:
	    outw(0x0000, DAC); // Set the DAC channel
	    outw(count, ADD);  // Write the data
	    outw(0x0001, DAC); // trigger the conversion 
	    break;
	case 1:
	    outw(0x0002, DAC);
	    outw(count, ADD);
	    outw(0x0003, DAC); // Need to keep the correct channel.
	    break;
	case 2:
	    outw(0x0004, DAC);
	    outw(count, ADD);
	    outw(0x0005, DAC);
	    break;
	case 3:
	    outw(0x0006, DAC);
	    outw(count, ADD);
	    outw(0x0007, DAC);
	    break;
	default:
	return -EFAULT;
    }
    inw(WDC); // Give watchdog timer a kick.
    return 0;
}

int s526DacWriteAll(int max, double * volts)
{
    int t;

    if ((max > EMCMOT_MAX_AXIS) || (max < 0))
	return -EFAULT;

    for (t = 0; t < S526_MAX_AXIS; t++){
	s526DacWrite(t, volts[t]);
    }
    return 0;
}

int s526EncoderSetIndexModel(unsigned int model)
{
    return -ENOSYS;
}

int s526EncoderNum(void)
{
    return S526_MAX_AXIS;
}

int s526DioMaxInputs(void)
{
    return S526_MAX_DIO_INPUTS;
}


int s526DioMaxOutputs(void)
{
    return S526_MAX_DIO_OUTPUTS;
}

int s526EncoderRead(int axis, double * counts)
{
    int status = 0;
    union pos_tag {
	__s32 data;
	struct word_tag {__s16 lo; __s16 hi;} word;
    } pos;

    pos.data = (__s32) *counts;
    pos.word.hi &= 0xff00; 

    switch(axis){
	case 0:
	    status = inw(C0C);
	    /* What happens if the count rolls over between reading
	       the status flags and latching the count ?
	       A large jump in readings that hopefully gets trapped
	       by the servo loop...*/
	    pos.word.lo = inw(C0L);
	    pos.word.hi |= inw(C0H);
	    if (status & 0x0008){	// Check for counter over-flow
		outw(0x0008, C0C);	// Clear the OF flag.
	    }
	    break;
	case 1:
	    status = inw(C1C);
	    pos.word.lo = inw(C1L);
	    pos.word.hi |= inw(C1H);
	    if (status & 0x0008){
		outw(0x0008, C1C);
	    }
	    break;
	case 2:
	    status = inw(C2C);
	    pos.word.lo = inw(C2L);
	    pos.word.hi |= inw(C2H);
	    if (status & 0x0008){
		outw(0x0008, C2C);
	    }
	    break;
	case 3:
	    status = inw(C3C);
	    pos.word.lo = inw(C3L);
	    pos.word.hi |= inw(C3H);
	    if (status & 0x0008){
		outw(0x0008, C3C);
	    }
	    break;
	default:
	return -EFAULT;
    }

    if ((status & 0x0008)){
	if (pos.word.hi & 0x0080){	// Check the sign of the 24bit count..
	    pos.word.hi = pos.word.hi - 0x0100;
	} else {
	    pos.word.hi = pos.word.hi + 0x0100;
	}
    }

    *counts = (double) pos.data;
    return 0;
}

int s526EncoderReadAll(int max, double * counts)
{
    int t;
/* FIXME - emcmot needs to read MAX_AXIS from the driver and use that number when calling Read/WriteAll functions */
    if ((max > EMCMOT_MAX_AXIS) || (max < 0))
	return -EFAULT;
    for (t = 0; t < S526_MAX_AXIS; t++){
	s526EncoderRead(t, &counts[t]);
    }
    return 0;
}


int s526EncoderResetIndex(int axis)
{
    /* Clear ICAP+ and ICAP- */
    switch(axis){
	case 0:
	    outw(0x0006, C0C);
	    break;
	case 1:
	    outw(0x0006, C1C);
	    break;
	case 2:
	    outw(0x0006, C2C);
	    break;
	case 3:
	    outw(0x0006, C3C);
	    break;
	default:
	return -EFAULT;
    }
    return 0;
}

int s526EncoderReadLatch(int axis, int * flag)
{
    int status = 0;

    switch(axis){
	case 0:
	    status = inw(C0C);
	    break;
	case 1:
	    status = inw(C1C);
	    break;
	case 2:
	    status = inw(C2C);
	    break;
	case 3:
	    status = inw(C3C);
	    break;
	default:
	return -EFAULT;
    }
    /* Use ICAP- (the falling edge of the index pulse. */
    *flag = (status & 0x0002) ? 1:0;
    return 0;
}

int s526EncoderReadLevel(int axis, int * flag)
{
    int status = 0;

    switch(axis){
	case 0:
	    status = inw(C0C);
	    break;
	case 1:
	    status = inw(C1C);
	    break;
	case 2:
	    status = inw(C2C);
	    break;
	case 3:
	    status = inw(C3C);
	    break;
	default:
	return -EFAULT;
    }
    /* AND bit 5 to get the current status of the index */
    *flag = (status & 0x0020) ? 1:0;
    return 0;
}

int s526MinLimitSwitchRead(int axis, int * flag)
{
    if ((axis > S526_MAX_AXIS) || (axis < 0))
	return -EFAULT;
    return s526DioRead(S526_MIN_LIMIT_INDEX, flag);
}

int s526MaxLimitSwitchRead(int axis, int * flag)
{
    if ((axis > S526_MAX_AXIS) || (axis < 0))
	return -EFAULT;
    return s526DioRead(S526_MAX_LIMIT_INDEX, flag);
}

int s526HomeSwitchRead(int axis, int * flag)
{
    if ((axis > S526_MAX_AXIS) || (axis < 0))
	return -EFAULT;
    s526EncoderResetIndex(axis);
    return s526DioRead(S526_HOME_INDEX, flag);
}

int s526AmpEnable(int axis, int enable)
{
    if ((axis > S526_MAX_AXIS) || (axis < 0))
	return -EFAULT;
    return s526DioWrite(S526_AMP_ENABLE_INDEX, enable);

}

int s526AmpFault(int axis, int * fault)
{
    if ((axis > S526_MAX_AXIS) || (axis < 0))
	return -EFAULT;
    return s526DioRead(S526_AMP_FAULT_INDEX, fault);

}

int s526DioRead(int index, int *value)
{
    int status = 0;

    if ((index > s526DioMaxInputs()) || (index < 0))
	return -EFAULT;

    /* Assume inputs start at INPUT_BASE */
    status = inw(DIO);
    *value = (status >> (index + INPUT_BASE)) & 0x00001;
    return 0;
}

int s526DioWrite(int index, int value)
{
    int status = 0;
    int bit = 0;

    if ((index > s526DioMaxOutputs()) || (index < 0))
	return -EFAULT;

    status = inw(DIO) & 0x00ff;
    bit = value ? 1:0;
    /* shift the bit to the correct location
       Assume outputs start at OUTPUT_BASE */
    bit = bit << (index + OUTPUT_BASE);
    status = status & (~bit); // Clear just the indexed bit

    /* Set the output bit and the IO config bits */
    status = status | bit | DIO_CONFIG;
    outw(status, DIO);

    return 0;
}

int s526DioCheck(int index, int *value)
{
    return s526DioRead(index, value);
}

/* Stub the Analogue input functions until they are needed.
   At present, EMC does not require any analogue inputs, but
   servo amps like the Copley have Voltage and Current monitor
   outputs that could be used in tuning & monitoring.
   Need some code in EMC to handle this, and maybe some GUI
   widgets to display the data..*/
int s526AioMaxInputs(void){return 8;}
int s526AioMaxOutputs(void){return 4;}
int s526AioStart(int index)
{
    __u16 control;

    if ((index > s526AioMaxOutputs()) || (index < 0))
	return -EFAULT;
    control = (0x002 << index) | 0x8001;

    outw(control, ADC);		// Trigger conversion
    return 0;
}

int s526AioReady(void){

    if ((inw(ISR) & 0x0004) == 1){
        return -EBUSY;
    }
    return 0;
}

int s526AioRead(int index, double *volts)
{
    __s16 data = 0;
    __u16 control;

    if ((index > s526AioMaxOutputs()) || (index < 0))
	return -EFAULT;

    /* This routine should be using signed ints rather than
       floats - At some point, this func. will be called in
       normal kernel space when the FP flags have not been
       saved.
    */
    return -EIO;	// Exit NOW.

    control = (0x002 << index) | 0x8001;

    outw(control, ADC);		// Trigger conversion
    do {} while((inw(ISR) & 0x0004) == 0); // wait ~25uSec
    data = inw(ADD);		// Read ADC register

    if(data > 0){
	 *volts = (float) (data+1) / 3276.8;
    }else{
	*volts = (float) data / 3276.8;
    }

    return 0;

}
int s526AioWrite(int index, double volts){return -ENOSYS;}
int s526AioCheck(int index, double *volts){return -ENOSYS;}
