/********************************************************************
*
* Description: emcioctl.c
*
*	devfs interface that will in time provide a sane and safe
*	method for passing data to and from the realtime module.
*
*	One of the first functions to be implemented will be a
*	call to start/stop the RT threads - This will provide a 
*	robust shutdown proceedure.
*
*    
* Author: Paul Corner
* License: GPL Version 2
* System: Linux
*    
* Last change: 
* $Revision: 1.2 $
* $Author: paul $
* $Date: 2005/08/12 11:37:42 $
*
* Copyright (c) 2004  All rights reserved.
*
********************************************************************/
#ifndef __KERNEL__
#define __KERNEL__
#endif
#ifndef MODULE
#define MODULE
#endif

#include <linux/module.h>
#include <linux/fs.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/fcntl.h>
#include <asm/uaccess.h>
#include "extintf.h"
#include "emcioctl.h"
#include "diagnostics.h"

/* ident tag */
#ifndef __GNUC__
#ifndef __attribute__
#define __attribute__(x)
#endif
#endif

static char __attribute__ ((unused)) ident[] =
    "$Id: emcioctl.c,v 1.2 2005/08/12 11:37:42 paul Exp $";

int emc_ioctl(struct inode *inode, struct file *filp,
	      unsigned int cmd, unsigned long arg)
{
    int index = -1;
    int volt = 56;
    int ret = 0;

    if (_IOC_TYPE(cmd) != EMC_IOC_MAGIC)
	return -ENOTTY;
    if (_IOC_DIR(cmd) & _IOC_READ) {
	ret = !access_ok(VERIFY_WRITE, (void *) arg, _IOC_SIZE(cmd));
    } else if (_IOC_DIR(cmd) & _IOC_WRITE) {
	ret = !access_ok(VERIFY_READ, (void *) arg, _IOC_SIZE(cmd));
    }
    if (ret)
	return -EFAULT;
    ret = 0;


    switch (cmd) {
/* NOTE: need to add code to index, bit, and volt to/from usr space.
    At the moment, we are just fleshing out the code to show how
    data can be exchanged.
*/
    case EMC_IOW_DIO_READ:
	get_user(index, (int *) arg);
	diagnostics("extDioRead %x\n", index);
	index = index - arg;
	put_user(index, (int *) arg);
	break;
    case EMC_IOW_DIO_SET:
	get_user(index, (int *) arg);
	diagnostics("extDioWrite %x\n", index);
	break;
    case EMC_IOW_DIO_CLEAR:
	get_user(index, (int *) arg);
	diagnostics("extDioWrite %x\n", index);
	break;
    case EMC_IOW_AIO_READ:
	/* Analogue data is going to have to be passed as signed ints,
	   not because of any restriction on data types, but because
	   the values are subject to further math in kernel space.
	   The FP flags are not saved when entering the kernel via
	   ioctl calls so NO floating point math can be done. */
	ret = get_user(index, (int *) arg);
	if (ret < 0)
	    break;
	//ret = extAioRead(index, &volt);
	if (ret < 0)
	    break;
	ret = put_user(volt, (int *) arg);
	break;
    case EMC_IOW_AIO_WRITE:
	ret = get_user(index, (int *) arg);
	if (ret < 0)
	    break;
	__get_user(volt, (int *) arg + 1);
	diagnostics("extAioWrite %x %x\n", index, volt);
	break;
    case EMC_IOW_AIO_READY:
	ret = extAioReady();
	break;
    case EMC_IOR_MODEL:
	ret = copy_to_user((char *) arg, extModel(), strlen(extModel()));
	break;

    default:
	return -ENOTTY;
    }
    return ret;

}


int emc_open(struct inode *inode, struct file *filp)
{
    int num = MINOR(inode->i_rdev);

    diagnostics("check the device number\n");
    if (filp->f_mode & FMODE_READ)
	diagnostics("Open for read %i\n", num);
    if (filp->f_mode & FMODE_WRITE)
	diagnostics("Open for write %i\n", num);
    return 0;
}

int emc_release(struct inode *inode, struct file *filp)
{
    diagnostics("Closing device.\n");
    return 0;
}

struct file_operations emc_fops = {
  open:emc_open,
  release:emc_release,
  ioctl:emc_ioctl,
};

int emc_major = 0;

int dev_init(void)
{
    int t = 0;
    t = register_chrdev(emc_major, "emc", &emc_fops);
    if (t < 0) {
	diagnostics("Failed to register /dev/emc %d\n", emc_major);
	return t;
    }
    if (emc_major == 0)
	emc_major = t;
    diagnostics("Registered /dev/emc with major %d\n", emc_major);
    return t;
}


void dev_clean(void)
{
    unregister_chrdev(emc_major, "emc");
    return;
}
