/********************************************************************
*
* Description: emcprocfs.c
*
*	Implimentation for the virtual proc file system.
*       Note: This runs in a kernel context so floats and
*       doubles can not be printed. It is not a good idea
*       to do any floating point math either as the FPU
*       state is not saved.
*
*    
* Author: Paul Corner
* License: GPL Version 2
* System: Linux
*    
* Last change: 
* $Revision: 1.3 $
* $Author: paul $
* $Date: 2005/09/08 02:14:20 $
*
* Copyright (c)  All rights reserved.
*
********************************************************************/

#ifndef __KERNEL__
#define __KERNEL__
#endif

#ifndef MODULE
#define MODULE
#endif
#ifndef EXPORT_SYMTAB
#define EXPORT_SYMTAB
#endif

#include <linux/ctype.h>
#include <asm/uaccess.h>

#include "diagnostics.h"
#include "emcmot.h"
#include "emcprocfs.h"		/* macros for read functions */

/* ident tag */
#ifndef __GNUC__
#ifndef __attribute__
#define __attribute__(x)
#endif
#endif
static char __attribute__ ((unused)) ident[] =
    "$Id: emcprocfs.c,v 1.3 2005/09/08 02:14:20 paul Exp $";

#ifdef CONFIG_PROC_FS

static struct proc_dir_entry *emc_dir = 0;	/* /proc/emc directory */
static struct proc_dir_entry *status_file = 0;	/* /proc/emc/status */
static struct proc_dir_entry *debug_file = 0;	/* /proc/emc/debug */

/** The following are callback functions for the /proc filesystem
    When someone reads a /proc file, the appropriate function below
    is called, and it must generate output for the reader on the fly.
*/
extern char loByte;
extern int IO_BASE_ADDRESS;

static int proc_read_status(char *page, char **start, off_t off,
			    int count, int *eof, void *data)
{
    int axis = 0;

    PROC_PRINT_VARS;
    PROC_PRINT("******* EMC STATUS ********\n\n");
    PROC_PRINT("Axis\t Status \t      Error \t InPos\t  Raw Count \t Raw Output\n");
    for (axis = 0; axis < EMCMOT_MAX_AXIS; axis++) {
	PROC_PRINT(" %i\t%8s\t%8s\t%8s%11i%11i\n", axis,
		   (GET_AXIS_ENABLE_FLAG(axis) ? "ENABLED" : "DISABLED"),
		   (GET_AXIS_ERROR_FLAG(axis) ? "ERROR" : "OK"),
		   (GET_AXIS_INPOS_FLAG(axis) ? "Yes" : "No"),
		   (int) emcmotDebug->rawInput[axis],
		   (int) emcmotDebug->rawOutput[axis]);
    }
    PROC_PRINT("\n");
    PROC_PRINT("Motion\n Status       Error  InPos\n");
	PROC_PRINT(" %8s%8s%8s\n",
            (GET_MOTION_ENABLE_FLAG() ? "ENABLED" : "DISABLED"),
            (GET_MOTION_ERROR_FLAG() ? "ERROR" : "OK"),
            (GET_MOTION_INPOS_FLAG() ? "Yes" : "No") );
    
    PROC_PRINT("\n");
    PROC_PRINT("\nHeartbeat = %i\n", emcmotStatus->heartbeat);
    PROC_PRINT("\n");
    PROC_PRINT_DONE;
}

/* can set these anywhere, for compile,test... debugging cycles */
int extdbg1,extdbg2,extdbg3,extdbg4;

static int proc_read_debug(char *page, char **start, off_t off,
			   int count, int *eof, void *data)
{
    int tc_queue = (DEFAULT_TC_QUEUE_SIZE + 10);
    PROC_PRINT_VARS;
    PROC_PRINT("******* EMC DEBUG ******\n\n");
    PROC_PRINT("Debug level: 0x%x\n", emcmotConfig->debug);
    PROC_PRINT("IO_BASE_ADDRESS: 0x%x\n\n", IO_BASE_ADDRESS);
    PROC_PRINT("\n");
    PROC_PRINT("sizeof(EMCMOT_COMMAND)    %i\n",  sizeof(EMCMOT_COMMAND));
    PROC_PRINT("sizeof(EMCMOT_STATUS)     %i\n", sizeof(EMCMOT_STATUS));
    PROC_PRINT("sizeof(EMCMOT_CONFIG)     %i\n", sizeof(EMCMOT_CONFIG));
    PROC_PRINT("sizeof(EMCMOT_DEBUG)      %i\n", sizeof(EMCMOT_DEBUG));
    PROC_PRINT("sizeof(EMCMOT_ERROR)      %i\n", sizeof(EMCMOT_ERROR));
    PROC_PRINT("sizeof(EMCMOT_LOG)        %i\n", sizeof(EMCMOT_LOG));
    PROC_PRINT("sizeof(EMCMOT_IO)         %i\n", sizeof(EMCMOT_IO));
    PROC_PRINT("sizeof(EMCMOT_COMP)       %i\n",  sizeof(EMCMOT_COMP) * EMCMOT_MAX_AXIS);
    PROC_PRINT("    Total                ----------\n");
    PROC_PRINT("sizeof(EMCMOT_STRUCT)     %i\n", sizeof(EMCMOT_STRUCT));
    PROC_PRINT("Space used by the assorted queues (part of EMCMOT_DEBUG)\n");
    PROC_PRINT("freeAxis[]        %i\n", sizeof(TP_STRUCT) * EMCMOT_MAX_AXIS);
    PROC_PRINT("cubic[]           %i\n", sizeof(CUBIC_STRUCT) * EMCMOT_MAX_AXIS);
    PROC_PRINT("queueTcSpace[]    %i\n", sizeof(TC_STRUCT) * tc_queue);
    PROC_PRINT("freeAxisTcSpace[] %i\n", sizeof(TC_STRUCT) * EMCMOT_MAX_AXIS * FREE_AXIS_QUEUE_SIZE);
    PROC_PRINT("\n");
    PROC_PRINT("Current TC queue usage: %i out of %i\n", tpQueueDepth(&emcmotDebug->queue), tc_queue);
    PROC_PRINT("\n");
    PROC_PRINT("Size of some of the other structs.\n");
    PROC_PRINT("sizeof(TP_STRUCT) %i\n", sizeof(TP_STRUCT));
    PROC_PRINT("sizeof(PID_STRUCT) %i\n", sizeof(PID_STRUCT));
    PROC_PRINT("sizeof(TC_STRUCT) %i\n", sizeof(TC_STRUCT));
    PROC_PRINT("\n");
    PROC_PRINT("General Purpose extdbg: %i %i %i %i\n", extdbg1,extdbg2,extdbg3,extdbg4);
    PROC_PRINT("\n");


    PROC_PRINT_DONE;
}

static int proc_write_debug(struct file *file,
			    const char *buffer, unsigned long count,
			    void *data)
{
    char c;

    /* copy 1 byte from user space */
    if (copy_from_user(&c, buffer, 1)) {
	return -1;
    }
    /* check it is a digit */
    if (isdigit(c)) {
	diagnostics("EMCMOT: rt-timer stopped.\n");
	stop_rt_timer();
    }
    /* tell whoever called us that we used all the data, even though we
       really only used the first byte */
    return count;
}

/** proc_init() initializes the /proc filesystem entries,
    creating the directory and files, and linking them
    to the appropriate callback functions. This function
    is called from the init_module() function of the
    RTAPI implementation.
*/

int proc_init(void)
{
    /* create the emc directory "/proc/emc" */
    emc_dir = create_proc_entry("emc", S_IFDIR, NULL);
    if (emc_dir == 0) {
	return -1;
    }

    /* create read only file "/proc/emc/status" */
    status_file = create_proc_entry("status", S_IRUGO, emc_dir);
    if (status_file == NULL) {
	return -1;
    }
    status_file->read_proc = proc_read_status;

    /* create read/write file "/proc/emc/debug" */
    debug_file = create_proc_entry("debug", S_IRUGO | S_IWUGO, emc_dir);
    if (debug_file == NULL) {
	return -1;
    }
    debug_file->data = NULL;
    debug_file->read_proc = proc_read_debug;
    debug_file->write_proc = proc_write_debug;
    return 0;
}

/** proc_clean() is called from the cleanup_module function of
    of the RTAPI implementation.  It removes the emc entries
    from the /proc filesystem.  Failing to remove a /proc
    entry before the module is removed may cause kernel panics.
*/

void proc_clean(void)
{
    /* remove /proc entries, only if they exist */
    if (emc_dir != NULL) {
	if (status_file != NULL) {
	    remove_proc_entry("status", emc_dir);
	    status_file = NULL;
	}
	if (debug_file != NULL) {
	    remove_proc_entry("debug", emc_dir);
	    debug_file = NULL;
	}
	remove_proc_entry("emc", NULL);
    }
}

#endif				/* CONFIG_PROC_FS */
