/********************************************************************
*
* Description: emcstepper.c
*
*	The freqfunc() and task creation routines split out from
*	emcmot.c - Concentrate on just one algorithm and get it right.
*
*	TODO Clean up the mess of the outer if(foo) loop and impliment
*	a viable user configurable method for step/dir definitions.
*
	Some confusion can arise when reading the documentation, ini files,
	and the source code. The term "pin" in the docs refer to the physical
	pin on the DB25 parallel port connector. When reading an ini file, the
	term INDEX relates to the individual bits of each port - Bits 0-7 of
	the first byte (base_addr) are INDEX 0-7, and bits 0-3 of the second
	output byte (base_addr +2) on the parallel port are INDEX 8-11.
	To confuse thins further, the inputs from the status port (base_addr +1),
	bits 3-7 are refered to as INDEX 0-5 in the ini file...

	The docs are outside the remit of the source code, and to change the
	INDEX to fit the needs of a byte wide parallel port would be time
	consuming. The INDEX model works reasonably well for a limited number
	of bits and ports.. The difficulties to be concerned with here is how
	to map an INDEX to a Step or Direction bit and translate it to a
	physical output.


FIXME - Re-word and impliment some sample code.

	Suggestion: An array is used to hold the definitions and state of each
	pin. So the byte assembly code in freqfunc would no longer do any bit
	shifts, instead, INDEX[axis.0.step]= 0 ? 1
	Then extDioByteWrite would be responsible for converting INDEX to bytes.
*
*    
* Author:
* License: GPL Ver. 2
* System: Linux
*    
* Last change: 
* $Revision: 1.4 $
* $Author: paul $
* $Date: 2005/09/17 15:27:30 $
*
* Copyright (c) 2004  All rights reserved.
*
*History
* 24Feb06   Matt Timmermans
*	    added realtime_* calls to set up timer stuff for emcmodule.c
* 19Mar05   lpg added module param for stepping type
*           0 == step/dir, 2 == quad drive, 3 = 4 phase
*               added code for quadrature drive: quadfunc()
*               and raw 4 phase l/r systems:     lr4func()
* 21Mar05   lpg added some I/O cleanup and GPL notice
********************************************************************/


#ifndef __KERNEL__
#define __KERNEL__
#endif
//#undef _RTAI_SHM_H_

#ifndef MODULE
#define MODULE
#endif
#ifndef EXPORT_SYMTAB
#define EXPORT_SYMTAB
#endif

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/version.h>

#include "rtai.h"
#include "rtai_sched.h"
#include "rtai_shm.h"

#include "emcmotcfg.h"
#include "emcmotglb.h"
#include "emcmot.h"
#include "extintf.h"

#include "diagnostics.h"

#define FREQ_AXES 6		/* how many axes are supported */
#define DEFAULT_FREQ_TASK_PRIORITY 1
#define DEFAULT_FREQ_TASK_STACK_SIZE 4096
#define DEFAULT_STEPPING_TYPE 0	/* step/dir */

int FIND_IO_BASE_ADDRESS;
extern int DEBUG_EMCMOT;
extern int IO_BASE_ADDRESS;
extern int module_error_code;

static RTIME rtaiTickPeriod;
static long rtaiTickPeriod_long;
static int ticksPerPeriod = 48;

int PERIOD = 48000;
module_param(PERIOD, int, S_IRUGO);
MODULE_PARM_DESC(PERIOD, "Base period for the clock");

static int FREQ_TASK_PRIORITY = DEFAULT_FREQ_TASK_PRIORITY;
module_param(FREQ_TASK_PRIORITY, int, S_IRUGO);
MODULE_PARM_DESC(FREQ_TASK_PRIORITY, " No description");

static int STEPPING_TYPE = DEFAULT_STEPPING_TYPE;
module_param(STEPPING_TYPE, int, S_IRUGO);
MODULE_PARM_DESC(STEPPING_TYPE, " 0=step/dir, 2=quadrature, 3=4 phase");

static int FREQ_TASK_STACK_SIZE = DEFAULT_FREQ_TASK_STACK_SIZE;
module_param(FREQ_TASK_STACK_SIZE, int, S_IRUGO);
MODULE_PARM_DESC(FREQ_TASK_STACK_SIZE, " No description");
MODULE_VERSION(" $Id: emcstepper.c,v 1.4 2005/09/17 15:27:30 paul Exp $");


static RT_TASK freqTask;
static int freqTask_created = 0;


/*
  smController runs the stepper motor pulse task calculations at a
  relatively high frequency. The time interval between calls to this
  task is governed by the maximum speed ([TRAJ] MAX_VELOCITY) and the
  max pulses-per-unit among all motors ([AXIS_#] INPUT_SCALE). The task
  runs at twice this time since it must issue both down pulses and up
  pulses, and can only do one per motor per call.

  For each motor, the task looks at the commanded position, in integer
  counts, as stored in the global smSteps[]. This is set by the
  emcmotController task. If the accumulated number of steps is greater
  (or less) than half a pulse, a pulse is sent in the proper direction and
  the accumulated count is incremented (or decremented).

  The timing of this task does not vary with the speed of each motor.

  The parallel port is used for outputs, using the extDioByteWrite()
  function. Four axes (step and direction) are put on the low byte, and
  the remaining two axes are put on the high byte.
*/

/*
  freqfunc() takes 'pdmult' period multiplier global and uses two
  counters to time the up and down portions of the duty cycle.

  Writes global int array 'emcmotDebug->stepperCount', defined in extsmmot.c or
  other interface dispatchers, which is used as the "encoder feedback"
  for open-loop emcmotDebug->stepping. (stepping here does not refer to single stepping through the motion queue !
*/

char loByte = 0;
char hiByte = 0;

extern int setup_time[];
extern int hold_time[];

/*
*  freqfunc()
*  This code gets run if you install module with STEPPING_TYPE = 0 (default)
*/
void freqfunc(long i)
{
    int axis = 0;
    int pdmag = 0;
    /* these need not be static since this function never goes out of scope */
    int oldpdmult[] = {[0 ... FREQ_AXES] = 0 };	// Creates an array of FREQ_AXES+1
    char direction[] = {[0 ... FREQ_AXES] = 0 };	// but allows all elements to be
    int upmag[] = {[0 ... FREQ_AXES] = 0 };	// initialised.
    int uptime[] = {[0 ... FREQ_AXES] = 0 };
    int upcount[] = {[0 ... FREQ_AXES] = 100 };
    int downcount[] = {[0 ... FREQ_AXES] = 0 };
#if 0
    int downmag[] = {[0 ... FREQ_AXES] = 0 };
    int reversal[] = {[0 ... FREQ_AXES] = 0 };
#endif
    static char oldLoByte = 0, oldHiByte = 0;

    diagnostics("freqFunc started.\n");

    while (1) {
#if 0
	/* This is the original freqfunc before setup & hold timings were added. */
	for (axis = 0; axis < FREQ_AXES; axis++) {
	    if (emcmotDebug->enable[axis]) {
		/* lpg - only axis that need correction will have enable[] == true */
		reversal[axis] = 0;	/* clear flag, it's set later */
		/* set mag and direction of period multiplier, if changed */
		if (emcmotDebug->pdmult[axis] != oldpdmult[axis]) {
		    pdmag = abs(emcmotDebug->pdmult[axis]);
		    if (emcmotDebug->pdmult[axis] == 0) {
			pdmag = 0;
			/* leave direction[] as it was */
		    } else if (emcmotDebug->pdmult[axis] < 0) {
			if (direction[axis] == 0) {
			    reversal[axis] = 1;
			}
			direction[axis] = 1;
		    } else {
			if (direction[axis] != 0) {
			    reversal[axis] = 1;
			}
			direction[axis] = 0;
		    }

		    /* set load values of up and down counters */
		    if (pdmag > 1) {
			downmag[axis] = pdmag / 2;
			upmag[axis] = pdmag - downmag[axis];
		    } else {
			upmag[axis] = downmag[axis] = 0;
		    }

		    /* knock down counters if they're above the new ceiling */
		    if (upcount[axis] > upmag[axis]) {
			upcount[axis] = upmag[axis];
		    }
		    if (downcount[axis] > downmag[axis]) {
			downcount[axis] = downmag[axis];
		    }

/* NOTE: Xylotex and other drivers based on the Allegro A3977 chip
   require the Direction signal to be stable 200nSec before the up
   transition of the Step signal. The Dir also needs to be held for
   200nSec (min) after the up transition of the Step..

   With this in mind...
    Set the Dir bits and write to the port.
    Set the Step bits and do a second write.
    (Save the state...)
   Each IO write will take 1 to 2 uSec and thus satisfy the timing
   constraints of the Allegro chip.

   NOTE: A direction change can only occur on or before an up transition. Bear
   this in mind when simulating the feedback !
*/
		    if (reversal[axis] && (--uptime[axis] < 0)) {
			/* set direction bit */
			if (direction[axis]) {
			    if (axis < 4) {
				loByte |= (0x01 << (axis << 1));
			    } else {
				hiByte |= (0x01 << ((axis - 4) << 1));
			    }
			} else {
			    if (axis < 4) {
				loByte &= ~(0x01 << (axis << 1));
			    } else {
				hiByte &= ~(0x01 << ((axis - 4) << 1));
			    }
			}

			oldpdmult[axis] = emcmotDebug->pdmult[axis];
			uptime[axis] = hold_time[axis];
		    }
		}

		/* end of if (emcmotDebug->pdmult != oldpdmult) */
		/* if we're not reversing direction, decrement counters
		   and run transition logic */
		if (!reversal[axis]) {
		    if (upcount[axis] > 0) {
			/* keep delaying */
			--upcount[axis];
		    } else if (upcount[axis] == 0) {
			/* timer expired-- do the down transition */
			if (axis < 4) {
			    loByte &= ~(0x02 << (axis << 1));
			} else {
			    hiByte &= ~(0x02 << ((axis - 4) << 1));
			}
			/* decr upcount so we don't come back here */
			--upcount[axis];
		    } else if (downcount[axis] > 0) {
			/* keep delaying */
			--downcount[axis];
		    } else if (downcount[axis] == 0) {
			/* timer expired-- do the up transition */
			if (axis < 4) {
			    loByte |= (0x02 << (axis << 1));
			} else {
			    hiByte |= (0x02 << ((axis - 4) << 1));
			}
			/* accumulate position */
			if (direction[axis]) {
			    --emcmotDebug->stepperCount[axis];
			} else {
			    ++emcmotDebug->stepperCount[axis];
			}
			/* reload counters */
			upcount[axis] = upmag[axis];
			downcount[axis] = downmag[axis];
		    }		/* end of decrement counter if-then-else logic */
		}		/* end of if (! reversal) */
	    }			/* end of if (enable) */
	}			/* end of for (axis) loop */

#else				/* This is the new(er) freqfunc with setup & hold times */

	for (axis = 0; axis < FREQ_AXES; axis++) {
	    /* Only run IF axis is enabled - Gotta sort out this test. It is a mess ! */
	    if (GET_AXIS_ENABLE_FLAG(axis)
		&& (emcmotDebug->enable[axis]
		    || (upcount[axis] < setup_time[axis]))) {

		/* lpg - only axis that need correction will have enable[] == true */
		/* set mag and direction of period multiplier, if changed */
		if (emcmotDebug->pdmult[axis] != oldpdmult[axis]) {
		    if (emcmotDebug->pdmult[axis] == 0) {
			pdmag = 0;
			/* leave direction[] as it was */
		    } else if (emcmotDebug->pdmult[axis] < 0) {
			pdmag = -emcmotDebug->pdmult[axis];
			direction[axis] = 1;
		    } else {
			pdmag = emcmotDebug->pdmult[axis];
			direction[axis] = 0;
		    }

		    if (pdmag > 0) {
			upmag[axis] = pdmag - hold_time[axis];
			if (upmag[axis] < setup_time[axis])
			    upmag[axis] = setup_time[axis];
		    } else {
			upmag[axis] = -1;
		    }

		    // adjust upcount
		    if (upcount[axis] >= setup_time[axis]
			&& upmag[axis] > 0) {
			upcount[axis] = upmag[axis] - uptime[axis];
			// make sure we don't miss setup_time!
			if (upcount[axis] < setup_time[axis]) {
			    upcount[axis] = setup_time[axis];
			}
		    }		// end of if (upcount[axis] > 0)
		    oldpdmult[axis] = emcmotDebug->pdmult[axis];
		}		// end of if (emcmotDebug->pdmult != oldpdmult)

		/* decrement counters and run transition logic */
		if (upcount[axis] > 0) {
		    if (upcount[axis] == setup_time[axis]) {
			/* set direction bit */
			if (direction[axis]) {
			    if (axis < 4) {
				loByte |= (0x01 << (axis << 1));
			    } else {
				hiByte |= (0x01 << ((axis - 4) << 1));
			    }
			} else {
			    if (axis < 4) {
				loByte &= ~(0x01 << (axis << 1));
			    } else {
				hiByte &= ~(0x01 << ((axis - 4) << 1));
			    }
			}
		    }
		    /* keep delaying */
		    --upcount[axis];
		    ++uptime[axis];
		} else if (upcount[axis] == 0) {
		    /* timer expired -- do the down transition */
		    if (axis < 4) {
			loByte &= ~(0x02 << (axis << 1));
		    } else {
			hiByte &= ~(0x02 << ((axis - 4) << 1));
		    }
		    /* accumulate position */
		    if (direction[axis]) {
			--emcmotDebug->stepperCount[axis];
		    } else {
			++emcmotDebug->stepperCount[axis];
		    }
		    /* decr upcount only so we don't come back here (upcount[axis] = -1) */
		    --upcount[axis];
		} else if (downcount[axis] > 0) {
		    /* keep delaying */
		    --downcount[axis];
		} else if (downcount[axis] <= 0) {
		    /* timer expired -- do the up transition */
		    if (axis < 4) {
			loByte |= (0x02 << (axis << 1));
		    } else {
			hiByte |= (0x02 << ((axis - 4) << 1));
		    }

		    /* reload counters */
		    uptime[axis] = 0;
		    upcount[axis] = upmag[axis];
		    downcount[axis] = hold_time[axis];
		}		/* end of decrement counter if-then-else logic */
	    }			/* end of if (enable) */
	}			/* end of for (axis) loop */
#endif

	/* write the output */
	if (oldLoByte != loByte) {
	    extDioByteWrite(0, loByte);
	    oldLoByte = loByte;
	}
	if (emcmotConfig->numAxes > 4) {
	    if (oldHiByte != hiByte) {
		extDioByteWrite(1, hiByte);
		oldHiByte = hiByte;
	    }
	}

	rt_task_wait_period();
    }				/* end of while (1) */
}

/*
*======================================================================
*  quadfunc()
*  Routine to compute and output the phase drive bits needed by quadrature
*  phase stepper drive systems like the TMG (The Motion Group) electronics
*  package. The command input resides in the global array pdmult[] which
*  represents the number of cycles delay between phase changes. The
*  output is a global array emcmotDebug->stepperCount[] which is used as encoder feedback
*  to the servo loop controller. This routine
*  uses table lookup for the state transitions. 
*  This code gets run if you install module with STEPPING_TYPE = 1.
*  The faster this routine executes, the shorter you will be able to set
*  your PERIOD parameter without locking up the computer and the better
*  speed resolution you will have with your stepper system.
*  Remember:  this code has to execute in less than PERIOD seconds.
*  written 26-Oct-2000 by Lawrence P Glaister VE7IT 
*  ported to bdi 4.20 19mar05 by LPG
*  The beauty of quad drive is that there are NO setup and hold requirements
*  and the steppers could potentially move every PERIOD.
*======================================================================
*/
void quadfunc(long i)
{
    int axis;

#define PHASE_LENGTH 4
    static int phase[PHASE_LENGTH] = { 0, 1, 3, 2 };	/* phase bit lookup table       */
    static int active_phase[EMCMOT_MAX_AXIS] = { 0 };	/* cur index to phase table */
    static int downcount[EMCMOT_MAX_AXIS] = { 0 };	/* for timing phase xsitions */
    static unsigned int phasedrive = 0;	/* packed phase bits for output */
    /* supports up to 5phase*6axis  */
    static unsigned int lastdrive = 0;	/* for minimizing o/p writes    */

    int req_dly = 0;		/* temp. required delay         */

    diagnostics("quadfunc() started.\n");
    while (1) {
	/* Keep the following code SIMPLE and FAST                        */
	for (axis = 0; axis < EMCMOT_MAX_AXIS; axis++) {
	    if (emcmotDebug->enable[axis]) {
		req_dly = emcmotDebug->pdmult[axis];
		if (req_dly < 0)
		    req_dly = -req_dly;
		/* limit speed jumps at very highend... will stall steppers   */
		if (req_dly < 2)
		    req_dly = 2;
		if (downcount[axis] > 0) {
		    /* we are waiting for phase interval to expire              */
		    --downcount[axis];
		    /* be sure to keep up with rapid speed increases            */
		    if (req_dly < downcount[axis])
			downcount[axis] = req_dly;
		} else {
		    /* reload phase delay counter                               */
		    downcount[axis] = req_dly;

		    /* we need to update the phase drive bits and position      */
		    /* clear space for bits to drop into existing drive word    */
		    /* the following code assumes 2 bits / axis in output word  */
		    phasedrive &= ~(3 << (axis << 1));

		    if (emcmotDebug->pdmult[axis] >= 0) {
			/* FWD drive: adjust phase index and check for wraparound */
			++active_phase[axis];
			if (active_phase[axis] >= PHASE_LENGTH)
			    active_phase[axis] = 0;

			/* lookup and drop new phase bits into place              */
			phasedrive |=
			    phase[active_phase[axis]] << (axis << 1);

			/* let the world know that stepper moved                  */
			++emcmotDebug->stepperCount[axis];
		    } else {
			/* REV drive: adjust phase index and check for wraparound */
			--active_phase[axis];
			if (active_phase[axis] < 0)
			    active_phase[axis] = PHASE_LENGTH - 1;

			/* lookup and drop new phase bits into place              */
			phasedrive |=
			    phase[active_phase[axis]] << (axis << 1);

			/* let the world know that stepper moved                  */
			--emcmotDebug->stepperCount[axis];
		    }
		}		/* end of if downcount phase timing                        */
	    }			/* end of if (enable)                                         */
	}			/* end of for (axis) loop                                       */

	/* write the output to the parallel port if changed               */
	/* where are those 32 bit I/O ports when ya need them?            */
	if ((phasedrive & 0x000000ff) != (lastdrive & 0x000000ff))
	    extDioByteWrite(0, phasedrive);

	/* save a write to i/o for low axis count machines */
	if ((phasedrive & 0x0000ff00) != (lastdrive & 0x0000ff00))
	    extDioByteWrite(1, phasedrive >> 8);

	lastdrive = phasedrive;

	/* reschedule ourselves to run in another PERIOD seconds          */
	rt_task_wait_period();
    }				/* end of infinite while (1) loop */
#undef PHASE_LENGTH
}

/*
*======================================================================
*  lr4func()
*  Routine to compute and output the phase drive bits needed by 4 phase l/r
*  stepper drive systems like the Taig/Microproto electronics package.
*  The beauty of 4 phase drive is that there are NO setup and hold 
*  requirements and a VERY cheap motor driver is possible for low end systems.
*  This routine functions very similar to the quad drive above, but the 
*  number of phase states is longer.
*  Note that 4 i/o pins / axis are used. This has only been used with 3 axis.
*  Be wary of inverted pins... see the end of routine for adjustments.
*======================================================================
*/
void lr4func(long i)
{
    int axis;

#define PHASE_LENGTH 8
    static int phase[PHASE_LENGTH] = { 1, 3, 2, 6, 4, 12, 8, 9 };	/* phase bit lookup table */
    static int active_phase[EMCMOT_MAX_AXIS] = { 0 };	/* cur index to phase table */
    static int downcount[EMCMOT_MAX_AXIS] = { 0 };	/* for timing phase xsitions */
    static unsigned int phasedrive = 0;	/* packed phase bits for output */
    /* supports up to 5phase*6axis  */
    static unsigned int lastdrive = 0;	/* for minimizing o/p writes    */

    int req_dly = 0;		/* temp. required delay         */

    diagnostics("lr4func() started.\n");
    while (1) {
	/* Keep the following code SIMPLE and FAST                        */
	for (axis = 0; axis < EMCMOT_MAX_AXIS; axis++) {
	    if (emcmotDebug->enable[axis]) {
		req_dly = emcmotDebug->pdmult[axis];
		if (req_dly < 0)
		    req_dly = -req_dly;
		/* limit speed jumps at very highend... will stall steppers   */
		if (req_dly < 2)
		    req_dly = 2;
		if (downcount[axis] > 0) {
		    /* we are waiting for phase interval to expire              */
		    --downcount[axis];
		    /* be sure to keep up with rapid speed increases            */
		    if (req_dly < downcount[axis])
			downcount[axis] = req_dly;
		} else {
		    /* reload phase delay counter                               */
		    downcount[axis] = req_dly;

		    /* we need to update the phase drive bits and position      */
		    /* clear space for bits to drop into existing drive word    */
		    /* the following code assumes 4 bits / axis in output word  */
		    phasedrive &= ~(0x0f << (axis << 2));

		    if (emcmotDebug->pdmult[axis] >= 0) {
			/* FWD drive: adjust phase index and check for wraparound */
			++active_phase[axis];
			if (active_phase[axis] >= PHASE_LENGTH)
			    active_phase[axis] = 0;

			/* lookup and drop 4 new phase bits into place              */
			phasedrive |=
			    phase[active_phase[axis]] << (axis << 2);

			/* let the world know that stepper moved                  */
			++emcmotDebug->stepperCount[axis];
		    } else {
			/* REV drive: adjust phase index and check for wraparound */
			--active_phase[axis];
			if (active_phase[axis] < 0)
			    active_phase[axis] = PHASE_LENGTH - 1;

			/* lookup and drop new phase bits into place              */
			phasedrive |=
			    phase[active_phase[axis]] << (axis << 2);

			/* let the world know that stepper moved                  */
			--emcmotDebug->stepperCount[axis];
		    }
		}		/* end of if downcount phase timing                        */
	    }			/* end of if (enable)                                         */
	}			/* end of for (axis) loop                                       */

	/* write the output to the parallel port if changed               */
	/* where are those 32 bit I/O ports when ya need them?            */
	if ((phasedrive & 0x000000ff) != (lastdrive & 0x000000ff))
	    extDioByteWrite(0, phasedrive);

	/* save a write to i/o for low axis count machines */
	if ((phasedrive & 0x0000ff00) != (lastdrive & 0x0000ff00)) {
	    /* TNB - Pin 16 is inverted, which means the third bit of the   */
	    /* third motor nibble needs to be inverted.                     */
	    extDioByteWrite(1, (phasedrive ^ 0x00000400) >> 8);
	}
	lastdrive = phasedrive;

	/* reschedule ourselves to run in another PERIOD seconds          */
	rt_task_wait_period();
    }				/* end of infinite while (1) loop */
#undef PHASE_LENGTH
}



void freqsig(void)
{
}

/* stepper_tasks() starts up one of the real time tasks that manage */
/* outputing the signals to control the stepper motor driver.       */
int stepper_tasks(void)
{
    RTIME now;

    switch (STEPPING_TYPE) {
    case 0:			/* step direction interface */
	diagnostics("emcmot: initializing freqTask for step/dir\n");
	module_error_code = rt_task_init(&freqTask,	/* RT_TASK */
					 freqfunc,	/* task func */
					 1,	/* arg */
					 FREQ_TASK_STACK_SIZE,	/* stack */
					 FREQ_TASK_PRIORITY, 0,	/* Don't use fpu */
					 freqsig);	/* signal func */
	break;
    case 2:			/* 2 wire quadrature drive (like encoder) */
	diagnostics
	    ("emcmot: initializing freqTask for quadrature drive\n");
	module_error_code = rt_task_init(&freqTask,	/* RT_TASK */
					 quadfunc,	/* task func */
					 1,	/* arg */
					 FREQ_TASK_STACK_SIZE,	/* stack */
					 FREQ_TASK_PRIORITY, 0,	/* Don't use fpu */
					 freqsig);	/* signal func */

	break;

    case 3:			/* 4 wire raw l/r phase drive */
	diagnostics
	    ("emcmot: initializing freqTask for 4 wire raw l/r drive\n");
	module_error_code = rt_task_init(&freqTask,	/* RT_TASK */
					 lr4func,	/* task func */
					 1,	/* arg */
					 FREQ_TASK_STACK_SIZE,	/* stack */
					 FREQ_TASK_PRIORITY, 0,	/* Don't use fpu */
					 freqsig);	/* signal func */

	break;
	/* add new stepping mode cases in here... */

    default:
	diagnostics
	    ("emcmot: rats... invalid module param for stepping type\n");
	module_error_code = -999;	/* fall through to abort with another msg */
	break;
    }
    if (module_error_code != 0) {
	diagnostics
	    ("rt_task_init failed for freqTask module_error_code = %d\n",
	     module_error_code);
	return -1;
    }
    diagnostics("emcmot: rtaiTickPeriod=%d\n", (int) rtaiTickPeriod);
    if (rtaiTickPeriod > 0) {
	now = rt_get_time();
	diagnostics
	    ("emcmot: making freqTask periodic at a rate of %d ticks or %d nanoseconds \n",
	     ((int) rtaiTickPeriod_long),
	     ((int) count2nano(rtaiTickPeriod)));
	module_error_code =
	    rt_task_make_periodic(&freqTask, now + 5 * rtaiTickPeriod,
				  rtaiTickPeriod);
	if (module_error_code != 0) {
	    diagnostics
		("rt_task_make_periodic failed for freqTask module_error_code = %d\n",
		 module_error_code);
	    return -1;
	}
    }
    freqTask_created = 1;
    return 0;
}

void stepper_stop(void)
{
    if (freqTask_created) {
	diagnostics("emcmot: deleting freqTask.\n");
	rt_task_delete(&freqTask);
	freqTask_created = 0;
    }
}

void scaleOutputs(int max)
{
    int axis = 0;
    double pdmultfloat;
    int periodns = count2nano(rtaiTickPeriod);

    if (max > 6)
	max = 6;

    for (axis = 0; axis < max; axis++) {
	/* lpg - stepper motors, rawoutput is typically -10->0->+10 with
	 * zero indicating no correction required. For very small corrections,
	 * the enable gets shut off so no motion takes place and gets used below
	 * to avoid divide by zero calc (which would give infinite stepper periods)
	 */
	emcmotDebug->rawOutput[axis] = emcmotStatus->output[axis];
	if (emcmotDebug->rawOutput[axis] < +OUTPUT_DELTA &&
	    emcmotDebug->rawOutput[axis] > -OUTPUT_DELTA) {
	    emcmotDebug->enable[axis] = 0;
	} else {

	    pdmultfloat = 1e9 /
		(emcmotDebug->rawOutput[axis] *
		 ((double) emcmotStatus->outputScale[axis]) *
		 ((double) periodns));

	    /* round it */
	    if (pdmultfloat < 0.0) {
		emcmotDebug->pdmult[axis] = (int) (pdmultfloat - 0.5);
		emcmotDebug->enable[axis] = 1;
		/* lpg */
		/*
		   rounding may have caused a problem for very high gain/high
		   error calcs. If negative number close to zero get rounded to 0, 
		   we lose the sign information on pdmult. Be sure it gets restored.
		 */
		if (emcmotDebug->pdmult[axis] == 0) {
		    emcmotDebug->pdmult[axis] = -1;
		}
	    } else {
		emcmotDebug->pdmult[axis] = (int) (pdmultfloat + 0.5);
		emcmotDebug->enable[axis] = 1;
		/* lpg - make sure pdmult > 0 as zero means infinite step rate */
		if (emcmotDebug->pdmult[axis] == 0) {
		    emcmotDebug->pdmult[axis] = 1;
		}
	    }
	}
    }
    return;
}

/******************************************************************
 * Real-time timer management.
 * 
 * This module runs the timer at period much shorter than the servo
 * cycle time, to generate step pulses
 * 
 * ***************************************************************/


/* establishes the realtime time base, and starts up one of the real
 * time tasks that manage outputing the signals to control the stepper
 * motor driver.   */
int realtime_startup(int servoCycleTicks)
{
    int periodsPerServoCycle;

    diagnostics("defartinit - compiled on %s at %s\n", __DATE__, __TIME__);

    diagnostics("emcmot:  PERIOD = %dns\n", PERIOD);

    ticksPerPeriod = nano2count(PERIOD);
    rtaiTickPeriod = ticksPerPeriod;
    rtaiTickPeriod_long = ticksPerPeriod;

    if (ticksPerPeriod < 2) {
	diagnostics("emcmot:  bad period setting (%d)\n", PERIOD);
	return (-1);
    }

    diagnostics("emcmot:  ticks per PERIOD: %d\n", (int) ticksPerPeriod);

    if (servoCycleTicks < ticksPerPeriod)
	servoCycleTicks = ticksPerPeriod;

    periodsPerServoCycle =
	(servoCycleTicks + (ticksPerPeriod >> 1)) / ticksPerPeriod;

    diagnostics("emcmot:  periodsPerServoCycle=%d\n",
		(int) periodsPerServoCycle);

    servoCycleTicks = periodsPerServoCycle * ticksPerPeriod;

    diagnostics("emcmot:  starting RT timer: %d\n", (int) ticksPerPeriod);
    start_rt_timer(ticksPerPeriod);
    stepper_tasks();
    return (servoCycleTicks);
}

void realtime_shutdown(void)
{
    stop_rt_timer();
    stepper_stop();
}

int realtime_adjust(int servoCycleTicks)
{
    int periodsPerServoCycle;
    if (servoCycleTicks < ticksPerPeriod)
	servoCycleTicks = ticksPerPeriod;

    periodsPerServoCycle =
	(servoCycleTicks + (ticksPerPeriod >> 1)) / ticksPerPeriod;

    servoCycleTicks = periodsPerServoCycle * ticksPerPeriod;
    return (servoCycleTicks);
}
