/*
   pathseg.c

   Motion path segments.
   

   Author: Matt Timmermans

   Adapted from the original EMC tc.c, by Fred Proctor

   License: GPL Ver. 2
   System: Linux
    
   $Date: 2006-03-14 20:00:00 $

*/

#include <linux/types.h>
#ifdef realtime
#ifdef __attribute_used__
#undef __attribute_used__
#endif
#ifdef __attribute_pure__
#undef __attribute_pure__
#endif
#endif

#include <sys/cdefs.h>
#include <math.h>
#include "posemath.h"
#include "emcpos.h"
#include "pathseg.h"

/* ident tag */
#ifndef __GNUC__
#ifndef __attribute__
#define __attribute__(x)
#endif
#endif

static char __attribute__ ((unused)) ident[] =
    "$Id: pathseg.c,v 1.0 2006/03/13 20:00:00 mtimmerm Exp $";


void pathsegInitLinear(PathSeg * seg, EmcPose start, EmcPose end)
{
    double tmag, abc_mag, x, y, z;
    char *s = (char *) (void *) (&seg->circle);
    char *d = s + sizeof(seg->circle);
    for (; s < d; *s++ = 0);
    seg->start = start;
    seg->end = end;

    /* set arcLength to be scalar difference */
    x = seg->end.tran.x - seg->start.tran.x;
    y = seg->end.tran.y - seg->start.tran.y;
    z = seg->end.tran.z - seg->start.tran.z;
    tmag = sqrt(x * x + y * y + z * z);
    x = seg->end.a - seg->start.a;
    y = seg->end.b - seg->start.b;
    z = seg->end.c - seg->start.c;
    abc_mag = sqrt(x * x + y * y + z * z);

    if ((tmag < 1e-6) && (abc_mag > tmag)) {
	seg->length = abc_mag;
	seg->segType = PATHSEG_ROTATIONAL;
    } else {
	seg->length = tmag;
	seg->segType = PATHSEG_LINEAR;
    }
}

void pathsegInitCircular(PathSeg * seg, EmcPose start, EmcPose end,
			 PmCartesian center, PmCartesian normal, int turn)
{
    double tmag;
    seg->start = start;
    seg->end = end;

    pmCircleInit(&seg->circle, start.tran, end.tran, center, normal, turn);

    /* get the helical component, normal to the circle's plane */
    pmCartMag(seg->circle.rHelix, &tmag);
    tmag =
	pmSqrt(pmSq(seg->circle.angle * seg->circle.radius) + pmSq(tmag));
    seg->length = tmag;
    if (tmag > 1e-6) {
	seg->segType = PATHSEG_CIRCULAR;
    } else {
	seg->segType = PATHSEG_LINEAR;
    }
}

EmcPose pathsegGetPos(PathSeg * seg, double arcpos)
{
    EmcPose v;
    double relpos;

    if (0 == seg) {
	v.tran.x = v.tran.y = v.tran.z = v.a = v.b = v.c = 0.0;
	return v;
    }
    if (arcpos < 0.0)
	relpos = 0.0;
    else if (arcpos + 1e-10 >= seg->length)
	relpos = 1.0;
    else
	relpos = arcpos / seg->length;
    if (seg->segType == PATHSEG_CIRCULAR) {
	pmCirclePoint(&seg->circle, seg->circle.angle * relpos, &v.tran);
    } else {
	pmCartCartInterpolate(seg->start.tran, seg->end.tran, relpos,
			      &v.tran);
    }
    v.a = seg->start.a + ((seg->end.a - seg->start.a) * relpos);
    v.b = seg->start.b + ((seg->end.b - seg->start.b) * relpos);
    v.c = seg->start.c + ((seg->end.c - seg->start.c) * relpos);

    return v;
}

extern PmCartesian pathsegGetStartDirection(PathSeg * seg)
{
    PmCartesian ret;
    switch (seg->segType) {

    default:
	ret.x = ret.y = ret.z = 0.0;
	break;

    case PATHSEG_LINEAR:
	pmCartCartSub(seg->end.tran, seg->start.tran, &ret);
	pmCartScalMult(ret, 1.0 / seg->length, &ret);
	break;

    case PATHSEG_CIRCULAR:
	pmCircleDerivative(&seg->circle, 0.0, &ret);
	pmCartUnit(ret, &ret);

    }
    return ret;
}


extern PmCartesian pathsegGetEndDirection(PathSeg * seg)
{
    PmCartesian ret;
    switch (seg->segType) {

    default:
	ret.x = ret.y = ret.z = 0.0;
	break;

    case PATHSEG_LINEAR:
	pmCartCartSub(seg->end.tran, seg->start.tran, &ret);
	pmCartScalMult(ret, 1.0 / seg->length, &ret);
	break;

    case PATHSEG_CIRCULAR:
	pmCircleDerivative(&seg->circle, seg->circle.angle, &ret);
	pmCartUnit(ret, &ret);
    }
    return ret;
}
