/********************************************************************
*
* Description: quickstep.c
*
*	Faster stepper driver for EMC.  Replaces emcstepper.c, with
*	quickmod.c to form the quickstep module.
*
*	Stepping speed is maximized by installing a fast handler for
*	the timer IRQ underneath RTAI to do per-step work.  The real
*	timer frequency is divided to call RTAI's handler at
*	appropriate intervals.  Good for 30KHz (maybe over 50KHz)
*	stepping on a P266
*
*	THIS ONLY WORKS for kernels built for single processors, no local APIC,
*	and RTAI in immediate mode.
*
* Author: Matt Timmermans
* License: GPL Ver. 2
* System: Linux
*    
* $Date: 2006-02-10 20:00:00 $
*
* CHANGES:
* 2006-02-26    Matt Timmermans
*       	- Added quadrature step support.  respsects freqmod-style
* 		stepping_type=quad, but also allows the type of each
* 		axis to be specified independently.
* 2006-02-24    Matt Timmermans
*       	- Added realtime_* timer setup
* 2006-02-19	Matt Timmermans
*		- Strictly limited accelerations to set maximum + 5%,
*		because they aren't always respected by servo-centric
*		commands.  We don't need to worry about compensating
*		for load changes, so a non-linear (i.e., non-PID-like)
*		algorithm is used to maximize following accuracy and
*		responsiveness under the acceleration limits. If
*		commands stay within accel and velocity limits, it
*		introduces zero error.
*		Axes 0-2 use max(axis accel, 2*traj accel) as the set
*		limit. Axis 3 uses axis accel.
*		- simplified rt timer period calculation in quickmod.c
*		- latched encoder outputs at expected servo cycle
*		start to make velocity logs accurate even wehn emcmot's
*		schedule jitters.
*
********************************************************************/


#ifndef __KERNEL__
#define __KERNEL__
#endif
//#undef _RTAI_SHM_H_

#ifndef MODULE
#define MODULE
#endif
#ifndef EXPORT_SYMTAB
#define EXPORT_SYMTAB
#endif

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/interrupt.h>
#include <linux/irq.h>

#include "rtai.h"
#include "rtai_sched.h"
#include "rtai_shm.h"

#include "emcmotcfg.h"
#include "emcmotglb.h"
#include "emcmot.h"
#include "extintf.h"

#include "diagnostics.h"

#ifdef realtime
#ifdef __attribute_used__
#undef __attribute_used__
#endif
#ifdef __attribute_pure__
#undef __attribute_pure__
#endif
#endif

#include <sys/cdefs.h>
#include <float.h>
#include <math.h>

#ifdef CONFIG_X86_LOCAL_APIC
#error This module only works with CONFIG_X86_LOCAL_APIC off
#endif

MODULE_VERSION
    (" $Id: quickstep.c,v 1.0 2006/02/10 20:00:00 mtimmerm Exp $");

//===============================================================================
//      Parameters we get from elsewhere
//===============================================================================

extern int module_error_code;

//this is PERIOD, in clock ticks
static int ticksPerPeriod;
//how many PERIODs in the motion controller cycle time
static int periodsPerServoCycle;

int FIND_IO_BASE_ADDRESS;
extern int IO_BASE_ADDRESS;

//This gets returned by our interceptor to forward the interrupt to RTAI
static void *pold_timer_handler = 0;

//===============================================================================
//      Parameters we define
//===============================================================================

int PERIOD = 48000;
module_param(PERIOD, int, S_IRUGO);
MODULE_PARM_DESC(PERIOD, "Base period for the clock");

//Each nibble specifies a parport D0 bit (0-7), and polarity (+8 for normally high)
//Starting at the LSB, they are DIR0, STEP0, DIR1, STEP1... STEP3
int QS_PINOUT = 0x76543210;
module_param(QS_PINOUT, int, S_IRUGO);
MODULE_PARM_DESC(QS_PINOUT, "Step/direction pin mapping");

int QS_MIN_STEP_WIDTH = 5000;
module_param(QS_MIN_STEP_WIDTH, int, S_IRUGO);
MODULE_PARM_DESC(QS_MIN_STEP_WIDTH, "Minimum width of step pulse (ns)");

int QS_MIN_STEP_OFF = 5000;
module_param(QS_MIN_STEP_OFF, int, S_IRUGO);
MODULE_PARM_DESC(QS_MIN_STEP_OFF,
		 "Minimum off time between step pulses (ns)");

int QS_DIR_SETUP_HOLD = 80000;
module_param(QS_DIR_SETUP_HOLD, int, S_IRUGO);
MODULE_PARM_DESC(QS_DIR_SETUP_HOLD,
		 "Minimum time between DIR line changes and STEP line changes");

//Each byte specifies the stepping type for 1 axis, LSB for axis 0
//Meanings are:
//0 => use previous type from previous axis (step/dir for axis 0)
//This supports independent configs for all axes, while supporting
//stepping types from freqmod-based inis
int STEPPING_TYPE;
module_param(STEPPING_TYPE, int, S_IRUGO);
MODULE_PARM_DESC(STEPPING_TYPE, "stepping type for axes");

#define STEP_TYPE_STEPDIR 1
#define STEP_TYPE_QUAD 2

//===============================================================================
//      The stepper driver.  This runs once every PERIOD to generate steps.
//      No need to do all that PID stuff.
//      We take absolute positions from the motion controller and smooth out
//      the motion with a low-pass filter.
//      It's all integer, and perfectly predictable, so feedback is unnecessary.
//===============================================================================

// max number of axes that fit into one outb() 
#define NAXES 4

//The driver state for each axis
struct AxisDriver {
    //this is the target position, in steps, established by
    //the motion controleler
    int target;
    //the last target position accounted for by the driver.
    int targetack;
    //steps waiting to feedback to emc
    int delta_out_pending;

    int history[1024];
    int a, v;			//acceleration, velocity
    //position delta from last step output.
    //this is adjusted every step, which keeps it from overflowing
    int p;
    //periods since 
    int idletime;
    //this axis' step bit
    int step_bit;
    //this axis' dir bit
    int dir_bit;
    //this axis' stepping type
    int step_type;
} axs[NAXES];

//It takes step_lag_half*2-1 PERIODs for changes in the targetack
//position to fully manifest at the motors.
//This should be long enough to cover a few servo cycles to ensure
//smooth motion

static int step_lag_half;

//each unit step of targetack ends up changing p by exactly this amount
//(equals step_lag_half squared)
//when p changes by more than this amount, we take a step and adjust p
//to prevent overflow.

static int step_driver_gain;

//the step line has to be idle for this many periods before we will
//change the direction line.
//this can be lengthy, because we will be moving slowly while
//changing direction.

static int min_periods_pulse_to_dir = 4;

//extra time to hold off next pulse after a direction change
//again, this can be relatively long

static int min_periods_extra_dir_setup = 3;

//minimum width of a pulse

static int min_periods_pulse_on = 1;

//minimum off time

static int min_periods_pulse_off = 1;

//on the port, this means all dirs forward, all step pulses inactive
static int outbits_ground = 0;

//what goes out to the parallel port is outbits_current^outbits_ground
static int outbits_current = 0;

static int get_timing(const char *desc, int v)
{
    int pnanos = (int) count2nano(ticksPerPeriod);
    int pval = (v + (pnanos - 1)) / pnanos;
    diagnostics("EMCQS - %s time %dns -> %dns (%d periods)\n", desc, v,
		pval * pnanos, pval);
    return pval;
}

static void init_timings(void)
{
    min_periods_pulse_on = get_timing("STEP PULSE ON", QS_MIN_STEP_WIDTH);
    min_periods_pulse_off = get_timing("STEP PULSE OFF", QS_MIN_STEP_OFF);
    min_periods_pulse_to_dir =
	get_timing("DIR SETUP&HOLD", QS_DIR_SETUP_HOLD);
    if (min_periods_pulse_to_dir < min_periods_pulse_off) {
	min_periods_extra_dir_setup = 0;
    } else {
	min_periods_extra_dir_setup =
	    min_periods_pulse_to_dir + min_periods_pulse_off;
    }
}

//keep track of bits allocated, so we can write out errors for duplicate
//signals
static int outbits_used = 0;

static int get_pinout_bit(const char *desc, int axis, int pinout)
{
    int bitnum = pinout & 7;
    int bit = 1 << bitnum;
    const char *active = "HIGH";
    if (pinout & 8) {
	outbits_ground |= bit;
	active = "LOW";
    }

    diagnostics("EMCQS - AXIS %d %s 0x%02X D%d pin%d Active %s\n",
		axis, desc, bit, bitnum, bitnum + 2, active);
    if (outbits_used & bit) {
	diagnostics("EMCQS - BAD PINOUT -- TWO SIGNALS ON SAME PIN\n");
    }
    outbits_used |= bit;
    return bit;
}

static int get_stepping_type(int axis, int typespec, int prevtype)
{
    const char *desc;
    typespec &= 255;
    if (!typespec)
	typespec = prevtype;
    switch (typespec) {
    case STEP_TYPE_STEPDIR:
	desc = "STEP/DIR";
	break;

    case STEP_TYPE_QUAD:
	desc = "QUADRATURE";
	break;

    default:
	desc = "ERROR - UNKNOWN";
	break;
    }
    diagnostics("EMCQS - AXIS %d stepping type is %s (%d)\n", axis, desc,
		typespec);
    return typespec;
}

static void init_axes(int halflag)
{
    int i;
    int pinout = QS_PINOUT;
    int steptypes = STEPPING_TYPE;
    int stype = STEP_TYPE_STEPDIR;
    char *p = (char *) (void *) &(axs[0]);
    char *pend = p + sizeof(axs);

    for (; p < pend; *p++ = 0);

    if (halflag > 512)
	halflag = 512;
    if (halflag < 2)
	halflag = 2;
    step_lag_half = halflag;
    step_driver_gain = halflag * halflag;
    outbits_ground = 0;
    for (i = 0; i < NAXES; i++) {
	//this is the neutral middle of the current step
	axs[i].p = step_driver_gain >> 1;

	stype = get_stepping_type(i, steptypes, stype);
	axs[i].step_type = stype;
	steptypes >>= 8;

	if (stype == STEP_TYPE_STEPDIR) {
	    axs[i].dir_bit = get_pinout_bit(" DIR", i, pinout);
	    pinout >>= 4;
	    axs[i].step_bit = get_pinout_bit("STEP", i, pinout);
	    pinout >>= 4;
	} else {
	    axs[i].dir_bit = get_pinout_bit("Q0", i, pinout);
	    pinout >>= 4;
	    axs[i].step_bit = get_pinout_bit("Q1", i, pinout);
	    pinout >>= 4;
	}
    }
    outbits_current = outbits_ground;
}

//This is returned by our interceptor to return from the interrupt, when it
//doesn't get forwarded to RTAI

void QUICKSTEP_TIMER_HANDLED(void);
__asm__("\n" __ALIGN_STR "\n\t"
	"QUICKSTEP_TIMER_HANDLED:\n\t"
	"iret\n\t" "iret\n\t" "iret\n\t" "iret");



static int rt_skips = 0;	//countdown to RTAI tick
static int hpos = 0;		//current position in the history buffers
void *quickstep_interceptor(void)
{
    int axi, incr, pdelta;

    //do the parallel port output at the start, where timing is less
    //variable.

    outb(outbits_current ^ outbits_ground, IO_BASE_ADDRESS);

    hpos = (hpos - 1) & 1023;
    for (axi = 0; axi < NAXES; axi++) {
	//This is the lowpass filter that turns motion commands into steps.
	//We use a 3-tap filter and two integrators to turn each position change
	//into two periods of equal and opposite accelleration.  Outputs from
	//all the incremental changes sum linearly and losslessly to produce the
	//smooth motion.
	struct AxisDriver *ax = &(axs[axi]);
	incr = ax->target - ax->targetack;
	ax->targetack += incr;
	ax->a += incr - (ax->history[(hpos + step_lag_half) & 1023] << 1)
	    + ax->history[(hpos + step_lag_half + step_lag_half) & 1023];
	ax->history[hpos] = incr;
	ax->v += ax->a;
	ax->p += ax->v;

	//Tada, p is updated
	//now modulate the output bits, according to accumulated
	//steps and time constraints
	pdelta = ax->p;
	incr = ax->idletime;
	incr -= ((incr - 0x00FFFFFF) >> 31);	//saturatng increment

	switch (ax->step_type) {

	case STEP_TYPE_STEPDIR:

	    if (outbits_current & ax->step_bit) {
		//step is high -- next thing this axis does is bring it down
		if (incr >= min_periods_pulse_on) {
		    outbits_current &= ~ax->step_bit;
		    incr = 0;
		}
	    } else if (pdelta >= step_driver_gain) {
		//step up is pendng
		if (outbits_current & ax->dir_bit) {
		    //need to change direction
		    if (incr >= min_periods_pulse_to_dir) {
			outbits_current &= ~ax->dir_bit;
			incr = -min_periods_extra_dir_setup;
		    }
		} else {
		    //step if we can
		    if (incr >= min_periods_pulse_off) {
			ax->p = pdelta - step_driver_gain;
			outbits_current |= ax->step_bit;
			incr = 0;
			//emc "encoder feedback"
			ax->delta_out_pending++;
		    }
		}
	    } else if (pdelta < 0) {
		//step down is pending
		if (!(outbits_current & ax->dir_bit)) {
		    //need to change direction
		    if (incr >= min_periods_pulse_to_dir) {
			outbits_current |= ax->dir_bit;
			incr = -min_periods_extra_dir_setup;
		    }
		} else {
		    //step if we can
		    if (incr >= min_periods_pulse_off) {
			ax->p = pdelta + step_driver_gain;
			outbits_current |= ax->step_bit;
			incr = 0;
			//emc "encoder feedback"
			ax->delta_out_pending--;
		    }
		}
	    }
	    break;

	case STEP_TYPE_QUAD:
	    if (incr >= min_periods_pulse_on) {
		int phase;
		//enough time has elapsed since the last change
		if (pdelta >= step_driver_gain) {
		    //step up
		    ax->p = pdelta - step_driver_gain;
		    ax->delta_out_pending++;
		    incr = 0;
		} else if (pdelta < 0) {
		    //step down
		    ax->p = pdelta + step_driver_gain;
		    ax->delta_out_pending--;
		    incr = 0;
		}
		//compute the phase directly from the reported step count
		phase = emcmotDebug->stepperCount[axi] +
		    ax->delta_out_pending;
		if (phase & 2)
		    outbits_current |= ax->step_bit | ax->dir_bit;
		else
		    outbits_current &= ~(ax->step_bit | ax->dir_bit);
		if (phase & 1)
		    outbits_current ^= ax->dir_bit;
	    }

	    break;
	}
	ax->idletime = incr;

	if (!rt_skips) {
	    //only sent output to emcmot when rtai should be about to
	    //schedule it.  This normally avoids errors in velocity logs
	    //due to servo cycle scheduling jitter
	    emcmotDebug->stepperCount[axi] += ax->delta_out_pending;
	    ax->delta_out_pending = 0;
	}

    }

    //we return the address of the interrupt completion handler
    if (--rt_skips < 0) {
	//give this one to RTAI so it can schedule the motion controller
	rt_skips += periodsPerServoCycle;
	return pold_timer_handler;
    } else {
	//this one was just for us -- ack it and have done
	outb(0x20, 0x20);
	return QUICKSTEP_TIMER_HANDLED;
    }
}


//===============================================================================
//      Timer IRQ interceptor
//===============================================================================


//Lifted from RTAI
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
#define __LXRT_GET_DATASEG(reg) "movl $" STR(__KERNEL_DS) ",%" #reg "\n\t"
#else				/* KERNEL_VERSION >= 2.6.0 */
#ifndef STR
#define __STR(x) #x
#define STR(x) __STR(x)
#endif
#define __LXRT_GET_DATASEG(reg) "movl $" STR(__USER_DS) ",%" #reg "\n\t"
#endif				/* KERNEL_VERSION < 2.6.0 */

void quickstep_timer_handler(void);
__asm__("\n" __ALIGN_STR "\n\t"
	"quickstep_timer_handler:\n\t"
	"cld\n\t"
	"pushl $0\n\t"
	"pushl %eax\n\t"
	"pushl %es\n\t"
	"pushl %ds\n\t"
	"pushl %ebp\n\t"
	"pushl %edi\n\t"
	"pushl %esi\n\t"
	"pushl %edx\n\t"
	"pushl %ecx\n\t" "pushl %ebx\n\t" __LXRT_GET_DATASEG(ecx)
	"movl %ecx, %ds\n\t"
	"movl %ecx, %es\n\t"
	"call quickstep_interceptor\n\t"
	"popl %ebx\n\t"
	"popl %ecx\n\t"
	"popl %edx\n\t"
	"popl %esi\n\t"
	"popl %edi\n\t"
	"popl %ebp\n\t"
	"popl %ds\n\t"
	"popl %es\n\t" "movl %eax, 4(%esp)\n\t" "popl %eax\n\t" "ret");

static struct desc_struct old_timer_save;
static int last_tick_period_set = 0;

static void set_timer_freq(unsigned tick)
{
    outb(0x34, 0x43);
    outb(tick & 0xff, 0x40);
    outb(tick >> 8, 0x40);
    last_tick_period_set = tick;
}

static void set_timer_vector(void)
{
    unsigned long flags;
    if (!pold_timer_handler) {
	unsigned x1, x2;
	unsigned vector = FIRST_EXTERNAL_VECTOR + RTAI_TIMER_8254_IRQ;
	void *handler = quickstep_timer_handler;
	rtai_save_flags_and_cli(flags);
	old_timer_save = idt_table[vector];
	x1 = old_timer_save.a & 0x0000FFFF;
	x2 = old_timer_save.b & 0xFFFF0000;
	pold_timer_handler = (void *) (x1 | x2);
	idt_table[vector].a =
	    (__KERNEL_CS << 16) | ((unsigned) handler & 0x0000FFFF);
	idt_table[vector].b =
	    ((unsigned) handler & 0xFFFF0000) | (0x8000 + (14 << 8));
	set_timer_freq((int) ticksPerPeriod);
	rtai_restore_flags(flags);
	diagnostics("EMCQS - set timer vector.  Period=%d\n",
		    last_tick_period_set);
    }
}

static void reset_timer_vector(void)
{
    unsigned long flags;
    if (pold_timer_handler) {
	unsigned vector = FIRST_EXTERNAL_VECTOR + RTAI_TIMER_8254_IRQ;
	rtai_save_flags_and_cli(flags);
	idt_table[vector] = old_timer_save;
	set_timer_freq((int) ticksPerPeriod * periodsPerServoCycle);
	rtai_restore_flags(flags);
	pold_timer_handler = 0;
	diagnostics("EMCQS - restored timer vector.  Period=%d\n",
		    last_tick_period_set);
    }
}


//===============================================================================
//      Move axes in servo cycle
//===============================================================================

//For any particular positional error and accelleration limit, there is
//an ideal relative velocity that we should try to match.  This is the
//velocity that will get us to the command position at the command
//velocity as quickly as possible.  It is the maximum speed from which
//we can brake in time.
//We assume constant velocity during each cycle, which makes this
//curve piecewise-linear.
//The accelleration limit is given in x units/cycle/cycle, and
//the velocity returned is in x units/cycle.

static double get_best_v(double e, double alimit)
{
    double vl, vh;
    if (alimit <= 0.0)
	return 0.0;
    e /= alimit;
    if (e < 0.0) {
	e = -e;
	alimit = -alimit;
    }
    vl = floor((sqrt(8.0 * e + 1.0) - 1.0) * 0.5);
    vh = vl + 1.0;
    e -= vl * vh * 0.5;
    return (alimit * (vl + e / vh));
}


static struct JointState {
    double commandPosPrev;
    double stepPosPrev, stepPosOld;
} jointStates[NAXES];

static void init_joint_states(void)
{
    int axi;
    struct JointState *js;
    for (axi = 0; axi < NAXES; ++axi) {
	js = &(jointStates[axi]);
	js->commandPosPrev = 0.0;
	js->stepPosPrev = 0.0;
	js->stepPosOld = 0.0;
    }
}

//emcmot calls this

void scaleOutputs(int max)
{
    int axi;
    double trajacc = emcmotStatus->acc * 2.0;

    for (axi = 0; axi < NAXES; ++axi) {
	double commandPos;
	double stepPos;
	struct JointState *js = &(jointStates[axi]);

	//get the velocity and acceleration limits for this axis, in
	//our natural units.  We add 5% to allow for differences in interpolation
	//methods between this code and the rest of EMC
	double vfac = 1.05 * fabs(emcmotStatus->inputScale[axi])
	    * emcmotConfig->servoCycleTime;
	double vlimit = emcmotConfig->axisLimitVel[axi] * vfac;	// steps/cycle
	double alimit = emcmotConfig->axisAcc[axi];
	if (alimit < trajacc && axi < 3)	//any other way to tell angular from linear?
	    alimit = trajacc;
	alimit *= vfac * emcmotConfig->servoCycleTime;	// steps/cycle/cycle

	if ((axi < max) && (GET_AXIS_ENABLE_FLAG(axi))) {
	    //ignore the PID outputs, and use get the target position
	    //in steps directly
	    commandPos = emcmotDebug->outJointPos[axi]
		* emcmotStatus->inputScale[axi] +
		emcmotStatus->inputOffset[axi];
	    if (vlimit < 0.0)
		vlimit = 0.0;
	} else {
	    commandPos = js->commandPosPrev;
	    vlimit = 0.0; //decelerate to a stop
	}

	if (alimit > 0.0) {
	    //get the commanded velocity
	    double vc = commandPos - js->commandPosPrev;
	    //and the old real velocity
	    double vPrev = js->stepPosPrev - js->stepPosOld;

	    double vBest;

	    //clamp commanded velocity to limit
	    if (vc > vlimit)
		vc = vlimit;
	    else if (vc < -vlimit)
		vc = -vlimit;
	    //get best velocity -- ain't steppers grand!
	    vBest =
		get_best_v(js->commandPosPrev - js->stepPosPrev,
			   alimit * 0.5) + vc;

	    //clamp it
	    if (vBest > vlimit)
		vBest = vlimit;
	    else if (vBest < -vlimit)
		vBest = -vlimit;

	    //get as close to it as we can without the accel bound
	    vc = vBest - vPrev;
	    if (vc > alimit)
		vBest = vPrev + alimit;
	    else if (vc < -alimit)
		vBest = vPrev - alimit;

	    //set next cycle target
	    stepPos = js->stepPosPrev + vBest;
	} else {
	    stepPos = js->stepPosPrev;
	}

	js->stepPosOld = js->stepPosPrev;
	js->stepPosPrev = stepPos;
	js->commandPosPrev = commandPos;

	if (stepPos >= 0.0)
	    axs[axi].target = (int) (stepPos + 0.5);
	else
	    axs[axi].target = (int) (stepPos - 0.5);

    }
}

//===============================================================================
//      Init/destroy
//===============================================================================

int realtime_startup(int servoCycleTicks)
{
    diagnostics("EMCQS - compiled on %s at %s\n", __DATE__, __TIME__);
    diagnostics("EMCQS - initializing quickstep realtime\n");

    diagnostics("EMCQS - PERIOD = %dns\n", PERIOD);

    ticksPerPeriod = nano2count(PERIOD);

    if (ticksPerPeriod < 2) {
	diagnostics("EMCQS - bad period setting (%d)\n", PERIOD);
	return (-1);
    }

    diagnostics("EMCQS - ticks per PERIOD: %d\n", (int) ticksPerPeriod);

    if (servoCycleTicks < ticksPerPeriod)
	servoCycleTicks = ticksPerPeriod;

    periodsPerServoCycle =
	(servoCycleTicks + (ticksPerPeriod >> 1)) / ticksPerPeriod;

    diagnostics("EMCQS - periodsPerServoCycle=%d\n",
		(int) periodsPerServoCycle);

    servoCycleTicks = periodsPerServoCycle * ticksPerPeriod;

    diagnostics("EMCQS - starting RT timer: %d\n", (int) servoCycleTicks);
    start_rt_timer(servoCycleTicks);

    module_error_code = 0;

    init_timings();
    diagnostics("EMCQS - IO_BASE_ADDRESS=0x%04X\n", IO_BASE_ADDRESS);
    diagnostics("EMCQS - setting up axes\n");
    init_axes(periodsPerServoCycle * 2 + 1);
    diagnostics("EMCQS - step driver output lag = %d ticks\n",
		step_lag_half * 2 - 1);
    init_joint_states();
    set_timer_vector();
    diagnostics("EMCQS - timer vector set.\n");

    return (servoCycleTicks);
}

void realtime_shutdown(void)
{
    diagnostics("EMCQS - shutting down quickstep\n");
    reset_timer_vector();
    diagnostics("EMCQS - timer vector reset\n");
    diagnostics("EMCQS - stopping RT timer\n");
    stop_rt_timer();
}

int realtime_adjust(int servoCycleTicks)
{
    int newppsc;

    if (servoCycleTicks < ticksPerPeriod)
	servoCycleTicks = ticksPerPeriod;

    newppsc = (servoCycleTicks + (ticksPerPeriod >> 1)) / ticksPerPeriod;
    servoCycleTicks = newppsc * ticksPerPeriod;

    if (newppsc != periodsPerServoCycle) {
	periodsPerServoCycle = newppsc;

	diagnostics("EMCQS - changing realtime timer\n");

	diagnostics("EMCQS - resetting timer vector\n");
	reset_timer_vector();
	diagnostics("EMCQS - stopping RT timer\n");
	stop_rt_timer();

	diagnostics("EMCQS - periodsPerServoCycle=%d\n",
		    (int) periodsPerServoCycle);

	diagnostics("EMCQS - starting RT timer: %d\n",
		    (int) servoCycleTicks);
	start_rt_timer(servoCycleTicks);

	module_error_code = 0;

	diagnostics("EMCQS - setting timer vector\n");
	set_timer_vector();
	diagnostics("EMCQS - realtime timer changed\n");
    }
    return (servoCycleTicks);
}
