/*
  Trajectory planning based on path blending.

  Derived from the original motion-blending trajectory planner
  for EMC.  The original algorithm filleted paths with quadratic
  splines by superimposing equal-timed portions of distinct
  trapezoidal-velocity-profile motions.  That was a remarkably simple
  way to generate a filleted path, but path errors dependend upon
  acceleration settings, and it didn't achieve the fastest traversals
  over multitudes of small segments.
  
  This version creates a similarly filleted path by superimposing
  and interpolating between equal-arc-length portions of path segments,
  and planning motion along the new path after the fact.  Path errors
  are limited by explicit tolerances, and lookahead allows for the 
  most efficient traversal.

  Author (BDI trajplan.c): Matt Timmermans
  
  Derived from EMC tp&tc (motion blending) by Fred Proctor, with
  	contributions from Will Shackleford and Paul Corner
  	
  Derived source code should include all of these attributions
  
  License: GPL Version 2
  System: Linux
  
  Changes:

  25-Mar-2006 Matt Timmermans switched to fast unbounded lookahead based on
              "energy restrictions"
  18-Mar-2006 Matt Timmermans added many-segment lookahead based on
  	      "lookahead stop" segments.
  14-Mar-2006 Matt Timmermans rewrote most of this file to convert from
  	      motion blending to path blending

  EMC tp.c Change log:
  
  16-Nov-2002 P.C. Changed dout to use extDioWrite so that one bit is used
  at a time - Handy when IO is limited to one port (or card).
  13-Nov-2002  FMP moved some dereferences to thisTc in tpRunCycle to be
  after a check for thisTc being NULL
  14-Dec-2001  FMP fixed some compile warnings: comparison of unsigned char
  < zero, implicit def of extMotDout()
  7-Dec-2001  FMP took hard-coded 1e-6 values out and moved them
  into tp.h as EPSILON defines; added setting of dout byte when aborted
  17-Aug-2001  FMP added aout,dout motion IO
  16-Nov-2000 WPS modified the algorithm to compute preAMax and
  preVMax, to elminate some velocity spikes, and ensure that the
  acceleration and velocity maximums were really honored.
  13-Mar-2000 WPS added unused attribute to ident to avoid 'defined
  but not used' compiler warning.
  28-Feb-2000 WPS eliminated #include <stdlib.h> when compiling for rtlinux
  stdlib.h may be incompatible with one of the header files in rtlinux-2.0.
  23-Sep-1999 WPS replaced printf with rcs_print  which is supported under CE
  8-Jun-1999  FMP added tpSetVlimit(), vLimit
  8-Mar-1999  FMP added tcSpace arg to tpCreate()
  25-Jun-1998  FMP added v to premax
  15-Jun-1998  FMP added check for TC_TERM_BLEND before blending next
  move, in tpRunCycle(); tpSet,GetTermCond()
  22-Jan-1998  FMP honored return value from tcqCreate, in tpCreate
  18-Dec-1997  FMP changed to EmcPose
  15-Nov-1997 FMP set vScale to vRestore in tpClear()
  5-Sep-1997  WPS modified tpRunCycle to prevent motions from being
  removed from the queue when later smaller motions complete before
  an earlier longer motion, and use the execId of the earliest incomplete
  motion rather than the last one being blended since this makes the
  agreement between the motionLine and the position look much better.
  25-Jul-1997  FMP cleared execId after done
  16-Jul-1997  FMP added ids
  14-Jul-1997  FMP added C posemath changes (PM_POSE -> EmcPose)
  13-Jun-1997  FMP added call to tcInit in tpAddMotion; added
  tpSetVscale, tpIsPaused
  2-Apr-1997  FMP changed MOTION_QUEUE to TC_QUEUE
  13-Mar-1997  FMP added return value to init
  29-Jan-1997  FMP changed vec.h to posemath.h
  23-Jan-1997  FMP created from tc.c
*/


#include <linux/types.h>
#ifdef realtime
#ifdef __attribute_used__
#undef __attribute_used__
#endif
#ifdef __attribute_pure__
#undef __attribute_pure__
#endif
#else
#include <stdio.h>
#endif

#include <sys/cdefs.h>
#include <float.h>
#include <math.h>

#include "posemath.h"
#include "emcmot.h"
#include "pathseg.h"
#include "trajplan.h"
#include "extintf.h"		/* extMotDout() */
#include "diagnostics.h"

/* ident tag */
#ifndef __GNUC__
#ifndef __attribute__
#define __attribute__(x)
#endif
#endif

int BLEND_TOL_REL_NS = 4000000;

static char __attribute__ ((unused)) ident[] =
    "$Id: trajplan.c,v 1.10 2006/03/14 20:20:20 mtimmerm Exp $";

#define tpQItem(tp,i) ((tp)->tcqBuf+(((tp)->tcqStart+(i))%(tp)->tcqBufSize))

int tpCreate(TP_STRUCT * tp, int _queueSize, TC_STRUCT * tcSpace)
{
    if (0 == tp) {
	return -1;
    }

    tp->tcqBuf = tcSpace;
    tp->tcqBufSize = _queueSize;
    tp->tcqStart = tp->tcqLen = 0;
    tp->lhqStart = tp->lhqLen = 0;
    tp->energyCount0 = tp->energyCount1 = 0;

    /* init the rest of our data */
    return tpInit(tp);
}

int tpDelete(TP_STRUCT * tp)
{
    if (0 == tp) {
	return 0;
    }
    tp->tcqBuf = 0;
    tp->tcqBufSize = 0;
    tp->tcqStart = tp->tcqLen = 0;
    tp->lhqStart = tp->lhqLen = 0;
    return 0;
}

/*
  tpClear() is a "soft init" in the sense that the TP_STRUCT configuration
  parameters (cycleTime, vMax, and aMax) are left alone, but the queue is
  cleared, and the flags are set to an empty, ready queue. The currentPos
  is left alone, and goalPos is set to this position.

  This function is intended to put the motion queue in the state it would
  be if all queued motions finished at the current position.
 */
int tpClear(TP_STRUCT * tp)
{
    if (tp->doutPos >= 0 && tp->tcqLen > tp->doutPos) {
	TC_STRUCT *tc = tpQItem(tp, tp->doutPos);
	if (tc->params.douts)
	    extDioWrite(tc->params.doutIndex, tc->params.doutends);
    }

    tp->tcqStart = tp->tcqLen = 0;
    tp->lhqStart = tp->lhqLen = 0;
    tp->energyCount0 = tp->energyCount1 = 0;
    tp->doutPos = -1;
    tp->currentArcPos = tp->nextArcPos = 0.0;
    tp->goalPos = tp->currentPos;
    tp->nextId = 0;
    tp->execId = 0;
    tp->params.blend_error_max = -1.0;
    tp->activeDepth = 0;
    tp->aborting = 0;
    tp->pausing = 0;

    return 0;
}

int tpInit(TP_STRUCT * tp)
{
    tp->cycleTime = 0.0;
    tp->params.vLimit = 0.0;
    tp->params.aMax = 0.0;
    tp->params.vMax = 0.0;
    tp->params.blend_error_max = 0.0;
    tp->params.douts = 0;
    tp->params.doutIndex = 0;
    tp->params.doutstarts = tp->params.doutends = 0;

    tp->vScale = 1.0;
    tp->wMax = 0.0;
    tp->wDotMax = 0.0;
    tp->default_blend_tol_base = 0.0;
    tp->default_blend_tol_relative = 1.0e-9 * (double) BLEND_TOL_REL_NS;

    tp->currentPos.tran.x = 0.0;
    tp->currentPos.tran.y = 0.0;
    tp->currentPos.tran.z = 0.0;
    tp->currentPos.a = 0.0;
    tp->currentPos.b = 0.0;
    tp->currentPos.c = 0.0;

    return tpClear(tp);
}

int tpSetCycleTime(TP_STRUCT * tp, double secs)
{
    if (0 == tp || secs <= 0.0) {
	return -1;
    }

    tp->cycleTime = secs;

    return 0;
}

int tpSetVmax(TP_STRUCT * tp, double vMax)
{
    if (0 == tp || vMax <= 0.0) {
	return -1;
    }

    tp->params.vMax = vMax;

    return 0;
}

int tpSetWmax(TP_STRUCT * tp, double wMax)
{
    if (0 == tp || wMax <= 0.0) {
	return -1;
    }

    tp->wMax = wMax;

    return 0;
}



int tpSetVlimit(TP_STRUCT * tp, double vLimit)
{
    if (0 == tp || vLimit <= 0.0) {
	return -1;
    }

    tp->params.vLimit = vLimit;

    return 0;
}

int tpSetVscale(TP_STRUCT * tp, double scale)
{
    if (0 == tp) {
	return -1;
    }

    if (scale < 0.0) {
	/* clamp it */
	scale = 0.0;
    }

    tp->vScale = scale;

    return 0;
}

int tpSetAmax(TP_STRUCT * tp, double aMax)
{
    if (0 == tp || aMax <= 0.0) {
	return -1;
    }

    tp->params.aMax = aMax;

    return 0;
}

int tpSetWDotmax(TP_STRUCT * tp, double wdotmax)
{
    if (0 == tp || wdotmax <= 0.0) {
	return -1;
    }

    tp->wDotMax = wdotmax;

    return 0;
}

/*
  tpSetId() sets the id that will be used for the next appended motions.
  nextId is incremented so that the next time a motion is appended its id
  will be one more than the previous one, modulo a signed int. If
  you want your own ids for each motion, call this before each motion
  you append and stick what you want in here.
  */
int tpSetId(TP_STRUCT * tp, int id)
{
    if (0 == tp) {
	return -1;
    }

    tp->nextId = id;

    return 0;
}

/*
  tpGetNextId() returns the id that will be used for the next appended motion.
  */
int tpGetNextId(TP_STRUCT * tp)
{
    if (0 == tp) {
	return -1;
    }

    return tp->nextId;
}

/*
  tpGetExecId() returns the id of the last motion that is currently
  executing.
  */
int tpGetExecId(TP_STRUCT * tp)
{
    if (0 == tp) {
	return -1;
    }

    return tp->execId;
}

/* use exact stops for subsquent moves */
int tpSetModeExactStop(TP_STRUCT * tp)
{
    if (0 == tp) {
	return -1;
    }
    tp->params.blend_error_max = 0.0;
    return 0;
}

/* use default blend tolerance for subsequent moves */
int tpSetModeDefaultBlend(TP_STRUCT * tp)
{
    if (0 == tp) {
	return -1;
    }
    tp->params.blend_error_max = -1.0;
    return 0;
}

/* use specified blend tolerance for subsequent moves */
int tpSetModeLimitedBlend(TP_STRUCT * tp, double tol)
{
    if (0 == tp) {
	return -1;
    }
    if (tol <= 0.0)
	tol = 0.0;
    tp->params.blend_error_max = tol;
    return 0;
}


/* set tolerances for default blend mode */
int tpSetDefaultBlendTolerance(TP_STRUCT * tp, double base, double rel)
{
    if (base <= 0.0)
	base = 0.0;
    if (rel <= 0.0)
	rel = 0.0;
    tp->default_blend_tol_base = base;
    tp->default_blend_tol_relative = rel;
    return 0;
}


int tpSetPos(TP_STRUCT * tp, EmcPose pos)
{
    if (0 == tp) {
	return -1;
    }

    tp->currentPos = pos;
    tp->goalPos = pos;

    return 0;
}


/*
 * get the first segment in the lookahead queue, or null if
 * it's empty
 */

static TC_STRUCT *lhqFirst(TP_STRUCT * tp)
{
    if (tp->lhqLen <= 0)
	return 0;
    return tp->tcqBuf[tp->lhqStart].lhentry;
}


/*
 * Add a new segment to the end of the lookahead queue.  This function
 * first removes all the existing entries with higher energy limits
 * They will never be the most restrictive entry, so we needn't remember
 * them.
 *
 * Energy limits in the lh queue will therefore increase monotonically,
 * which lets us find the truncation point with a binary search.
 */

static void lhqAppend(TP_STRUCT * tp, TC_STRUCT * newtc)
{
    TC_STRUCT *buf = tp->tcqBuf;
    int countphase = (tp->energyCount0 >= tp->energyCount1 ? 0 : 1);
    double nextElimit = newtc->energyLimits[countphase];
    int bufsz = tp->tcqBufSize;
    int minpos = tp->lhqStart % bufsz;
    int limpos = minpos + tp->lhqLen;
    int testpos;
    int correction = -minpos;

    /* check the wraparound point first so we don't have to
       do the modulus on every iteration */

    if (limpos > bufsz) {
	if (buf[0].lhentry->energyLimits[countphase] >= nextElimit) {
	    /* stuff after wraparound goes */
	    limpos = bufsz;
	} else {
	    /* some of the wraparound stays */
	    correction += bufsz;
	    minpos = 0;
	    limpos -= bufsz;
	}
    }

    while (limpos > minpos) {
	testpos = (limpos + minpos) >> 1;
	if (buf[testpos].lhentry->energyLimits[countphase] >= nextElimit) {
	    /* test goes */
	    limpos = testpos;
	} else {
	    /* test stays */
	    minpos = testpos + 1;
	}
    }

    tp->lhqLen = minpos + correction + 1;
    buf[minpos % bufsz].lhentry = newtc;
}


/*
 * remove the first segment from the lookahead queue and return the new
 * first segment
 */

static TC_STRUCT *lhqAdvance(TP_STRUCT * tp)
{
    int i;
    if (tp->lhqLen < 2) {
	tp->lhqStart = tp->lhqLen = 0;
	return 0;
    }
    i = (tp->lhqStart + 1) % tp->tcqBufSize;
    tp->lhqStart = i;
    tp->lhqLen--;
    return tp->tcqBuf[i].lhentry;
}

static void setBlendSize(TC_STRUCT * tc0, TC_STRUCT * tc1, double minstart)
{
    double bsize, lim, emag, vmax, amax;
    PmCartesian v0, v1, verr;

    /* don't blend with purely rotational motions */
    if ((tc0->seg.segType == PATHSEG_ROTATIONAL)
	|| (tc1->seg.segType == PATHSEG_ROTATIONAL)) {
	tc0->blend_size = 0.0;
	tc0->blend_vmax = 0.0;
	return;
    }

    /* blend at most half of each move */
    bsize = tc0->seg.length - minstart;
    lim = tc1->seg.length * 0.5;
    if (lim < bsize)
	bsize = lim;

    /* only blend an approximately straight portion of circles */
    if (tc0->seg.segType == PATHSEG_CIRCULAR) {
	lim = 0.4 * tc0->seg.length / (tc0->seg.circle.angle);
	if (lim < bsize)
	    bsize = lim;
    }
    if (tc1->seg.segType == PATHSEG_CIRCULAR) {
	lim = 0.4 * tc1->seg.length / (tc1->seg.circle.angle);
	if (lim < bsize)
	    bsize = lim;
    }

    /* only blend when deceleration in necessary */
    vmax = tc0->params.vMax;
    lim = tc1->params.vMax;
    if (lim < vmax)
	vmax = lim;
    amax = tc0->params.aMax;
    lim = tc1->params.aMax;
    if (lim < amax)
	amax = lim;
    if ((amax < 1e-10) || (vmax < 1e-10)) {
	tc0->blend_size = 0.0;
	tc0->blend_vmax = 0.0;
	return;
    }
    lim = vmax * vmax * 0.5 / amax;
    if (lim < bsize)
	bsize = lim;


    /* check corner angle and error limits */
    v0 = pathsegGetEndDirection(&tc0->seg);
    v1 = pathsegGetStartDirection(&tc1->seg);
    pmCartCartSub(v1, v0, &verr);
    emag = pmCartGetMag(verr);

    if (emag > 0.00001) {
	/* distance from programmed corner to blend path is bsize*emag/4 */
	/* this must be less than the error tolerance */
	lim = tc0->params.blend_error_max;
	if (tc1->params.blend_error_max < lim)
	    lim = tc1->params.blend_error_max;
	if (lim <= 0.0) {
	    tc0->blend_size = 0.0;
	    tc0->blend_vmax = 0.0;
	    return;
	}
	lim = lim * 4.0 / emag;
	if (lim < bsize)
	    bsize = lim;

	/* adjust max velocity through blend to limit acceleration
	 * total |deltaV| = v*emag
	 * time = bsize*2/v
	 * accel = v*v*emag/(bsize * 2)
	 * we want accel <= amax =>
	 * v*v <= amax * bsize * 2.0 / emag 
	 */
	lim = sqrt(amax * bsize * 2.0 / emag);
	/* NOTE! commanded tangential acceleration is NOT entirely
	 * orthogonal to this accel.  tpRuncycle must take that into account */
	if (lim < vmax) {
	    vmax = lim;
	}
    }

    if (bsize <= 1e-6) {
	bsize = 0.0;
	vmax = 0.0;
    }

    tc0->blend_size = bsize;
    tc0->blend_vmax = vmax;
}

static int tpAppend(TP_STRUCT * tp, PathSeg * seg)
{
    TC_STRUCT *tc, *prevtc;

    if (tp->tcqLen > 0)
	prevtc = tpQItem(tp, tp->tcqLen - 1);
    else {
	tp->currentArcPos = tp->nextArcPos = 0.0;
	prevtc = 0;
    }

    if (tp->tcqLen >= tp->tcqBufSize)
	return -1;

    tc = tpQItem(tp, tp->tcqLen);

    tc->seg = *seg;
    tc->params = tp->params;
    tc->id = tp->nextId;
    tc->blend_vmax = 0.0;
    tc->blend_size = 0.0;
    tc->energyLimits[0] = tc->energyLimits[1] = 0.0;

    if (tc->seg.segType == PATHSEG_ROTATIONAL) {
	if (tp->wMax > 0.0 || tc->params.vMax <= 0.0)
	    tc->params.vMax = tp->wMax;
	if (tp->wDotMax > 0.0 || tc->params.aMax <= 0.0)
	    tc->params.aMax = tp->wDotMax;
    }

    if (tc->params.blend_error_max < -1e-10) {
	tc->params.blend_error_max = tp->default_blend_tol_base +
	    tp->params.vMax * tp->default_blend_tol_relative;
    }

    tp->tcqLen++;

    tp->params.douts = 0;
    tp->params.doutstarts = 0;
    tp->params.doutends = 0;

    tp->goalPos = seg->end;
    tp->nextId++;

    if (prevtc) {
	double minstart = prevtc->seg.length * 0.5;
	double enerinc;
	if ((tp->tcqLen == 2) && (tp->currentArcPos >= minstart)) {
	    minstart = tp->currentArcPos;
	}
	setBlendSize(prevtc, tc, minstart);

	/* update energy */
	enerinc = 2.0 * (prevtc->seg.length -
			 prevtc->blend_size) * prevtc->params.aMax;
	tc->energyPos[0] = prevtc->energyPos[0] + enerinc;
	tc->energyPos[1] = prevtc->energyPos[1] + enerinc;

	tp->energyCount0++;
	tp->energyCount1++;
	if (tp->energyCount0 >= tp->energyCount1) {
	    if (tp->energyCount1 > tp->tcqLen) {
		tp->energyCount0 = 1;
		tc->energyPos[0] = 0.0;
	    }
	} else {
	    if (tp->energyCount0 > tp->tcqLen) {
		tp->energyCount1 = 1;
		tc->energyPos[1] = 0.0;
	    }
	}

	enerinc = tp->cycleTime * prevtc->params.aMax;
	enerinc = prevtc->blend_vmax * (prevtc->blend_vmax - enerinc);
	prevtc->energyLimits[0] = tc->energyPos[0] + enerinc;
	prevtc->energyLimits[1] = tc->energyPos[1] + enerinc;

	lhqAppend(tp, prevtc);
    } else {
	tp->energyCount0 = tp->energyCount1 = 1;
	tc->energyPos[0] = tc->energyPos[1] = 0.0;
	tc->energyLimits[0] = tc->energyLimits[1] = 0.0;
    }

    return 0;
}


int tpAddLine(TP_STRUCT * tp, EmcPose end)
{
    PathSeg seg;

    if (0 == tp) {
	return -1;
    }

    if (tp->aborting) {
	return -1;
    }

    pathsegInitLinear(&seg, tp->goalPos, end);
    return tpAppend(tp, &seg);
}

int tpAddCircle(TP_STRUCT * tp, EmcPose end,
		PmCartesian center, PmCartesian normal, int turn)
{
    PathSeg seg;

    if (0 == tp) {
	return -1;
    }

    if (tp->aborting) {
	return -1;
    }

    pathsegInitCircular(&seg, tp->goalPos, end, center, normal, turn);
    return tpAppend(tp, &seg);
}


static double getSafeV(double energy, double w, double a, double T)
{
    /* v(v+aT) = energy + w(w-aT) */
    /* v2 + vaT - w2 + waT - energy = 0 */
    double Bneg = -a * T;
    double C = w * a * T - w * w - energy;

    double det = Bneg * Bneg - 4.0 * C;
    double v;
    if (det <= 0.0)
	return w;
    v = (Bneg + sqrt(det)) * 0.5;
    if (v < w)
	v = w;
    return v;
}

typedef struct {
    int index;
    char val;
} DOutBuf;

static void dobInit(DOutBuf * buf)
{
    buf->index = -1;
    buf->val = 0;
}

static void dobPutStart(DOutBuf * buf, TC_STRUCT * tc)
{
    if (tc->params.douts) {
	if ((buf->index >= 0) && (buf->index != tc->params.doutIndex)) {
	    extDioWrite(buf->index, buf->val);
	}
	buf->index = tc->params.doutIndex;
	buf->val = tc->params.doutstarts;
    }
}

static void dobPutEnd(DOutBuf * buf, TC_STRUCT * tc)
{
    if (tc->params.douts) {
	if ((buf->index >= 0) && (buf->index != tc->params.doutIndex)) {
	    extDioWrite(buf->index, buf->val);
	}
	buf->index = tc->params.doutIndex;
	buf->val = tc->params.doutends;
    }
}

static void dobFlush(DOutBuf * buf)
{
    if (buf->index >= 0) {
	extDioWrite(buf->index, buf->val);
	buf->index = -1;
	buf->val = 0;
    }
}

int tpRunCycle(TP_STRUCT * tp)
{
    EmcPose newPos;
    TC_STRUCT *thisTc = 0, *nextTc = 0, *lhTc, *endTc;
    double arcPos, blendStart, blendSize;
    double curEnergy;
    double T, vtarget, vcurrent;
    double segVmax, segAmax, segDmax, segVlimit;
    DOutBuf dob;
    int doutPosNew;
    int countphase;

    if (0 == tp) {
	return -1;
    }

    dobInit(&dob);
    arcPos = tp->nextArcPos;
    countphase = (tp->energyCount0 >= tp->energyCount1 ? 0 : 1);

    /* advance to current segment */

    lhTc = lhqFirst(tp);

    for (;;) {

	if (tp->tcqLen <= 0) {
	    /* queue is empty */
	    thisTc = 0;
	    tp->doutPos = -1;
	    tp->currentArcPos = tp->nextArcPos = 0.0;
	    tp->activeDepth = 0;
	    tp->currentPos = tp->goalPos;
	    tp->execId = 0;
	    if (tp->aborting) {
		tp->tcqStart = tp->tcqLen = 0;
		tp->lhqStart = tp->lhqLen = 0;
		tp->aborting = 0;
		tpResume(tp);
	    }
	    dobFlush(&dob);
	    return (0);
	}

	thisTc = tpQItem(tp, 0);
	tp->execId = thisTc->id;

	if (arcPos < (thisTc->seg.length + thisTc->blend_size))
	    break;

	/* remove this segment */
	if (lhTc == thisTc) {
	    lhTc = lhqAdvance(tp);
	}
	if (thisTc->params.douts && (tp->doutPos <= 0))
	    dobPutEnd(&dob, thisTc);
	if (tp->doutPos >= 0)
	    tp->doutPos--;
	arcPos -= thisTc->seg.length;
	tp->tcqStart = (tp->tcqStart + 1) % tp->tcqBufSize;
	tp->tcqLen--;
    }

    blendSize = thisTc->blend_size;
    blendStart = thisTc->seg.length - blendSize;
    vcurrent = tp->nextArcPos - tp->currentArcPos;
    doutPosNew = tp->doutPos;

    if ((tp->tcqLen > 1) && (arcPos > blendStart)) {
	/* in blend */
	EmcPose p2;
	double bapos = (arcPos - blendStart) * 0.5;
	double iprop = bapos / blendSize;

	if (iprop >= 1.0) {
	    bapos = blendSize;
	    iprop = 1.0;
	}
	/* get path-blended position */
	nextTc = tpQItem(tp, 1);
	newPos = pathsegGetPos(&thisTc->seg, blendStart + bapos);
	p2 = pathsegGetPos(&nextTc->seg, bapos);
	pmCartCartSub(p2.tran, newPos.tran, &p2.tran);
	p2.a -= newPos.a;
	p2.b -= newPos.b;
	p2.c -= newPos.c;
	pmCartScalMAC(p2.tran, iprop, &newPos.tran);
	newPos.a += iprop * p2.a;
	newPos.b += iprop * p2.b;
	newPos.c += iprop * p2.c;

	curEnergy = nextTc->energyPos[countphase];
	if (lhTc == thisTc) {	/* lh at least to nexttc blend */
	    lhTc = lhqAdvance(tp);
	}

	/* maximum V for blended region */
	segVmax = thisTc->blend_vmax * tp->vScale;

	/* stuff that depends on which half of the blend we're in. */
	/* deceleration and acceleration are limited in the first and
	 * last halves, respectively, because blend accel isn't
	 * orthogonal to this accel */
	if (iprop >= 0.5) {
	    doutPosNew = 1;
	    segVlimit = nextTc->params.vLimit;
	    segDmax = nextTc->params.aMax;
	    segAmax = segDmax * 0.41;
	    curEnergy += (arcPos - thisTc->seg.length) * 2.0 * segDmax;
	} else {
	    doutPosNew = 0;
	    segVlimit = thisTc->params.vLimit;
	    segAmax = thisTc->params.aMax;
	    segDmax = segAmax * 0.41;
	}

	tp->activeDepth = 2;
    } else {
	/* not in blend */
	newPos = pathsegGetPos(&thisTc->seg, arcPos);
	segVmax = thisTc->params.vMax * tp->vScale;
	segAmax = segDmax = thisTc->params.aMax;
        segVlimit = thisTc->params.vLimit;
	doutPosNew = 0;
	tp->activeDepth = 1;
	curEnergy = thisTc->energyPos[countphase] + 2.0 * arcPos * segDmax;
    }

    if (segVmax > segVlimit)
        segVmax = segVlimit; // Clamp velocity if feed override is too high.

    /* set new point */
    tp->currentPos = newPos;
    tp->currentArcPos = tp->nextArcPos = arcPos;

    /* calculate highest safe velocity from energy restrictions */
    T = tp->cycleTime;
    endTc = tpQItem(tp, tp->tcqLen - 1);
    if (!lhTc) {
	lhTc = endTc;
    } else {
	double endlimit = endTc->energyPos[countphase];
	endlimit += 2.0 * endTc->seg.length;
	if (endlimit < lhTc->energyLimits[countphase])
	    lhTc = endTc;
    }

    if (tp->pausing || tp->aborting) {
	vtarget = 0.0;
	curEnergy = 0.0;
    } else {
	double minv;
	curEnergy = lhTc->energyPos[countphase] - curEnergy;
	curEnergy += 2.0 *
	    (lhTc->seg.length - lhTc->blend_size) * lhTc->params.aMax;
	vtarget =
	    getSafeV(curEnergy, lhTc->blend_vmax, lhTc->params.aMax, T);
	minv = segAmax * T * 0.25;
	if (vtarget < minv)
	    vtarget = minv;
	if (vtarget > segVmax)
	    vtarget = segVmax;
    }
    tp->curSegVtarget = vtarget;	/* just instrumentation */
    tp->curSegVmax = segVmax;	/* just instrumentation */

    /* convert to cycletime */
    vtarget *= T;
    T *= T;
    segAmax *= T;
    segDmax *= T;


    /* do our best to meet the target */
    if (vtarget < (vcurrent - segDmax))
	vtarget = vcurrent - segDmax;
    if (vtarget > (vcurrent + segAmax))
	vtarget = vcurrent + segAmax;

    if (vtarget > 0.0)
	tp->nextArcPos = arcPos + vtarget;

    /* check for abort done */
    if (tp->aborting && (vtarget == 0.0)) {
	if (tp->doutPos >= 0 && tp->tcqLen > tp->doutPos) {
	    dobPutEnd(&dob, tpQItem(tp, tp->doutPos));
	}
	tp->tcqStart = tp->tcqLen = 0;
	tp->lhqStart = tp->lhqLen = 0;
	tp->doutPos = -1;
	doutPosNew = -1;
	tp->goalPos = tp->currentPos;
	tp->activeDepth = 0;
	tp->aborting = 0;
	tp->execId = 0;
	tpResume(tp);
    }

    /* do douts */
    if (tp->doutPos < doutPosNew) {
	do {
	    dobPutEnd(&dob, tpQItem(tp, tp->doutPos));
	    tp->doutPos++;
	} while (tp->doutPos < doutPosNew);
	dobPutStart(&dob, tpQItem(tp, tp->doutPos));
    }
    dobFlush(&dob);
    return 0;
}

int tpPause(TP_STRUCT * tp)
{
    if (0 == tp) {
	return -1;
    }

    tp->pausing = 1;

    return 0;
}

int tpResume(TP_STRUCT * tp)
{
    if (0 == tp) {
	return -1;
    }

    tp->pausing = 0;

    return 0;
}

int tpAbort(TP_STRUCT * tp)
{
    if (0 == tp) {
	return -1;
    }

    if (!tp->aborting) {
	/* to abort, signal a pause and set our abort flag */
	tpPause(tp);
	tp->aborting = 1;
    }
    return 0;
}

EmcPose tpGetPos(TP_STRUCT * tp)
{
    EmcPose retval;
    if (0 == tp) {
	retval.tran.x = retval.tran.y = retval.tran.z = 0.0;
	retval.a = retval.b = retval.c = 0.0;
	return retval;
    }

    return tp->currentPos;
}

int tpIsDone(TP_STRUCT * tp)
{
    if (0 == tp) {
	return 0;
    }

    return (tp->tcqLen <= 0 ? 1 : 0);
}

/*
  tpIsPaused() returns 1 only when all active queued motions are paused.
  This is necessary so that abort clears the queue when motions have really
  stopped. If there are no queued motions, it returns 1 if the pausing
  flag is set. */
int tpIsPaused(TP_STRUCT * tp)
{
    if (0 == tp) {
	return 0;
    }

    if (0 == tp->tcqLen) {
	return tp->pausing;
    }

    return ((tp->pausing
	     && (tp->currentArcPos == tp->nextArcPos)) ? 1 : 0);
}

int tpQueueDepth(TP_STRUCT * tp)
{
    if (0 == tp) {
	return 0;
    }

    return tp->tcqLen;
}

int tpIsQueueFull(TP_STRUCT * tp, int margin)
{
    if (tp && (tp->tcqLen < (tp->tcqBufSize - margin)))
	return 0;
    else
	return 1;
}

int tpActiveDepth(TP_STRUCT * tp)
{
    if (0 == tp) {
	return 0;
    }

    return tp->activeDepth;
}

void tpPrint(TP_STRUCT * tp)
{
}

int tpSetAout(TP_STRUCT * tp,
	      unsigned char index, double start, double end)
{
    /* FIXME-- unimplemented due to large size required for doubles */
    return 0;
}
int tpSetDout(TP_STRUCT * tp, int index,
	      unsigned char start, unsigned char end)
{
    if (0 == tp) {
	return -1;
    }

    tp->params.doutIndex = index;
    tp->params.doutstarts = start;
    tp->params.doutends = end;
    tp->params.douts = 1;
    return 0;
}
