#ifndef TRAJPLAN_H_
#define TRAJPLAN_H_

/*
  Trajectory planning based on path blending.

  Derived from the original motion-blending trajectory planner
  for EMC.  The original algorithm filleted paths with quadratic
  splines by superimposing equal-timed portions of distinct
  trapezoidal-velocity-profile motions.  That was a remarkably simple
  way to generate a filleted path, but path errors dependend upon
  acceleration settings, and it didn't achieve the fastest traversals
  over multitudes of small segments.
  
  This version creates a similarly filleted path by superimposing
  and interpolating between equal-arc-length portions of path segments,
  and planning motion along the new path after the fact.  Path errors
  are limited by explicit tolerances, and lookahead allows for the 
  most efficient traversal.

  Author (BDI trajplan.h): Matt Timmermans
  
  Derived from EMC tp&tc (motion blending) by Fred Proctor, with
  	contributions from Will Shackleford and Paul Corner.
  	
  Derived source code should include all of these attributions
  
  License: GPL Version 2
  System: Linux
  
  Changes:

  25-Mar-2006 Matt Timmermans switched to fast unbounded lookahead based on
              "energy restrictions"
  18-Mar-2006 Matt Timmermans added many-segment lookahead based on
  	      "lookahead stop" segments.
  14-Mar-2006 Matt Timmermans rewrote most of this file to convert from
  	      motion blending to path blending

  EMC tp.h Change log:
  
  22-Mar-2004  FMP added 'now' flag to tpSetDout/Aout
  7-Dec-2001  FMP took hard-coded 1e-6 values out of tp.c and moved them
  here as EPSILON defines.
  17-Aug-2001  FMP added aout,dout motion IO
  13-Mar-2000 WPS added unused attribute to tp_h to avoid
  'defined but not used' compiler warning.
  8-Jun-1999  FMP added tpSetVlimit(), vLimit
  8-Mar-1999  FMP added tcSpace arg to tpCreate()
  18-Dec-1997  FMP took out C++ interface
  18-Dec-1997  FMP changed to PmPose
  18-Jul-1997  FMP added active depth
  16-Jul-1997  FMP added ids
  14-Jul-1997  FMP added C posemath changes (PM_POSE -> PmPose)
  24-Jun-1997  FMP added tpClear()
  13-Jun-1997  FMP added tpSetVscale(), tpIsPaused(), abort stuff
  16-Apr-1997  FMP created from C and C++ headers
*/

#include "posemath.h"
#include "pathseg.h"

/* ident tag */
#ifndef __GNUC__
#ifndef __attribute__
#define __attribute__(x)
#endif
#endif

/* Parameters for trajectory elements.  The current values established
 * through the API arestructure are copied to individual elements as
 * they are added to the queue. */

typedef struct {
    double vMax;		/* max velocity */
    double aMax;		/* max accel */
    double vLimit;		/* abs vel limit, including scale */
    double blend_error_max;
    int doutIndex;		/* index for dout value */
    unsigned char douts;	/* mask for douts to set */
    unsigned char doutstarts;	/* mask for dout start vals */
    unsigned char doutends;	/* mask for dout end vals */
} MotSegParams;

/* structure for individual trajectory elements */

typedef struct tag_TC_STRUCT {
    PathSeg seg;
    MotSegParams params;
    int id;			/* id for motion segment */

    /* let: w be the greatest velocity at which we can cross the
     *        a point on the path.
     *      e be the arc-distance remaining to the point.
     *      a be the rate at which we can declerate.
     *      T be the cycle time.
     *
     * Then the highest velocity we can travel at is v, such that
     *
     *      v(v+aT) = 2ae + w(w-aT)
     *
     * This differs from the real-world physics formula, because
     * we make allowances for time quantization.
     *
     * The 2ae, however, still represents the amount of "energy"
     * that can be shed by decelrating along the e-length of path.
     *
     * We keep track of the total energy in the path before each
     * segment, as measured from two different reference points.
     * (see TP_STRUCT.energyCount)
     */
    double energyPos[2];

    /* Everything before here is initialized with its final value
     * when the segment is added to the queue */

    /* these are initialized to 0 when the segment is added
     * to the queue, and updated to their final values when
     * the next segment is appended */
    double blend_size;
    double blend_vmax;

    /*
     * Every velocity restriction limits the velocity that can be
     * safely used at every preceding position.  The most restrictive
     * point is the one with the smallest 2ae + w(w-aT), which
     * doesn't depend on where you measure from.  We measure this
     * limit for each segment's blend start from the two energy
     * measurement positions
     */
    double energyLimits[2];

    /* This is NOT part of the path segment, and is not reset when
     * the segments is added to the path queue.  It is an entry
     * in the lookahead queue and is managed as part of the lookahead
     * process */

    struct tag_TC_STRUCT *lhentry;
} TC_STRUCT;

typedef struct {
    TC_STRUCT *tcqBuf;
    int tcqBufSize;
    int tcqStart;
    int tcqLen;

    /* lookahead queue consists of lhentry members in tcqBuf */
    /* It is the subset of the tc queue containing the segments with
     * lower energy limits than all the segments that follow them.
     * The first such segment is always the most restrictive segment
     * in the queue. */
    int lhqStart, lhqLen;

    int doutPos;		/* index of last segment with processed douts, or -1 if none */

    EmcPose currentPos;
    EmcPose goalPos;
    double currentArcPos;
    double nextArcPos;
    double curSegVmax;		/* reports current segment velocity limit */
    double curSegVtarget;	/* reports current target v */

    MotSegParams params;

    double cycleTime;
    double vScale;
    double wMax;		/* rotational velocity max  */
    double wDotMax;		/* rotational accelleration max */

    /* default blend tolerance, used if params.blend_error_max < 0.0 */
    double default_blend_tol_base;
    double default_blend_tol_relative;	/* additional velocity-based
					   tolerance in seconds of feed */
    int nextId;
    int execId;
    int activeDepth;		/* number of motions blending */
    int aborting;
    int pausing;

    /* the total energy we can shed
     * before reaching each segment is measured from two
     * distinct start positions, which are reset every now
     * and then.
     * For X in {0,1}, the last energyCountX segments added
     * to the queue have their energies measured from the
     * same start position in their energyPos[X] members.
     * Whenever both energyCounts exceed the queue length, the
     * longest one is reset to 0, and its measurements are
     * restarted at 0.0.  This ensures that at least one set of
     * measurments will always cover the whole queue, and neither
     * will accumulate measurements, and errors, indefinitely.
     */
    int energyCount0, energyCount1;
} TP_STRUCT;

extern int tpCreate(TP_STRUCT * tp, int _queueSize, TC_STRUCT * tcSpace);
extern int tpDelete(TP_STRUCT * tp);
extern int tpClear(TP_STRUCT * tp);
extern int tpInit(TP_STRUCT * tp);
extern int tpSetCycleTime(TP_STRUCT * tp, double secs);
extern int tpSetVmax(TP_STRUCT * tp, double vmax);
extern int tpSetWmax(TP_STRUCT * tp, double vmax);
extern int tpSetVlimit(TP_STRUCT * tp, double limit);
extern int tpSetVscale(TP_STRUCT * tp, double scale);	/* 0.0 .. large */
extern int tpSetAmax(TP_STRUCT * tp, double amax);
extern int tpSetWDotmax(TP_STRUCT * tp, double amax);
extern int tpSetId(TP_STRUCT * tp, int id);
extern int tpGetNextId(TP_STRUCT * tp);
extern int tpGetExecId(TP_STRUCT * tp);
extern int tpSetModeExactStop(TP_STRUCT * tp);
extern int tpSetModeDefaultBlend(TP_STRUCT * tp);
extern int tpSetModeLimitedBlend(TP_STRUCT * tp, double tolerance);
extern int tpSetDefaultBlendTolerance(TP_STRUCT * tp, double base,
				      double rel);
extern int tpSetPos(TP_STRUCT * tp, EmcPose pos);
extern int tpAddLine(TP_STRUCT * tp, EmcPose end);
extern int tpAddCircle(TP_STRUCT * tp, EmcPose end,
		       PmCartesian center, PmCartesian normal, int turn);
extern int tpRunCycle(TP_STRUCT * tp);
extern int tpPause(TP_STRUCT * tp);
extern int tpResume(TP_STRUCT * tp);
extern int tpAbort(TP_STRUCT * tp);
extern EmcPose tpGetPos(TP_STRUCT * tp);
extern int tpIsDone(TP_STRUCT * tp);
extern int tpIsPaused(TP_STRUCT * tp);
extern int tpQueueDepth(TP_STRUCT * tp);
extern int tpIsQueueFull(TP_STRUCT * tp, int margin);
extern int tpActiveDepth(TP_STRUCT * tp);
extern void tpPrint(TP_STRUCT * tp);
extern int tpSetAout(TP_STRUCT * tp, unsigned char index, double start,
		     double end);
extern int tpSetDout(TP_STRUCT * tp, int index, unsigned char start,
		     unsigned char end);


#endif				/*TRAJPLAN_H_ */
