/********************************************************************
* Description: interp_o_word.cc
*
*
* Author: Kenneth Lerman
* License: GPL Version 2
* System: Linux
*    
* Copyright 2005 All rights reserved.
*
********************************************************************/

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "rs274ngc.hh"
#include "rs274ngc_return.hh"
#include "interp_internal.hh"

//========================================================================
// Functions for control stuff (O-words)
//========================================================================

/************************************************************************/
/*
   In the long run, this function will use a hash table or other
   fast data structure
*/
int Interp::control_save_offset(int line,	//!< (o-word) line number
				block_pointer block,	//!< pointer to a block of RS274/NGC instructions
				setup_pointer settings)	//!< pointer to machine settings
{
    static char name[] = "control_save_offset";
    int index;

    if (control_find_oword(line, settings, &index) == INTERP_OK) {
	// already exists
	return INTERP_OK;
    }

    CHK((settings->oword_labels >= INTERP_OWORD_LABELS),
	NCE_TOO_MANY_OWORD_LABELS);


    index = settings->oword_labels++;

    settings->oword_offset[index].o_word = line;
    settings->oword_offset[index].type = block->o_type;
    settings->oword_offset[index].offset = block->offset;

    // the sequence number has already been bumped, so save
    // the proper value
    settings->oword_offset[index].sequence_number =
	settings->sequence_number - 1;

    return INTERP_OK;
}

#if 0
static int control_skip_to(int line,	//!< (o-word) line number
			   setup_pointer settings)	//!< pointer to machine settings
{
}
#endif
int Interp::control_find_oword(int line,	//!< (o-word) line number
			       setup_pointer settings,	//!< pointer to machine settings
			       int *o_index)	//!< the index of o-word (returned)
{
    static char name[] = "control_find_oword";
    int i;

    for (i = 0; i < settings->oword_labels; i++) {
	if (settings->oword_offset[i].o_word == line) {
	    *o_index = i;
	    return INTERP_OK;
	}
    }

    ERP(NCE_UNKNOWN_OWORD_NUMBER);
}

int Interp::control_back_to(int line,	//!< (o-word) line number
			    setup_pointer settings) //!<  pointer to machine settings
{
    static char name[] = "control_back_to";
    int i;

    for (i = 0; i < settings->oword_labels; i++) {
	if (settings->oword_offset[i].o_word == line) {
	    if (settings->file_pointer == NULL) {
		ERP(NCE_FILE_NOT_OPEN);
	    }
	    fseek(settings->file_pointer,
		  settings->oword_offset[i].offset, SEEK_SET);

	    settings->sequence_number =
		settings->oword_offset[i].sequence_number;

	    return INTERP_OK;
	}
    }
    ERP(NCE_UNKNOWN_OWORD_NUMBER);
}

/************************************************************************/
/* convert_control_functions

Returned Value: int (INTERP_OK)
Side effects:
   Changes the flow of control.

Called by: execute_block

Calls: control_skip_to
       control_back_to
       control_save_offset
*/

int Interp::convert_control_functions(block_pointer block,	//!< pointer to a block of RS274/NGC instructions
				      setup_pointer settings)	//!< pointer to machine settings
{
    static char name[] = "convert_control_functions";
    int status;
    int index;
    int i;

    // must skip if skipping
    if (settings->skipping_o && (settings->skipping_o != block->o_number)) {
	return INTERP_OK;
    }

    switch (block->o_type) {
    case O_none:
	// not an error because we use this to signal that we
	// are not evaluating functions
	break;

    case O_sub:
	// if the level is not zero, this is a call
	// not the definition
	// if we were skipping, no longer
	settings->skipping_o = 0;
	if (settings->call_level == 0) {
	    // a definition
	    CHK((settings->defining_sub == 1), NCE_NESTED_SUBROUTINE_DEFN);
	    CHP(control_save_offset(block->o_number, block, settings));
	    settings->skipping_o = block->o_number;
	    settings->defining_sub = 1;
	}
	break;
    case O_endsub:
	// if level is not zero, in a call
	// otherwise in a defn
	// if we were skipping, no longer
	settings->skipping_o = 0;
	if (settings->call_level != 0) {
	    // in a call -- must do a return
	    // restore old values of parameters
	    // restore file position from context

	    settings->call_level--;
#if 0
	    for (i = 0; i < INTERP_SUB_PARAMS; i++) {
		settings->parameters[i + INTERP_FIRST_SUBROUTINE_PARAM] =
		    settings->sub_context[settings->call_level].
		    saved_params[i];
	    }
#endif
	    if (settings->file_pointer == NULL) {
		ERP(NCE_FILE_NOT_OPEN);
	    }

	    fseek(settings->file_pointer,
		  settings->sub_context[settings->call_level].position,
		  SEEK_SET);

	    settings->sequence_number =
		settings->sub_context[settings->call_level].
		sequence_number;

	} else {
	    // a definition
	    CHK((settings->defining_sub != 1), NCE_NOT_IN_SUBROUTINE_DEFN);
	    // no longer skipping or defining
	    settings->skipping_o = 0;
	    settings->defining_sub = 0;
	}
	break;
    case O_call:
	// copy parameters from context
	// save old values of parameters
	// save current file position in context
	// if we were skipping, no longer
	settings->skipping_o = 0;
	if (settings->call_level >= INTERP_SUB_ROUTINE_LEVELS) {
	    ERP(NCE_TOO_MANY_SUBROUTINE_LEVELS);
	}
#if 0
	for (i = 0; i < INTERP_SUB_PARAMS; i++) {
	    settings->sub_context[settings->call_level].saved_params[i] =
		settings->parameters[i + INTERP_FIRST_SUBROUTINE_PARAM];

	    settings->parameters[i + INTERP_FIRST_SUBROUTINE_PARAM] =
		block->params[i];
	}
#endif
	    if (settings->file_pointer == NULL) {
		ERP(NCE_FILE_NOT_OPEN);
	    }
	    settings->sub_context[settings->call_level].position =
		ftell(settings->file_pointer);

	    settings->sub_context[settings->call_level].sequence_number
		= settings->sequence_number;


	settings->call_level++;

	CHP(control_back_to(block->o_number, settings));
	break;
    case O_do:
	// if we were skipping, no longer
	settings->skipping_o = 0;
	// save the loop point
	CHP(control_save_offset(block->o_number, block, settings));
	break;

    case O_while:
	// if we were skipping, no longer
	settings->skipping_o = 0;
	status = control_find_oword(block->o_number, settings, &index);

	// test if not already seen OR
	// if seen and this is a while (alternative is that it is a do)
	if ((status != INTERP_OK) ||
	    (settings->oword_offset[index].type == block->o_type)) {
	    // this is the beginning of a 'while' loop

	    // add it to the table if not already there
	    if (status != INTERP_OK) {
		CHP(control_save_offset(block->o_number, block, settings));
	    }
	    // test the condition

	    if (settings->test_value == 0.0) {
		// false
		// skip forward
		settings->skipping_o = block->o_number;
	    }
	} else {
	    // this is the end of a 'do'
	    // test the condition
	    if (settings->test_value != 0.0) {
		// true
		// loop on back
		CHP(control_back_to(block->o_number, settings));
	    }
	}

	break;

    case O_if:
	if (settings->test_value != 0.0) {
	    //true
	    settings->skipping_o = 0;
	    settings->executed_if = 1;
	} else {
	    //false
	    settings->skipping_o = block->o_number;
	    settings->executed_if = 0;
	}
	break;

    case O_elseif:
	if (settings->skipping_o != block->o_number) {
	    // we were not skipping -- start skipping
	    settings->skipping_o = block->o_number;
	    return INTERP_OK;
	}
	// we were skipping
	// but were we ever not skipping
	if (settings->executed_if) {
	    // we have already executed, keep on skipping
	    //settings->skipping_0 = block->o_number;
	    return INTERP_OK;
	}

	if (settings->test_value != 0.0) {
	    //true -- start executing
	    settings->skipping_o = 0;
	    settings->executed_if = 1;
	}
	break;

    case O_else:
	// were we ever not skipping
	if (settings->executed_if) {
	    // we have already executed, skip
	    settings->skipping_o = block->o_number;
	    return INTERP_OK;
	}

	if (settings->skipping_o == block->o_number) {
	    // we were skipping so stop skipping
	    settings->executed_if = 1;
	    settings->skipping_o = 0;
	}
	break;

    case O_endif:
	// stop skipping if we were
	settings->skipping_o = 0;
	// the KEY -- outside if clearly must have executed
	// or this would not have executed
	settings->executed_if = 1;
	break;

    case O_break:
	// start skipping
	settings->skipping_o = block->o_number;
	//settings->doing_break = 1;
	break;

    case O_continue:

	// if already skipping, do nothing
	if (settings->skipping_o == block->o_number) {
	    return INTERP_OK;
	}
	// start skipping
	settings->skipping_o = block->o_number;
	settings->doing_continue = 1;
	break;

    case O_endwhile:
	// end of a while loop
	if (settings->skipping_o == block->o_number) {
	    // we were skipping, so this is the end
	    settings->skipping_o = 0;

	    if (settings->doing_continue) {
		settings->doing_continue = 0;

		// loop on back
		CHP(control_back_to(block->o_number, settings));
	    } else {
		// not doing continue, we are done
		return INTERP_OK;
	    }
	} else {
	    // loop on back
	    CHP(control_back_to(block->o_number, settings));
	}

	break;

    case O_return:
	if (settings->call_level == 0) {
	    // this is in the definition
	    break;
	}
	// if we were skipping, no longer
	settings->skipping_o = 0;

	// in a call -- must do a return
	// restore old values of parameters
	// restore file position from context

	settings->call_level--;
#if 0
	for (i = 0; i < INTERP_SUB_PARAMS; i++) {
	    settings->parameters[i] =
		settings->sub_context[settings->call_level].
		saved_params[i];
	}
#endif
	if (settings->file_pointer == NULL) {
	    ERP(NCE_FILE_NOT_OPEN);
	}

	fseek(settings->file_pointer,
	      settings->sub_context[settings->call_level].position,
	      SEEK_SET);

	settings->sequence_number =
	    settings->sub_context[settings->call_level].sequence_number;

	break;

    default:
	//!!!KL should probably be an error
	break;
    }

    return INTERP_OK;
}

//========================================================================
// End of functions for control stuff (O-words)
//========================================================================
