/*****************************************************************************
 *
 *      NAME:  Probe Engine
 *
 *      PURPOSE:  Standalone app to perform automated probing with EMC
 *                Reads a probing initialization file for
 *                probing regions and other setup information.
 *
 *      INPUT:    probe.ini, command line
 *
 *
 *      OUTPUT:
 *
 *      RESTRICTIONS:  Most of this code is borrowed from emcsh.cc
 *                      (tcl extensions)
 *
 *      REVISION: DLH - ORIGINAL VERSION   1/20/02 11:14AM
        WPS -- made some silly "style" changes, ( I think it is 
clearer if only pointers are set or compared with NULL characters are set
or compared with 0 , and increments
and storing data in pointers is done on seperate lines.)
 *
 *****************************************************************************
 *
 *
 ****************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <signal.h>
#include <math.h>
#include <ctype.h>
#include <time.h>
#include <sys/wait.h>
#include <sys/types.h>
#include <unistd.h>		// exec(),fork(),getopt() etc. -- added by WPS  3-Mar-2004
#define TRUE 1
#define FALSE 0

#include "rcs.hh"		// etime()
#include "posemath.h"		// PM_POSE
#include "emc.hh"		// EMC NML
#include "emcglb.h"		// EMC_NMLFILE, TRAJ_MAX_VELOCITY, etc.
#include "emccfg.h"		// DEFAULT_TRAJ_MAX_VELOCITY
#include "inifile.hh"		// INIFILE
#include "hmi_base.hh"
using namespace Hmi;

// ident tag
#ifndef __GNUC__
#ifndef __attribute__
#define __attribute__(x)
#endif
#endif

static char __attribute__ ((unused)) ident[] =
    "$Id: emcprobe.cc,v 1.4 2005/08/08 22:56:39 paul Exp $";

/*

  THe following comments are from the original emcsh

  Using emcsh:

  emcsh {<filename>} {-- -ini <ini file>}

  With filename, it opens NML buffers to the EMC, runs the script, closes
  the buffers, and quits.

  With -- -ini <inifile>, uses inifile instead of generic.ini. Note that
  the two dashes prevents Tcl from looking at the remaining args, which
  would otherwise trigger a Tcl error that it doesn't understand what
  -ini means.

  Without filename, it runs interactively.

  The files (or manual input) are Tcl scripts, extended with EMC-specific
  commands. These commands are all prefixed with "emc_", which makes them
  somewhat inconvenient for typing but avoids name conflicts, e.g., open.

  Some commands take 0 or more arguments. 0 arguments means they return
  the associated value; the argument would be to set the value.

  Commands are sent to the EMC, and control resumes immediately. You can
  call a timed wait until the command got there, or a timed wait until the
  command completed, or not wait at all.

  EMC commands:

  emc_plat
  Returns the platform for which this was compiled, e.g., linux_2_0_36

  emc_ini <var> <section>
  Returns the string value of <var> in section <section>, in EMC_INIFILE

  emc_set_wait none | received | done
  Set the wait for commands to return to be right away (none), after the
  command was sent and received (received), or after the command was
  done (done).

  emc_wait received | done
  Force a wait for the previous command to be received, or done. This lets
  you wait in the event that "emc_set_wait none" is in effect.

  emc_set_timeout <timeout>
  Set the timeout for commands to return to <timeout>, in seconds. Timeout
  is a real number. If it's <= 0.0, it means wait forever. Default is 0.0,
  wait forever.

  emc_update (none) | none | auto
  With no arg, forces an update of the EMC status. With "none", doesn't
  cause an automatic update of status with other emc_ words. With "auto",
  makes emc_ words automatically update status before they return values.

  emc_error
  Returns the current EMC error string, or "ok" if no error.

  emc_operator_display
  Returns the current EMC operator display string, or "ok" if none.

  emc_operator_text
  Returns the current EMC operator text string, or "ok" if none.

  emc_time
  Returns the time, in seconds, from the start of the epoch. This starting
  time depends on the platform.

  emc_estop_in
  Returns the estop input setting as "on" or "off".

  emc_machine (none) | on | off
  With no arg, returns the machine setting as "on" or "off". Otherwise,
  sends a machine on or off command.

  emc_mode (none) | manual | auto | mdi
  With no arg, returns the mode setting as "manual", "auto", or "mdi".
  Otherwise, sends a mode manual, auto, or mdi command.

  emc_mist (none) | on | off
  With no arg, returns the mist setting as "on" or "off". Otherwise,
  sends a mist on or off command.

  emc_flood (none) | on | off
  With no arg, returns the flood setting as "on" or "off". Otherwise,
  sends a flood on or off command.

  emc_lube (none) | on | off
  With no arg, returns the lubricant pump setting as "on" or "off".
  Otherwise, sends a lube on or off command.

  emc_lube_level
  Returns the lubricant level sensor reading as "ok" or "low".

  emc_spindle (none) | forward | reverse | increase | decrease | constant | off
  With no arg, returns the value of the spindle state as "forward",
  "reverse", "increase", "decrease", or "off". With arg, sends the spindle
  command. Note that "increase" and "decrease" will cause a speed change in
  the corresponding direction until a "constant" command is sent.

  emc_brake (none) | on | off
  With no arg, returns the brake setting. Otherwise sets the brake.

  emc_tool
  Returns the id of the currently loaded tool

  emc_tool_offset
  Returns the currently applied tool length offset

  emc_load_tool_table <file>
  Loads the tool table specified by <file>

  emc_home 0 | 1 | 2 | ...
  Homes the indicated axis.

  emc_jog_stop 0 | 1 | 2 | ...
  Stop the axis jog

  emc_jog 0 | 1 | 2 | ... <speed>
  Jog the indicated axis at <speed>; sign of speed is direction

  emc_jog_incr 0 | 1 | 2 | ... <speed> <incr>
  Jog the indicated axis by increment <incr> at the <speed>; sign of
  speed is direction

  emc_feed_override {<percent>}
  With no args, returns the current feed override, as a percent. With
  argument, set the feed override to be the percent value

  emc_abs_cmd_pos 0 | 1 | ...
  Returns double obj containing the XYZ-SXYZ commanded pos in abs coords,
  at given index, 0 = X, etc.

  emc_abs_act_pos
  Returns double objs containing the XYZ-SXYZ actual pos in abs coords

  emc_rel_cmd_pos 0 | 1 | ...
  Returns double obj containing the XYZ-SXYZ commanded pos in rel coords,
  at given index, 0 = X, etc., including tool length offset

  emc_rel_act_pos
  Returns double objs containing the XYZ-SXYZ actual pos in rel coords,
  including tool length offset

  emc_pos_offset X | Y | Z | R | P | W
  Returns the position offset associated with the world coordinate provided

  emc_joint_limit 0 | 1 | ...
  Returns "ok", "minsoft", "minhard", "maxsoft", "maxhard"

  emc_joint_fault 0 | 1 | ...
  Returns "ok" or "fault"

  emc_joint_homed 0 | 1 | ...
  Returns "homed", "not"

  emc_mdi <string>
  Sends the <string> as an MDI command

  emc_task_plan_init
  Initializes the program interpreter

  emc_open <filename>
  Opens the named file

  emc_run {<start line>}
  Without start line, runs the opened program from the beginning. With
  start line, runs from that line. A start line of -1 runs in verify mode.

  emc_pause
  Pause program execution

  emc_resume
  Resume program execution

  emc_step
  Step the program one line

  emc_program
  Returns the name of the currently opened program, or "none"

  emc_program_line
  Returns the currently executing line of the program

  emc_program_status
  Returns "idle", "running", or "paused"

  emc_program_codes
  Returns the string for the currently active program codes

  emc_log_open <file> <type> <size> <skip> {<axis>}
  Open a log into file named <file>. <type> is one of axis_pos, axes_inpos,
  axes_outpos, axis_vel, axes_ferror, traj_pos, traj_vel, traj_acc,
  pos_voltage. <size> is size in entries. <skip> is how many to skip.
  <axis> pertains only to axis_pos and axis_vel types.

  emc_log_start
  emc_log_stop
  Starts or stops logging into the log, which must be already opened
  with emc_log_open.

  emc_log_close
  Close the log and dump contents to the file specified in emc_log_open.

  emc_probe_index
  Which wire is the probe on or which bit of digital IO to use? (No args
  gets it, one arg sets it.)

  emc_probe_polarity
  Value to look for for probe tripped? (0 args gets it, one arg sets it.)

  emc_probe_clear
  Clear the probe tripped flag.

  emc_probe_tripped
  Has the probe been tripped since the last clear.

  emc_probe_value
  Value of current probe signal. (read-only)

  emc_probe
  Move toward a certain location. If the probe is tripped on the way stop
  motion, record the position and raise the probe tripped flag.
*/

#define MAX_REGIONS 100
#define MAX_PATH 512
#define MAX_LINE 512

typedef enum {
    cylinder,
    ball,
    point
} tiptype;

typedef struct probe_setup {
    int Quiet_Mode;
    char Ini_Filename[MAX_PATH];
    int Units;
    int Significant_Digits;
    tiptype Tip_Style;
    double Tip_Diameter;
    int Probe_Index;
    int Probe_Polarity;
    int Num_Regions;
    char Post_Process_Cmd[MAX_PATH];
};

typedef struct probe_region {
    int XYZ_Output_Gcode;
    char XYZ_Gcode_File[MAX_PATH];
    double XYZ_Gcode_ShiftX;
    double XYZ_Gcode_ShiftY;
    double XYZ_Gcode_ShiftZ;
    double XYZ_Start_X;
    double XYZ_Start_Y;
    double XYZ_Start_Z;
    double XYZ_End_X;
    double XYZ_End_Y;
    double XYZ_End_Z;
    double XYZ_Min_X;
    double XYZ_Min_Y;
    double XYZ_Min_Z;
    double XYZ_Max_X;
    double XYZ_Max_Y;
    double XYZ_Max_Z;
    double XYZ_Step_Size_X;
    double XYZ_Step_Size_Y;
    int XYZ_Probe_Axis;
    int XYZ_Probe_Method;
    int XYZ_Probe_Algorythm;
    double XYZ_Probe_Traverse_Speed;
    double XYZ_Probe_Z_Speed;
};

struct probe_setup Probe_Setup;
struct probe_region Probe_Region[MAX_REGIONS];

static void thisQuit(void)
{
    nmlQuit();
    exit(0);
}

static void sigQuit(int sig)
{
    thisQuit();
}

/**************************PROBE ENGINE CODE BEGINS HERE *******************/
/**************************PROBE ENGINE CODE BEGINS HERE *******************/
/*****************************************************************************
 *
 *      NAME: Some generic string routines
 *
 *      PURPOSE:
 *
 *      INPUT:
 *
 *
 *      OUTPUT:
 *
 *      RESTRICTIONS:
 *
 *      REVISION: DLH - ORIGINAL VERSION
 *
 *****************************************************************************
 *
 *
 ****************************************************************************/

char *strlwr(char *s)
{
    unsigned int i;

    for (i = 0; i < strlen(s); i++) {
	s[i] = tolower(s[i]);
    }

    return s;
}

char *strupr(char *s)
{
    unsigned int i;

    for (i = 0; i < strlen(s); i++) {
	s[i] = toupper(s[i]);
    }
    return s;
}

void stripit(char *instring)
{
    unsigned int loop1;

    for (loop1 = 0; loop1 <= strlen(instring); loop1++) {
	if ((instring[loop1] < 0x20)) {
	    instring[loop1] = 0x00;
	    return;
	}
    }
}

/*****************************************************************************
 *
 *      NAME:
 *
 *      PURPOSE: get a single line from file and throw away the CR
 *
 *      INPUT:
 *
 *
 *      OUTPUT:
 *
 *      RESTRICTIONS:
 *
 *      REVISION: DLH - ORIGINAL VERSION
 *
 *****************************************************************************
 *
 *
 ****************************************************************************/
char *fgets_nocr(char *s, int n, FILE * stream)
{

    fgets(s, n, stream);

    if (s != NULL)
	stripit(s);

    strlwr(s);

    return (s);
}

/*****************************************************************************
 *
 *      NAME:  Get_Ini_Setting
 *
 *      PURPOSE:  Retrieve string ini file setting
 *
 *      INPUT:
 *
 *
 *      OUTPUT:
 *
 *      RESTRICTIONS:
 *
 *      REVISION: DLH - ORIGINAL VERSION
 *
 *****************************************************************************
 *
 *
 ****************************************************************************/
int Get_Ini_Setting(char *isection, char *ikey, char *destvalue,
		    char *iinifilename)
{
    FILE *src_inifile;
    char inpstr[LINELEN];
    char *valueread;
    char section[LINELEN];
    char key[LINELEN];
    char inifilename[LINELEN];

    strcpy(section, isection);
    strcpy(key, ikey);
    strcpy(inifilename, iinifilename);
    strlwr(section);
    strlwr(key);
    strlwr(inifilename);

    memset(destvalue, 0x00, sizeof(destvalue));

    if ((src_inifile = fopen(inifilename, "rt")) == NULL) {
	printf("Get_Ini_Setting - Can't open inifile\n");
	return (0);
    }
    //get us to the section of interest

    do {
	memset(inpstr, 0x00, sizeof(inpstr));
	fgets_nocr(inpstr, LINELEN - 1, src_inifile);	//keep reading lines..
    }
    while ((!feof(src_inifile)) &&	//until we reach end of file..
	   (strstr(inpstr, section) == NULL)	// or find a match
	);

    if (feof(src_inifile))	//we reached end & could not find the section...
    {
	fclose(src_inifile);
	printf("Get_Ini_Setting - Section %s not found in %s\n", section,
	       inifilename);
	return (FALSE);
    }
    //now look in this section for keystring
    do {
	memset(inpstr, 0x00, sizeof(inpstr));
	fgets_nocr(inpstr, LINELEN - 1, src_inifile);

	if (strstr(inpstr, "[") != NULL)	//we hit the next section without finding key..
	    break;

	if (strstr(inpstr, key) != NULL) {
	    strtok(inpstr, "=");
	    valueread = strtok(NULL, "=");
	    strcpy(destvalue, valueread);
	    fclose(src_inifile);
	    return (1);
	}
    }
    while (!feof(src_inifile));

    fclose(src_inifile);
    printf
	("Get_Ini_Setting - Key Item %s not found in section %s of file %s\n",
	 key, section, inifilename);
    return (0);
}

/*****************************************************************************
 *
 *      NAME: read_probe_ini
 *
 *      PURPOSE: We read the file, if there is an reading error we
 *               tell why & return zero,  else return 1
 *
 *      INPUT:  The init file format as described in the following
 *               Notes:
 *                Items are case insensitive
 *                Comments are not permitted (yet)
 *
 *         [Probe Setup]
 *         Units=in
 *         Significant_Digits=3
 *         Tip_Style=Cylindrical
 *         Tip_Diameter=0.25
 *         Num_Regions=1
 *         Post_Process_Cmd=ls
 *
 *         [Region 1]
 *         Region_Type=XYZ
 *         XYZ_Start_X=0.00
 *         XYZ_Start_Y=0.00
 *         XYZ_Start_Z=0.00
 *         XYZ_Min_X=0.00
 *         XYZ_Min_Y=0.00
 *         XYZ_Min_Z=0.00
 *         XYZ_Max_X=1.00
 *         XYZ_Max_Y=1.00
 *         XYZ_Max_Z=0.50
 *         XYZ_Step_Size_X=0.10
 *         XYZ_Step_Size_Y=0.10
 *         XYZ_Probe_Axis=X
 *         XYZ_Probe_Method=0
 *         XYZ_Probe_Algorythm=0
 *         XYZ_Probe_Traverse_Speed=5
 *         XYZ_Probe_Z_Speed=5
 *
 *
 *
 *      OUTPUT:
 *
 *      RESTRICTIONS:
 *
 *      REVISION: DLH - ORIGINAL VERSION
 *
 *****************************************************************************
 *
 *
 ****************************************************************************/

static int read_probe_ini(char *inifilename)
{
    char readstring[MAX_LINE];
    char region_sec_str[MAX_LINE];
    int i;

    printf("Reading initialization file %s\n", inifilename);

    if (Get_Ini_Setting("Probe Setup", "Units", readstring, inifilename)) {
	if (strstr(readstring, "in") != NULL) {
	    Probe_Setup.Units = 1;
	} else {
	    Probe_Setup.Units = 0;
	}
    } else {
	return (0);
    }

    if (Get_Ini_Setting
	("Probe Setup", "Significant_Digits", readstring, inifilename)) {
	Probe_Setup.Significant_Digits = atoi(readstring);
    } else {
	return (0);
    }

    //the tip stuff is read in but not used by emcprobe at present
    if (Get_Ini_Setting
	("Probe Setup", "Tip_Style", readstring, inifilename)) {
	if (strstr(readstring, "cyl") != NULL) {
	    Probe_Setup.Tip_Style = cylinder;
	} else if (strstr(readstring, "ball") != NULL) {
	    Probe_Setup.Tip_Style = ball;
	} else if (strstr(readstring, "point") != NULL) {
	    Probe_Setup.Tip_Style = point;
	} else {
	    Probe_Setup.Tip_Style = cylinder;
	}
    } else {
	return (0);
    }

    if (Get_Ini_Setting
	("Probe Setup", "Tip_Diameter", readstring, inifilename)) {
	Probe_Setup.Tip_Diameter = atof(readstring);
    } else {
	return (0);
    }

    if (Get_Ini_Setting
	("Probe Setup", "Probe_Index", readstring, inifilename)) {
	Probe_Setup.Probe_Index = atoi(readstring);
    } else {
	return (0);
    }

    if (Get_Ini_Setting
	("Probe Setup", "Probe_Polarity", readstring, inifilename)) {
	Probe_Setup.Probe_Polarity = atoi(readstring);
    } else {
	return (0);
    }

    if (Get_Ini_Setting
	("Probe Setup", "Num_Regions", readstring, inifilename)) {
	Probe_Setup.Num_Regions = atoi(readstring);
    } else {
	return (0);
    }

    if (Get_Ini_Setting
	("Probe Setup", "Post_Process_Cmd", readstring, inifilename)) {
	strcpy(Probe_Setup.Post_Process_Cmd, readstring);
    } else {
	return (0);
    }

    for (i = 1; i <= Probe_Setup.Num_Regions; i++) {
	sprintf(region_sec_str, "Region %d", i);

	if (Get_Ini_Setting
	    (region_sec_str, "XYZ_Output_Gcode", readstring,
	     inifilename)) {
	    Probe_Region[i].XYZ_Output_Gcode = atoi(readstring);
	    Probe_Region[i].XYZ_Gcode_ShiftX = 0.0;
	    Probe_Region[i].XYZ_Gcode_ShiftX = 0.0;
	    Probe_Region[i].XYZ_Gcode_ShiftX = 0.0;
	    strcpy(Probe_Region[i].XYZ_Gcode_File, "");

	    if (Probe_Region[i].XYZ_Output_Gcode) {
		if (Get_Ini_Setting
		    (region_sec_str, "XYZ_Gcode_File", readstring,
		     inifilename)) {
		    strcpy(Probe_Region[i].XYZ_Gcode_File, readstring);
		} else {
		    sprintf(Probe_Region[i].XYZ_Gcode_File, "region%d.ngc",
			    i);
		}
		if (Get_Ini_Setting
		    (region_sec_str, "XYZ_Gcode_ShiftX", readstring,
		     inifilename)) {
		    Probe_Region[i].XYZ_Gcode_ShiftX = atof(readstring);
		}
		if (Get_Ini_Setting
		    (region_sec_str, "XYZ_Gcode_ShiftY", readstring,
		     inifilename)) {
		    Probe_Region[i].XYZ_Gcode_ShiftY = atof(readstring);
		}
		if (Get_Ini_Setting
		    (region_sec_str, "XYZ_Gcode_ShiftZ", readstring,
		     inifilename)) {
		    Probe_Region[i].XYZ_Gcode_ShiftZ = atof(readstring);
		}
	    }
	}

	if (Get_Ini_Setting
	    (region_sec_str, "XYZ_Start_X", readstring, inifilename)) {
	    Probe_Region[i].XYZ_Start_X = atof(readstring);
	} else {
	    return (0);
	}

	if (Get_Ini_Setting
	    (region_sec_str, "XYZ_Start_Y", readstring, inifilename)) {
	    Probe_Region[i].XYZ_Start_Y = atof(readstring);
	} else {
	    return (0);
	}

	if (Get_Ini_Setting
	    (region_sec_str, "XYZ_Start_Z", readstring, inifilename)) {
	    Probe_Region[i].XYZ_Start_Z = atof(readstring);
	} else {
	    return (0);
	}

	if (Get_Ini_Setting
	    (region_sec_str, "XYZ_End_X", readstring, inifilename)) {
	    Probe_Region[i].XYZ_End_X = atof(readstring);
	} else {
	    return (0);
	}

	if (Get_Ini_Setting
	    (region_sec_str, "XYZ_End_Y", readstring, inifilename)) {
	    Probe_Region[i].XYZ_End_Y = atof(readstring);
	} else {
	    return (0);
	}

	if (Get_Ini_Setting
	    (region_sec_str, "XYZ_End_Z", readstring, inifilename)) {
	    Probe_Region[i].XYZ_End_Z = atof(readstring);
	} else {
	    return (0);
	}

	if (Get_Ini_Setting
	    (region_sec_str, "XYZ_Min_X", readstring, inifilename)) {
	    Probe_Region[i].XYZ_Min_X = atof(readstring);
	} else {
	    return (0);
	}

	if (Get_Ini_Setting
	    (region_sec_str, "XYZ_Min_Y", readstring, inifilename)) {
	    Probe_Region[i].XYZ_Min_Y = atof(readstring);
	} else {
	    return (0);
	}

	if (Get_Ini_Setting
	    (region_sec_str, "XYZ_Min_Z", readstring, inifilename)) {
	    Probe_Region[i].XYZ_Min_Z = atof(readstring);
	} else {
	    return (0);
	}

	if (Get_Ini_Setting
	    (region_sec_str, "XYZ_Max_X", readstring, inifilename)) {
	    Probe_Region[i].XYZ_Max_X = atof(readstring);
	} else {
	    return (0);
	}

	if (Get_Ini_Setting
	    (region_sec_str, "XYZ_Max_Y", readstring, inifilename)) {
	    Probe_Region[i].XYZ_Max_Y = atof(readstring);
	} else {
	    return (0);
	}

	if (Get_Ini_Setting
	    (region_sec_str, "XYZ_Max_Z", readstring, inifilename)) {
	    Probe_Region[i].XYZ_Max_Z = atof(readstring);
	} else {
	    return (0);
	}

	if (Get_Ini_Setting
	    (region_sec_str, "XYZ_Step_Size_X", readstring, inifilename)) {
	    Probe_Region[i].XYZ_Step_Size_X = atof(readstring);
	} else {
	    return (0);
	}

	if (Get_Ini_Setting
	    (region_sec_str, "XYZ_Step_Size_Y", readstring, inifilename)) {
	    Probe_Region[i].XYZ_Step_Size_Y = atof(readstring);
	} else {
	    return (0);
	}

	if (Get_Ini_Setting
	    (region_sec_str, "XYZ_Probe_Axis", readstring, inifilename)) {
	    Probe_Region[i].XYZ_Probe_Axis = 0;
	    if (strstr(readstring, "x") != NULL) {
		Probe_Region[i].XYZ_Probe_Axis = 0;
	    } else if (strstr(readstring, "y") != NULL) {
		Probe_Region[i].XYZ_Probe_Axis = 1;
	    }
	} else {
	    return (0);
	}

	if (Get_Ini_Setting
	    (region_sec_str, "XYZ_Probe_Method", readstring,
	     inifilename)) {
	    Probe_Region[i].XYZ_Probe_Method = atoi(readstring);
	} else {
	    return (0);
	}

	if (Get_Ini_Setting
	    (region_sec_str, "XYZ_Probe_Algorythm", readstring,
	     inifilename)) {
	    Probe_Region[i].XYZ_Probe_Algorythm = atoi(readstring);
	} else {
	    return (0);
	}

	if (Get_Ini_Setting
	    (region_sec_str, "XYZ_Probe_Traverse_Speed", readstring,
	     inifilename)) {
	    Probe_Region[i].XYZ_Probe_Traverse_Speed = atof(readstring);
	} else {
	    return (0);
	}

	if (Get_Ini_Setting
	    (region_sec_str, "XYZ_Probe_Z_Speed", readstring,
	     inifilename)) {
	    Probe_Region[i].XYZ_Probe_Z_Speed = atof(readstring);
	} else {
	    return (0);
	}
    }
    return (1);
}

/*****************************************************************************
 *
 *      NAME: do_region
 *
 *      PURPOSE:  Probes one region
 *
 *      INPUT:
 *
 *
 *      OUTPUT:
 *
 *      RESTRICTIONS:
 *
 *      REVISION: DLH - ORIGINAL VERSION   1/20/02 11:09AM
 *
 *****************************************************************************
 *
 *
 ****************************************************************************/
int do_region(int i, FILE * dest_gcode_file)
{
    char mdistring[LINELEN];
    double step_x;
    double step_y;
    double xmove = 0;
    double ymove = 0;

    int num_x_steps;
    int num_y_steps;
    int j, k, num_j_steps, num_k_steps;
    int toggle_direction;

    time_t now;

    step_x = Probe_Region[i].XYZ_Step_Size_X;	//default our step size
    step_y = Probe_Region[i].XYZ_Step_Size_Y;

    if (Probe_Region[i].XYZ_Min_X > Probe_Region[i].XYZ_Max_X) { //unless limits are reversed
	step_x = -Probe_Region[i].XYZ_Step_Size_X;
    }
    if (Probe_Region[i].XYZ_Min_Y > Probe_Region[i].XYZ_Max_Y) {
	step_y = -Probe_Region[i].XYZ_Step_Size_Y;
    }
    //checkme - round up
    num_x_steps =
	(int) (fabs(Probe_Region[i].XYZ_Max_X - Probe_Region[i].XYZ_Min_X)
	       / fabs(Probe_Region[i].XYZ_Step_Size_X));
    num_y_steps =
	(int) (fabs(Probe_Region[i].XYZ_Max_Y - Probe_Region[i].XYZ_Min_Y)
	       / fabs(Probe_Region[i].XYZ_Step_Size_Y));

    if (Probe_Region[i].XYZ_Probe_Axis == 1) {
	num_j_steps = num_x_steps + 1;
	num_k_steps = num_y_steps + 1;
    } else {
	num_j_steps = num_y_steps + 1;
	num_k_steps = num_x_steps + 1;
    }

    //Lets set the speed
    sprintf(mdistring, "F%-4.4f",
	    Probe_Region[i].XYZ_Probe_Traverse_Speed);

    //Send the string
    printf("%s\n", mdistring);
    if (sendMdiCmd(mdistring) < 0) {
	printf("Setting traverse speed failed.\n");
	return (0);
    }

    //before we go to the probe region let's go up to the max z height first
    sprintf(mdistring, "G00 Z%-4.4f", Probe_Region[i].XYZ_Max_Z);

    //Send the string
    printf("%s\n", mdistring);
    if (sendMdiCmd(mdistring) < 0) {
	printf("Traverse to new Z failed.\n");
	return (0);
    }

    toggle_direction = 0;

    for (j = 0; j <= num_j_steps; j++) {	//could be x or y!
	if (Probe_Region[i].XYZ_Probe_Axis == 1) {	//if it's x
	    xmove = (j * step_x) + Probe_Region[i].XYZ_Min_X;
	} else {
	    ymove = (j * step_y) + Probe_Region[i].XYZ_Min_Y;
	}

	if (toggle_direction == 1) {
	    toggle_direction = 0;
	} else {
	    toggle_direction = 1;
	}

	for (k = 0; k <= num_k_steps; k++) {	//could be x or y!
	    if (Probe_Region[i].XYZ_Probe_Axis == 1) {
		if ((Probe_Region[i].XYZ_Probe_Method == 1)
		    && (toggle_direction == 0)) {
		    //ymove=Probe_Region[i].XYZ_Min_Y - ((num_k_steps-k)*step_y);
		    ymove =
			Probe_Region[i].XYZ_Min_Y +
			(step_y * num_k_steps) - (k * step_y);
		    printf("toggle=%d method=%d ymove=%-4.4f \n",
			   Probe_Region[i].XYZ_Probe_Method,
			   toggle_direction, ymove);
		} else {
		    ymove = (k * step_y) + Probe_Region[i].XYZ_Min_Y;
		}
	    } else {
		if ((Probe_Region[i].XYZ_Probe_Method == 1)
		    && (toggle_direction == 0)) {
		    //xmove=Probe_Region[i].XYZ_Min_X - ((num_k_steps-k)*step_x);
		    xmove =
			Probe_Region[i].XYZ_Min_X +
			(step_x * num_k_steps) - (k * step_x);
		    printf("toggle=%d method=%d xmove=%-4.4f \n",
			   Probe_Region[i].XYZ_Probe_Method,
			   toggle_direction, xmove);

		} else {
		    xmove = (k * step_x) + Probe_Region[i].XYZ_Min_X;
		}
	    }

	    //rapid move to xy z traverse
	    //compose the rapid move string
	    sprintf(mdistring, "G00 X%-4.4f Y%-4.4f Z%-4.4f",
		    xmove, ymove, Probe_Region[i].XYZ_Max_Z);

	    //Send the string
	    printf("%s\n", mdistring);
	    if (sendMdiCmd(mdistring) < 0) {
		printf("Traverse to new XYZ failed.\n");
		return (0);
	    }
	    //clear tripped flag
	    if (sendClearProbeTrippedFlag() < 0) {
		printf("Can't clear the tripped flag.\n");
		return (0);
	    }
	    //probe "down" to min_z
	    if (sendProbe(xmove, ymove, Probe_Region[i].XYZ_Min_Z) < 0) {
		printf("Can't perform the actual probe command.\n");
		return (0);
	    }
	    //loop check status tripped
	    //need timeout check here, also never probe to a machine limit!
	    //Hopefully a single probe will never take longer than 30 sec

	    now = time((time_t *) NULL);
	    while (!emcStatus->motion.traj.probe_tripped &
		   ((now + 30) < time((time_t *) NULL))
		) {
		updateStatus();
	    }

	    //print aquired to screen
	    printf("Probed:  X%-4.4f Y%-4.4f Z%-4.4f\n",
		   emcStatus->motion.traj.probedPosition.tran.x,
		   emcStatus->motion.traj.probedPosition.tran.y,
		   emcStatus->motion.traj.probedPosition.tran.z);

	    //and append into the ini file
	    //fprintf(dest_inifile,"X%-4.4f Y%-4.4f Z%-4.4f\n",
	    //emcStatus->motion.traj.probedPosition.tran.x,
	    //emcStatus->motion.traj.probedPosition.tran.y,
	    //emcStatus->motion.traj.probedPosition.tran.z);

	    //and optionally write the g-code
	    if (Probe_Region[i].XYZ_Output_Gcode) {
		if ((k == 0) && (Probe_Region[i].XYZ_Probe_Method == 0)) {
		    //if first one
		    //bring us up and back around to the the next row in the unidirectional case
		    fprintf(dest_gcode_file, "G00 Z%-4.4f\n",
			    Probe_Region[i].XYZ_Max_Z);
		    fprintf(dest_gcode_file, "G00 X%-4.4f Y%-4.4f\n",
			    (emcStatus->motion.traj.probedPosition.tran.x +
			     Probe_Region[i].XYZ_Gcode_ShiftX),
			    (emcStatus->motion.traj.probedPosition.tran.y +
			     Probe_Region[i].XYZ_Gcode_ShiftY));
		    fprintf(dest_gcode_file, "G01 Z%-4.4f\n",
			    (emcStatus->motion.traj.probedPosition.tran.z +
			     Probe_Region[i].XYZ_Gcode_ShiftZ));

		    //i.e., we start each pass with something like this:
		    //G00 Z0.5
		    //G00 X3.1955 Y3.7646
		    //G01 F1.0 Z0.25
		} else {
		    fprintf(dest_gcode_file,
			    "G01 X%-4.4f Y%-4.4f Z%-4.4f\n",
			    (emcStatus->motion.traj.probedPosition.tran.x +
			     Probe_Region[i].XYZ_Gcode_ShiftX),
			    (emcStatus->motion.traj.probedPosition.tran.y +
			     Probe_Region[i].XYZ_Gcode_ShiftY),
			    (emcStatus->motion.traj.probedPosition.tran.
			     z) + Probe_Region[i].XYZ_Gcode_ShiftZ);
		}
	    }
	    //back to where we started
	    if (sendMdiCmd(mdistring) < 0) {
		printf("Traverse back to new XYZ failed.\n");
		return (0);
	    }

	}
    }
    return 0;
}

/*****************************************************************************
 *
 *      NAME:
 *
 *      PURPOSE:  Iterate through the probe regions defined, calling do_region
 *
 *      INPUT:
 *
 *
 *      OUTPUT:
 *
 *      RESTRICTIONS:
 *
 *      REVISION: DLH - ORIGINAL VERSION   1/20/02 11:09AM
 *
 *****************************************************************************
 *
 *
 ****************************************************************************/
int do_regions()
{

    int i;
    //FILE           *dest_inifile;
    FILE *dest_gcode_file = NULL;

    //if ( ( dest_inifile = fopen(Probe_Setup.Ini_Filename, "a" ) ) == NULL )
    //{
    //    printf( "Can't open inifile to record data.\n" );
    //    return ( 0 );
    //}

    printf("Starting Probe Regions...\n");

    //ensure we are in MDI mode
    printf("Entering MDI mode\n");
    if (sendMdi() < 0) {
	printf("MDI mode- Error\n");
	return (0);
    }

    for (i = 1; i <= Probe_Setup.Num_Regions; i++) {
	if (Probe_Region[i].XYZ_Output_Gcode) {
	    if ((dest_gcode_file =
		 fopen(Probe_Region[i].XYZ_Gcode_File, "w")) == NULL) {
		printf("Can't create G-Code file to record data.\n");
	    }
	}
	//fprintf(dest_inifile,"\n[Aquired_Region%d]\n",i);
	do_region(i, dest_gcode_file);
	if (Probe_Region[i].XYZ_Output_Gcode) {
	    fclose(dest_gcode_file);
	}
    }

    //fclose(dest_inifile);
    return 0;
}

/*****************************************************************************
 *
 *      NAME:  machine_setup
 *
 *      PURPOSE: Initialize the machine before performing probe operation
 *
 *      INPUT:
 *
 *
 *      OUTPUT:
 *
 *      RESTRICTIONS:
 *
 *      REVISION: DLH - ORIGINAL VERSION   1/20/02 11:09AM
 *
 *****************************************************************************
 *
 *
 ****************************************************************************/
static int machine_setup()
{
    char mdistring[LINELEN];
    emcWaitType = EMC_WAIT_DONE;

    printf("Machine on\n");
    if (sendMachineOn() < 0) {
	printf("Machine on - Error\n");
	return (0);
    }

    printf("Estop reset\n");
    if (sendEstopReset() < 0) {
	printf("Estop reset - Error\n");
	return (0);
    }

    printf("Probe Index Setup\n");
    //fixme
    if (sendSetProbeIndex(Probe_Setup.Probe_Index) < 0) {
	printf("Probe Index Setup - Error\n");
	return (0);
    }

    printf("Probe Polarity Setup\n");
    //fixme
    if (sendSetProbePolarity(Probe_Setup.Probe_Polarity) < 0) {
	printf("Probe Polarity Setup - Error\n");
	return (0);
    }

    printf("MDI mode\n");
    if (sendMdi() < 0) {
	printf("MDI mode- Error\n");
	return (0);
    }

    printf("Setting Units\n");

    if (Probe_Setup.Units == 1) {
	strcpy(mdistring, "G20");
    } else {
	strcpy(mdistring, "G21");
    }

    if (sendMdiCmd(mdistring) < 0) {
	printf("Setting Units - Error\n");
	return (0);
    }

    printf("Manual mode\n");
    if (sendManual() < 0) {
	printf("MDI mode- Error\n");
	return (0);
    }
    /*
       printf("Homing axis 0\n");
       if ( sendHome(0))
       {
       printf("Homing axis 0 - Error\n");
       return (0);
       }

       printf("Homing axis 1\n");
       if ( sendHome(1))
       {
       printf("Homing axis 1 - Error\n");
       return (0);
       }

       printf("Homing axis 2\n");
       if ( sendHome(2))
       {
       printf("Homing axis 2 - Error\n");
       return (0);
       }
     */
    printf("Machine Init completed\n");
    return (1);
}

/*****************************************************************************
 *
 *      NAME:   abort_probe
 *
 *      PURPOSE:  Theis is where we go when something fails
 *
 *      INPUT:
 *
 *
 *      OUTPUT:
 *
 *      RESTRICTIONS:
 *
 *      REVISION: DLH - ORIGINAL VERSION   1/20/02 11:09AM
 *
 *****************************************************************************
 *
 *
 ****************************************************************************/
void abort_probe(void)
{
    printf("Aborting emcprobe\n");
    thisQuit();
    exit(1);
}

/*****************************************************************************
 *
 *      NAME: parse
 *
 *      PURPOSE: pases the command line to spawn into a form suitable for exec
 *
 *      INPUT:
 *
 *
 *      OUTPUT:
 *
 *      RESTRICTIONS:
 *
 *      REVISION: Borrowed from A. D. Marshall C tutorial
 *
 *****************************************************************************
 *
 *
 ****************************************************************************/
void parse(char *buf, char **args)
{
    while (*buf != 0) {
	/*
	 * Strip whitespace.  Use nulls, so
	 * that the previous argument is terminated
	 * automatically.
	 */
	while ((*buf == ' ') || (*buf == '\t')) {
	    *buf = 0;
	    buf++;
	}

	/*
	 * Save the argument.
	 */
	*args = buf;
	args++;

	/*
	 * Skip over the argument.
	 */
	while ((*buf != 0) && (*buf != ' ') && (*buf != '\t'))
	    buf++;
    }

    *args = 0;
}

/*****************************************************************************
 *
 *      NAME: execute
 *
 *      PURPOSE: spawn a child process and execute the program
 *
 *      INPUT:
 *
 *
 *      OUTPUT:
 *
 *      RESTRICTIONS:
 * 
 *         NOTE: The execv() vnd execvp versions of execl() are useful when the 
 *         number  of  arguments is unknown in advance;                         
 *         The arguments to execv() and execvp()  are the name                  
 *         of the file to be executed and a vector of strings  contain-         
 *         ing  the  arguments.   The last argument string must be fol-         
 *         lowed by a 0 pointer.                                                
 *                                                                              
 *         execlp() and execvp() are called with the same arguments  as         
 *         execl()  and  execv(),  but duplicate the shell's actions in         
 *         searching for an executable file in a list  of  directories.         
 *         The directory list is obtained from the environment. 
 * 
 *         NOTE: perror() produces a short  error  message  on  the  standard 
 *         error describing the last error encountered during a call to       
 *         a system or library function.                                      
 * 
 *              
 *      REVISION: Borrowed from A. D. Marshall C tutorial
 *
 *****************************************************************************
 *
 *
 ****************************************************************************/
void execute(char **args)
{
    int pid, status;

    // Get a child process.

    if ((pid = fork()) < 0) {
	perror("fork");
	exit(1);
    }

    // The child executes the code inside the if.

    if (pid == 0) {
	execvp(*args, args);
	perror(*args);
	exit(1);

    }
    // The parent executes the wait.     
    while (wait(&status) != pid);
}

/*****************************************************************************
 *
 *      NAME:
 *
 *      PURPOSE:
 *
 *      INPUT:
 *
 *
 *      OUTPUT:
 *
 *      RESTRICTIONS:
 *
 *      REVISION: DLH - ORIGINAL VERSION   1/20/02 11:09AM
 *
 *****************************************************************************
 *
 *
 ****************************************************************************/
int main(int argc, char *argv[])
{
    int c;
    char *args[64];

    while ((c = getopt(argc, argv, "qf:")) != EOF) {
	switch (c) {
	case 'f':
	    strcpy(Probe_Setup.Ini_Filename, optarg);
	    printf("Ini=%s\n", Probe_Setup.Ini_Filename);
	    break;

	case 'q':
	    Probe_Setup.Quiet_Mode = 1;
	    break;

	default:
	    printf("Invalid option specified: %c\n", optopt);
	    printf("%s [-f inifile] [-q]\n", argv[0]);
	    return 1;
	    break;
	}
    }

    // init NML
    printf("EMC Probe v0.1 Started...\n");

    printf("Attempting to connect to EMC via NML...\n");

    if (0 != tryNml("xemc")) {
	rcs_print_error("Can't connect to EMC\n");
	thisQuit();
	exit(1);
    }
    printf("NML Connection made..\n");

    // get current serial number, and save it for restoring when we quit
    // so as not to interfere with real operator interface
    updateStatus();
    emcCommandSerialNumber = emcStatus->echo_serial_number;
    saveEmcCommandSerialNumber = emcStatus->echo_serial_number;

    // attach our quit function to SIGINT
    signal(SIGINT, sigQuit);

    read_probe_ini(Probe_Setup.Ini_Filename);
    machine_setup();
    do_regions();

    //If given, Run the post process command

    parse(Probe_Setup.Post_Process_Cmd, args);
    execute(args);

    thisQuit();
    return (0);
}
