/*
  feedslider.cc

  Demonstrates an external process modifying the feedrate override.

  Run the EMC on one machine. Then run feedslider on another networked
  machine, and enter percent numbers at the prompt, e.g., 0, 100.

  The .nml file is hard-coded here as generic.nml. This process connects as
  "remote", so you should have something like this in your emc.nml file:

  B emcCommand            SHMEM   <actual name of machine running EMC> ...
  B emcStatus             SHMEM   <actual name of machine running EMC> ...
  B emcError              SHMEM   <actual name of machine running EMC> ...
  ...
  P remote emcCommand REMOTE localhost W 0 10.0 0 10
  P remote emcStatus  REMOTE localhost R 0 10.0 0 10
  P remote emcError   REMOTE localhost R 0 10.0 0 10
*/

#include <stdio.h>
#include "rcs.hh"
#include "emc.hh"

int main(int argc, char *argv[])
{
  char buffer[LINELEN];
  RCS_CMD_CHANNEL * emc_cmd_buffer;
  RCS_STAT_CHANNEL * emc_stat_buffer;
  EMC_STAT * emc_stat_ptr;
  EMC_NULL null_msg;
  int saved_serial_number;
  EMC_TRAJ_SET_SCALE feed_override_msg;
  double percent;

  emc_cmd_buffer = new RCS_CMD_CHANNEL(emcFormat, "emcCommand", "remote", "generic.nml");
  if (0 == emc_cmd_buffer ||
      ! emc_cmd_buffer->valid()) {
    fprintf(stderr, "can't connect to EMC\n");
    return 1;
  }

  emc_stat_buffer = new RCS_STAT_CHANNEL(emcFormat, "emcStatus", "remote", "generic.nml");
  if (0 == emc_stat_buffer ||
      ! emc_stat_buffer->valid() ||
      EMC_STAT_TYPE != emc_stat_buffer->peek()) {
    fprintf(stderr, "can't connect to EMC\n");
    return 1;
  }
  emc_stat_ptr = (EMC_STAT *) emc_stat_buffer->get_address();
  saved_serial_number = emc_stat_ptr->echo_serial_number;
  feed_override_msg.serial_number = saved_serial_number;

  while (! feof(stdin)) {
    printf("> ");
    fflush(stdout);
    if (NULL == fgets(buffer, LINELEN, stdin)) {
      break;
    }
    buffer[LINELEN - 1] = 0;	// make sure it's a null-terminated string

    if (1 != sscanf(buffer, "%lf", &percent)) {
      printf("?\n");
      continue;
    }
    feed_override_msg.scale = percent / 100.0; // scale to 0..1 range
    feed_override_msg.serial_number++; // flag a new message
    emc_cmd_buffer->write(feed_override_msg); // and write it
  }

  /* now restore old serial number, so the real EMC GUI won't possibly
     have its next command ignored */
  null_msg.serial_number = saved_serial_number;
  emc_cmd_buffer->write(null_msg);

  delete emc_stat_buffer;
  delete emc_cmd_buffer;

  return 0;
}

