/*
  inimot.c

  Initialize motion controller from INI file

  Modification history:

   2-Sep-2003 WPS - This file should explicitly include stdarg.h before
using va_list etc instead of depending on getting it indirectly.
  22-Jul-1999  FMP looped on [TRAJ] AXES from ini file, rather than
  EMCMOT_MAX_AXES
  15-Oct-1998  FMP changed emcLogError to emcOperatorError per change
  in emc.hh
  2-Sep-1998  FMP added local implementation of emcLogError()
  6-Jul-1998 FMP changed BASE_ADDRESS to SHMEM_BASE_ADDRESS
  1-Jul-1998  FMP added -ini, -shm, -base
  15-Jun-1998  FMP added emcStatus global since logging functions in
  interface code reference this
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>

#include "emc.hh"
#include "emcmot.h"
#include "usrmotintf.h"
#include "initraj.hh"
#include "iniaxis.hh"
#include "emcglb.h"             // EMC_INIFILE
#include "emcmotglb.h"          // SHMEM_KEY, SHMEM_BASE_ADDRESS
#include "emcmotcfg.h"          // EMCMOT_MAX_AXIS

// global EMC status-- some interface functions fill this in
static EMC_STAT emcStatusStruct;
EMC_STAT * emcStatus = 0;       // declared in emc.hh

// implementation of EMC error logger
int emcOperatorError(int id, const char *fmt, ...)
{
  char errorString[LINELEN];
  va_list ap;

  // prepend error code
  sprintf(errorString, "[%d] ", id);

  // append error string
  va_start(ap, fmt);
  vsprintf(&errorString[strlen(errorString)], fmt, ap);
  va_end(ap);

  // force a NULL at the end for safety
  errorString[LINELEN-1] = 0;

  // print it
  printf("%s\n", errorString);

  return 0;
}

static int numAxes = 0;         // [TRAJ] AXES from ini file, for looping
#define DEFAULT_NUM_AXES 3      // how many to try if it's not in ini file

static int iniLoad(const char *filename)
{
  Inifile inifile;
  const char *inistring;

  // open it
  if (-1 == inifile.open(filename)) {
    return -1;
  }

  if (NULL != (inistring = inifile.find("AXES", "TRAJ"))) {
    if (1 != sscanf(inistring, "%d", &numAxes)) {
      numAxes = DEFAULT_NUM_AXES;
    }
  }
  else {
    // not found, use default
    numAxes = DEFAULT_NUM_AXES;
  }

  // close it
  inifile.close();

  return 0;
}

int main(int argc, char *argv[])
{
  int retval = 0;
  int axis;

  emcStatus = &emcStatusStruct;

  // process command line args
  if (0 != emcGetArgs(argc, argv))
    {
      printf("error in argument list\n");
      exit(1);
    }

  // get some ini file params for us
  if (0 != iniLoad(EMC_INIFILE)) {
    printf("can't read INI file\n");
    exit(1);
  }

  // check ini file for emcmot params
  if (0 != usrmotIniLoad(EMC_INIFILE)) {
    printf("can't read motion control INI file\n");
    exit(1);
  }

  if (0 != usrmotInit()) {
    printf("can't initialize motion control interface\n");
    exit(1);
  }

  for (axis = 0; axis < numAxes; axis++) {
    printf("initing axis %d...", axis);
    fflush(stdout);
    if (0 == iniAxis(axis, EMC_INIFILE)) {
      printf("done\n");
    }
    else {
      printf("error\n");
      retval = 1;
    }
  }

  // initialize parameters from ini file
  printf("initing traj...");
  fflush(stdout);
  if (0 == iniTraj(EMC_INIFILE)) {
    printf("done\n");
  }
  else {
    printf("error\n");
    retval = 1;
  }

  usrmotExit();

  exit(retval);
}
