/********************************************************************
* Description: posemath.c
*    A minimal set of math functions primarily used in realtime code
*    to calculate position & velocity. There is scope for further
*    streamlining both here and in the code that uses this stuff.
*    \todo posemath: Restructure so that additional axes can be added
*       at a later date without resorting to crufty hacks - Support for
*       U, V, & W is desirable.
*
* Author:
* License: GPL Version 2
* System: Linux
*    
* Copyright (c) 2004 All rights reserved.
*
* Last change: 
* $Revision: 1.2 $
* $Author: paul $
* $Date: 2006/02/08 13:54:43 $
********************************************************************/
#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <linux/types.h>
#ifdef realtime
#ifdef __attribute_used__
#undef __attribute_used__
#endif
#ifdef __attribute_pure__
#undef __attribute_pure__
#endif
#endif
#include <sys/cdefs.h>
#include <float.h>
#include <math.h>
#include "posemath.h"

/* global error number */
int pmErrno = 0;

/* fuzz checker */
#define IS_FUZZ(a,fuzz) (fabs(a) < (fuzz) ? 1 : 0)

double pmSqrt(double x)
{
    if (x > 0.0) {
	return sqrt(x);
    }

    if (x > SQRT_FUZZ) {
	return 0.0;
    }

    return 0.0;
}

/* PmCartesian functions */

int pmCartCartCompare(PmCartesian v1, PmCartesian v2)
{
    if (fabs(v1.x - v2.x) >= V_FUZZ ||
	fabs(v1.y - v2.y) >= V_FUZZ || fabs(v1.z - v2.z) >= V_FUZZ) {
	return 0;
    }

    return 1;
}

int pmCartCartDot(PmCartesian v1, PmCartesian v2, double *d)
{
    *d = v1.x * v2.x + v1.y * v2.y + v1.z * v2.z;

    return pmErrno = 0;
}

int pmCartCartCross(PmCartesian v1, PmCartesian v2, PmCartesian * vout)
{
    vout->x = v1.y * v2.z - v1.z * v2.y;
    vout->y = v1.z * v2.x - v1.x * v2.z;
    vout->z = v1.x * v2.y - v1.y * v2.x;

    return pmErrno = 0;
}

int pmCartMag(PmCartesian v, double *d)
{
    *d = pmSqrt(pmSq(v.x) + pmSq(v.y) + pmSq(v.z));

    return pmErrno = 0;
}

double pmCartGetMag(PmCartesian v)
{
    return pmSqrt(v.x * v.x + v.y * v.y + v.z * v.z);
}


int pmCartCartDisp(PmCartesian v1, PmCartesian v2, double *d)
{
    *d = pmSqrt(pmSq(v2.x - v1.x) + pmSq(v2.y - v1.y) + pmSq(v2.z - v1.z));

    return pmErrno = 0;
}

int pmCartCartAdd(PmCartesian v1, PmCartesian v2, PmCartesian * vout)
{
    vout->x = v1.x + v2.x;
    vout->y = v1.y + v2.y;
    vout->z = v1.z + v2.z;

    return pmErrno = 0;
}

int pmCartCartInterpolate(PmCartesian v1, PmCartesian v2, double t,
			  PmCartesian * vout)
{
    vout->x = v1.x + (v2.x-v1.x)*t;
    vout->y = v1.y + (v2.y-v1.y)*t;
    vout->z = v1.z + (v2.z-v1.z)*t;

    return pmErrno = 0;
}

int pmCartCartSub(PmCartesian v1, PmCartesian v2, PmCartesian * vout)
{
    vout->x = v1.x - v2.x;
    vout->y = v1.y - v2.y;
    vout->z = v1.z - v2.z;

    return pmErrno = 0;
}

int pmCartScalMult(PmCartesian v1, double d, PmCartesian * vout)
{
    vout->x = v1.x * d;
    vout->y = v1.y * d;
    vout->z = v1.z * d;

    return pmErrno = 0;
}

int pmCartScalMAC(PmCartesian v1, double d, PmCartesian * vout)
{
    vout->x += v1.x * d;
    vout->y += v1.y * d;
    vout->z += v1.z * d;

    return pmErrno = 0;
}


int pmCartScalDiv(PmCartesian v1, double d, PmCartesian * vout)
{
    if (d == 0.0) {
	vout->x = DBL_MAX;
	vout->y = DBL_MAX;
	vout->z = DBL_MAX;

	return pmErrno = PM_DIV_ERR;
    }

    vout->x = v1.x / d;
    vout->y = v1.y / d;
    vout->z = v1.z / d;

    return pmErrno = 0;
}

int pmCartNeg(PmCartesian v1, PmCartesian * vout)
{
    vout->x = -v1.x;
    vout->y = -v1.y;
    vout->z = -v1.z;

    return pmErrno = 0;
}

int pmCartInv(PmCartesian v1, PmCartesian * vout)
{
    double size_sq = pmSq(v1.x) + pmSq(v1.y) + pmSq(v1.z);

    if (size_sq == 0.0) {
	vout->x = DBL_MAX;
	vout->y = DBL_MAX;
	vout->z = DBL_MAX;

	return pmErrno = PM_NORM_ERR;
    }

    vout->x = v1.x / size_sq;
    vout->y = v1.y / size_sq;
    vout->z = v1.z / size_sq;

    return pmErrno = 0;
}

// This used to be called pmCartNorm.
int pmCartUnit(PmCartesian v, PmCartesian * vout)
{
    double size = pmSqrt(pmSq(v.x) + pmSq(v.y) + pmSq(v.z));

    if (size == 0.0) {
	vout->x = DBL_MAX;
	vout->y = DBL_MAX;
	vout->z = DBL_MAX;

	return pmErrno = PM_NORM_ERR;
    }

    vout->x = v.x / size;
    vout->y = v.y / size;
    vout->z = v.z / size;

    return pmErrno = 0;
}

int pmCartIsNorm(PmCartesian v)
{
    return pmSqrt(pmSq(v.x) + pmSq(v.y) + pmSq(v.z)) - 1.0 <
	V_FUZZ ? 1 : 0;
}

int pmCartCartProj(PmCartesian v1, PmCartesian v2, PmCartesian * vout)
{
    int r1, r2, r3;
    double d;

    r1 = pmCartUnit(v2, &v2);
    r2 = pmCartCartDot(v1, v2, &d);
    r3 = pmCartScalMult(v2, d, vout);

    return pmErrno = r1 || r2 || r3 ? PM_NORM_ERR : 0;
}

int pmCartPlaneProj(PmCartesian v, PmCartesian normal, PmCartesian * vout)
{
    int r1, r2;
    PmCartesian par;

    r1 = pmCartCartProj(v, normal, &par);
    r2 = pmCartCartSub(v, par, vout);

    return pmErrno = r1 || r2 ? PM_NORM_ERR : 0;
}

int pmRpyMatConvert(PmRpy rpy, PmRotationMatrix * m)
{
    double sa, sb, sg;
    double ca, cb, cg;

    sa = sin(rpy.y);
    sb = sin(rpy.p);
    sg = sin(rpy.r);

    ca = cos(rpy.y);
    cb = cos(rpy.p);
    cg = cos(rpy.r);

    m->x.x = ca * cb;
    m->y.x = ca * sb * sg - sa * cg;
    m->z.x = ca * sb * cg + sa * sg;

    m->x.y = sa * cb;
    m->y.y = sa * sb * sg + ca * cg;
    m->z.y = sa * sb * cg - ca * sg;

    m->x.z = -sb;
    m->y.z = cb * sg;
    m->z.z = cb * cg;

    return pmErrno = 0;
}

int pmMatCartMult(PmRotationMatrix m, PmCartesian v, PmCartesian * vout)
{
    vout->x = m.x.x * v.x + m.y.x * v.y + m.z.x * v.z;
    vout->y = m.x.y * v.x + m.y.y * v.y + m.z.y * v.z;
    vout->z = m.x.z * v.x + m.y.z * v.y + m.z.z * v.z;

    return pmErrno = 0;
}

int pmLineInit(PmLine * line, PmCartesian start, PmCartesian end)
{
    line->start = start;
    line->end = end;
    return 0;
}

/*
  pmCircleInit() takes the defining parameters of a generalized circle
  and sticks them in the structure. It also computes the radius and vectors
  in the plane that are useful for other functions and that don't need
  to be recomputed every time.

  Note that the end can be placed arbitrarily, resulting in a combination of
  spiral and helical motion. There is an overconstraint between the start,
  center, and normal vector: the center vector and start vector are assumed
  to be in the plane defined by the normal vector. If this is not true, then
  it will be made true by moving the center vector onto the plane.
  */
int pmCircleInit(PmCircle * circle,
		 PmCartesian start, PmCartesian end,
		 PmCartesian center, PmCartesian normal, int turn)
{
    double dot;
    PmCartesian rEnd;
    PmCartesian v;
    double d;
    int r1;

    /* adjust center */
    pmCartCartSub(start, center, &v);
    r1 = pmCartCartProj(v, normal, &v);
    if (PM_NORM_ERR == r1) {
	/* bad normal vector-- abort */
	return -1;
    }
    pmCartCartAdd(v, center, &circle->center);

    /* normalize and redirect normal vector based on turns. If turn is less
       than 0, point normal vector in other direction and make turn positive, 
       -1 -> 0, -2 -> 1, etc. */
    pmCartUnit(normal, &circle->normal);
    if (turn < 0) {
	turn = -1 - turn;
	pmCartScalMult(circle->normal, -1.0, &circle->normal);
    }

    /* radius */
    pmCartCartDisp(start, circle->center, &circle->radius);

    /* vector in plane of circle from center to start, magnitude radius */
    pmCartCartSub(start, circle->center, &circle->rTan);
    /* vector in plane of circle perpendicular to rTan, magnitude radius */
    pmCartCartCross(circle->normal, circle->rTan, &circle->rPerp);

    /* do rHelix, rEnd */
    pmCartCartSub(end, circle->center, &circle->rHelix);
    pmCartPlaneProj(circle->rHelix, circle->normal, &rEnd);
    circle->spiral_proportion =
	(pmCartGetMag(rEnd) - circle->radius) / circle->radius;
    pmCartCartSub(circle->rHelix, rEnd, &circle->rHelix);
    pmCartUnit(rEnd, &rEnd);
    pmCartScalMult(rEnd, circle->radius, &rEnd);

    /* Patch for error spiral end same as spiral center */
    pmCartMag(rEnd, &d);
    if (d == 0.0) {
	pmCartScalMult(circle->normal, DOUBLE_FUZZ, &v);
	pmCartCartAdd(rEnd, v, &rEnd);
    }
    /* end patch 03-mar-1999 Dirk Maij */

    /* angle */
    pmCartCartDot(circle->rTan, rEnd, &dot);
    dot = dot / (circle->radius * circle->radius);
    if (dot > 1.0) {
	circle->angle = 0.0;
    } else if (dot < -1.0) {
	circle->angle = M_PIl;
    } else {
	circle->angle = acos(dot);
    }
    /* now angle is in range 0..PI . Check if cross is antiparallel to
       normal. If so, true angle is between PI..2PI. Need to subtract from
       2PI. */
    pmCartCartCross(circle->rTan, rEnd, &v);
    pmCartCartDot(v, circle->normal, &d);
    if (d < 0.0) {
	circle->angle = (M_PIl * 2.0) - circle->angle;
    }

    if (circle->angle > -(CIRCLE_FUZZ) && circle->angle < (CIRCLE_FUZZ)) {
	circle->angle = (M_PIl * 2.0);
    }

    /* now add more angle for multi turns */
    if (turn > 0) {
	circle->angle += turn * 2.0 * M_PIl;
    }

    return pmErrno = 0;
}

/*
  pmCirclePoint() returns the vector to the point at the given angle along
  the circle. If the circle is a helix or spiral or combination, the
  point will include interpolation off the actual circle.
  */
int pmCirclePoint(PmCircle * circle, double angle, PmCartesian * point)
{
    double scale;
    double rscale;

    /* get scale for spiral, helix interpolation */
    if (circle->angle == 0.0) {
	return pmErrno = PM_DIV_ERR;
    }
    scale = angle / circle->angle;
    rscale = 1.0 + scale * circle->spiral_proportion;

    /* start with center */
    *point = circle->center;

    /* add scaled vector in helix dir */
    pmCartScalMAC(circle->rHelix, scale, point);

    /* compute circular/spiral components rel to center */
    pmCartScalMAC(circle->rTan, cos(angle) * rscale, point);
    pmCartScalMAC(circle->rPerp, sin(angle) * rscale, point);

    return pmErrno = 0;
}

/*
  pmCircleVelocity() returns the derivative of pmCirclePoint by angle.
*/
int pmCircleDerivative(PmCircle * circle, double angle,
		       PmCartesian * velocity)
{
    double scale, dscale, drscale;
    double rscale;
    double cosangle, sinangle;

    /* get scale for spiral, helix interpolation */
    if (circle->angle == 0.0) {
	return pmErrno = PM_DIV_ERR;
    }
    scale = angle / circle->angle;
    dscale = 1.0 / circle->angle;
    rscale = 1.0 + scale * circle->spiral_proportion;
    drscale = dscale * circle->spiral_proportion;
    cosangle = cos(angle);
    sinangle = sin(angle);

    /* start helical part */
    pmCartScalMult(circle->rHelix, dscale, velocity);

    /* compute circular/spiral components rel to center */
    pmCartScalMAC(circle->rTan, cosangle * drscale - sinangle * rscale,
		  velocity);
    pmCartScalMAC(circle->rPerp, sinangle * drscale + cosangle * rscale,
		  velocity);

    return pmErrno = 0;
}
