/********************************************************************
* Description: posemath.h
*
*   Derived from a work by Fred Proctor & Will Shackleford
*
* Author:
* License: LGPL Version 2
* System: Linux
*    
* Copyright (c) 2004 All rights reserved.
*
* Last change: 
* $Revision: 1.2 $
* $Author: paul $
* $Date: 2005/08/01 10:32:35 $
********************************************************************/
#ifndef POSEMATH_H
#define POSEMATH_H

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#ifdef __cplusplus

extern "C" {
#endif

/* PmCartesian */
typedef struct {
    double x, y, z;
} PmCartesian;

#define  PmCartesian_zero {0,0,0}

/* line structure */

typedef struct {
    PmCartesian start;		/* where motion was started */
    PmCartesian end;		/* where motion is going */
} PmLine;

/* Generalized circle structure */

typedef struct {
    PmCartesian center;
    PmCartesian normal;
    PmCartesian rTan;
    PmCartesian rPerp;
    PmCartesian rHelix;
    double radius;
    double angle;
    double spiral_proportion;
} PmCircle;

/* PmRotationMatrix and PmRpy are not required for a trivial kinematics machine.
    However, they may be useful at a later date for implementing coordinate
    transforms for five axis machines that have rotating heads. */
typedef struct {
    PmCartesian x, y, z;
} PmRotationMatrix;

typedef struct {
    double r, p, y;
} PmRpy;

/* These two can be used for jacobean transforms on the two structs above. */
extern int pmRpyMatConvert(PmRpy, PmRotationMatrix *);
extern int pmMatCartMult(PmRotationMatrix m, PmCartesian v, PmCartesian * vout);

/* quicky macros */
#define pmSq(x) ((x)*(x))

#ifdef TO_DEG
#undef TO_DEG
#endif
#define TO_DEG (180.0/M_PIl)

#ifdef TO_RAD
#undef TO_RAD
#endif
#define TO_RAD (M_PIl/180.0)


/* DOUBLE_FUZZ is the smallest double, d, such that (1+d != 1) w/o FPC.
   DOUBLECP_FUZZ is the same only with the Floating Point CoProcessor */
#define DOUBLE_FUZZ 2.2204460492503131e-16
#define CART_FUZZ (0.000001)
/* how close a cartesian vector's magnitude must be for it to be considered
   a zero vector */
#define SQRT_FUZZ (-.000001)
/* how close to 0 before math_sqrt() is error */
#define CIRCLE_FUZZ (0.000001)
/* Bug fix for the missing circles problem */
#define V_FUZZ (0.000001)


/* global error number and errors */
extern int pmErrno;

#define PM_ERR             -1	/* unspecified error */
#define PM_IMPL_ERR        -2	/* not implemented */
#define PM_NORM_ERR        -3	/* arg should have been norm */
#define PM_DIV_ERR         -4	/* divide by zero error */

/* Scalar functions */

extern double pmSqrt(double x);


/* Arithmetic functions
*/

/* translation functions */
extern int pmCartCartCompare(PmCartesian, PmCartesian);
extern int pmCartCartDot(PmCartesian, PmCartesian, double *);
extern int pmCartCartCross(PmCartesian, PmCartesian, PmCartesian *);
extern int pmCartMag(PmCartesian, double *);
extern double pmCartGetMag(PmCartesian);
extern int pmCartCartDisp(PmCartesian v1, PmCartesian v2, double *d);
extern int pmCartCartAdd(PmCartesian, PmCartesian, PmCartesian *);
extern int pmCartCartInterpolate(PmCartesian v1, PmCartesian v2,
				 double t, PmCartesian * vout);
extern int pmCartCartSub(PmCartesian, PmCartesian, PmCartesian *);
extern int pmCartScalMult(PmCartesian, double, PmCartesian *);
extern int pmCartScalMAC(PmCartesian, double, PmCartesian *);
extern int pmCartScalDiv(PmCartesian, double, PmCartesian *);
extern int pmCartNeg(PmCartesian, PmCartesian *);
extern int pmCartUnit(PmCartesian v, PmCartesian * vout);
extern int pmCartIsNorm(PmCartesian v);
extern int pmCartInv(PmCartesian, PmCartesian *);
extern int pmCartCartProj(PmCartesian, PmCartesian, PmCartesian *);
extern int pmCartPlaneProj(PmCartesian v, PmCartesian normal,
			   PmCartesian * vout);

/* line functions */
extern int pmLineInit(PmLine * line, PmCartesian start, PmCartesian end);

/* circle functions */
extern int pmCircleInit(PmCircle * circle, PmCartesian start, PmCartesian end,
			PmCartesian center, PmCartesian normal, int turn);
extern int pmCirclePoint(PmCircle * circle, double angle, PmCartesian * point);
extern int pmCircleDerivative(PmCircle * circle, double angle, PmCartesian * point);

#ifdef __cplusplus
}
#endif
#endif
